@@ -1117,7 +1117,7 @@ namespace System.Windows.Forms
 			
 			if (textbox_ctrl != null) {
 				textbox_ctrl.SetSelectable (false);
-				textbox_ctrl.ShowSelection = true;
+				textbox_ctrl.ShowSelection = Enabled;
 				textbox_ctrl.ActivateCaret (true);
 				textbox_ctrl.SelectAll ();
 			}
@@ -1504,6 +1504,8 @@ namespace System.Windows.Forms
 			// No edit control, we paint the edit ourselves
 			if (dropdown_style == ComboBoxStyle.DropDownList) {
 				DrawItemState state = DrawItemState.None;
+				Color back_color = BackColor;
+				Color fore_color = ForeColor;
 				Rectangle item_rect = text_area;
 				item_rect.X += border;
 				item_rect.Y += border;
@@ -1513,10 +1515,12 @@ namespace System.Windows.Forms
 				if (Focused) {
 					state = DrawItemState.Selected;
 					state |= DrawItemState.Focus;
+					back_color = SystemColors.Highlight;
+					fore_color = SystemColors.HighlightText;
 				}
 				
 				state |= DrawItemState.ComboBoxEdit;
-				HandleDrawItem (new DrawItemEventArgs (dc, Font, item_rect, SelectedIndex, state, ForeColor, BackColor));
+				HandleDrawItem (new DrawItemEventArgs (dc, Font, item_rect, SelectedIndex, state, fore_color, back_color));
 			}
 			
 			if (show_dropdown_button) {
@@ -1598,6 +1602,12 @@ namespace System.Windows.Forms
 				listbox_ctrl.Dispose ();
 				listbox_ctrl = null;
 			}
+#if NET_2_0
+			 // The auto complete list could have been shown after the listbox,
+			 // so make sure it's hidden.
+			 if (textbox_ctrl != null)
+				 textbox_ctrl.HideAutoCompleteList ();
+#endif
 		}
 		
 		private int FindStringCaseInsensitive (string search)
@@ -2307,12 +2317,17 @@ namespace System.Windows.Forms
 			{
 				this.owner = owner;
 				ShowSelection = false;
-				HideSelection = false;
+				owner.EnabledChanged += OwnerEnabledChangedHandler;
 #if NET_2_0
 				owner.LostFocus += OwnerLostFocusHandler;
 #endif
 			}
 
+			void OwnerEnabledChangedHandler (object o, EventArgs args)
+			{
+				ShowSelection = owner.Focused && owner.Enabled;
+			}
+
 #if NET_2_0
 			void OwnerLostFocusHandler (object o, EventArgs args)
 			{
@@ -2522,11 +2537,11 @@ namespace System.Windows.Forms
 				else { // DropDown or DropDownList
 					
 					width = owner.DropDownWidth;
-					int count = (owner.Items.Count <= owner.MaxDropDownItems) ? owner.Items.Count : owner.MaxDropDownItems;
+					int visible_items_count = (owner.Items.Count <= owner.MaxDropDownItems) ? owner.Items.Count : owner.MaxDropDownItems;
 					
 					if (owner.DrawMode == DrawMode.OwnerDrawVariable) {
 						height = 0;
-						for (int i = 0; i < count; i++) {
+						for (int i = 0; i < visible_items_count; i++) {
 							height += owner.GetItemHeight (i);
 						}
 
@@ -2535,14 +2550,15 @@ namespace System.Windows.Forms
 					} else	{
 #if NET_2_0
 						if (owner.DropDownHeight == default_drop_down_height) { // ignore DropDownHeight
-							height = owner.ItemHeight * count;
+							height = owner.ItemHeight * visible_items_count;
 							show_scrollbar = owner.Items.Count > owner.MaxDropDownItems;
 						} else {
+							// ignore visible items count, and use manual height instead
 							height = owner.DropDownHeight;
-							show_scrollbar = (count * owner.ItemHeight) > height;
+							show_scrollbar = (owner.Items.Count * owner.ItemHeight) > height;
 						}
 #else		
-						height = owner.ItemHeight * count;
+						height = owner.ItemHeight * visible_items_count;
 						show_scrollbar = owner.Items.Count > owner.MaxDropDownItems;
 #endif
 					}
@@ -2613,9 +2629,13 @@ namespace System.Windows.Forms
 							continue;
 
 						DrawItemState state = DrawItemState.None;
+						Color back_color = owner.BackColor;
+						Color fore_color = owner.ForeColor;
 
 						if (i == HighlightedIndex) {
 							state |= DrawItemState.Selected;
+							back_color = SystemColors.Highlight;
+							fore_color = SystemColors.HighlightText;
 							
 							if (owner.DropDownStyle == ComboBoxStyle.DropDownList) {
 								state |= DrawItemState.Focus;
@@ -2623,7 +2643,7 @@ namespace System.Windows.Forms
 						}
 
 						owner.HandleDrawItem (new DrawItemEventArgs (dc, owner.Font, item_rect,
-							i, state, owner.ForeColor, owner.BackColor));
+							i, state, fore_color, back_color));
 					}
 				}
 			}
