@@ -1359,6 +1359,21 @@ namespace System.Windows.Forms {
 		_NET_WM_STATE_TOGGLE		= 2
 	}
 
+	internal enum NetWmMoveResize {
+		_NET_WM_MOVERESIZE_SIZE_TOPLEFT = 	0,
+		_NET_WM_MOVERESIZE_SIZE_TOP = 		1,
+		_NET_WM_MOVERESIZE_SIZE_TOPRIGHT = 	2,
+		_NET_WM_MOVERESIZE_SIZE_RIGHT = 	3,
+		_NET_WM_MOVERESIZE_SIZE_BOTTOMRIGHT = 	4,
+		_NET_WM_MOVERESIZE_SIZE_BOTTOM = 	5,
+		_NET_WM_MOVERESIZE_SIZE_BOTTOMLEFT =	6,
+		_NET_WM_MOVERESIZE_SIZE_LEFT = 		7,
+		_NET_WM_MOVERESIZE_MOVE = 		8,
+		_NET_WM_MOVERESIZE_SIZE_KEYBOARD = 	9,
+		_NET_WM_MOVERESIZE_MOVE_KEYBOARD = 	10,
+		_NET_WM_MOVERESIZE_CANCEL = 		11
+	}
+
 	[Flags]
 	internal enum XSizeHintsFlags  {
 		USPosition			= (1 << 0),
@@ -1478,16 +1493,6 @@ namespace System.Windows.Forms {
 		internal Rectangle	Area;			// The area the current grab is confined to
 	}
 
-	internal struct ClipboardStruct {
-		internal object		Source;
-		internal object		Item;			// Object on the clipboard
-		internal object		Type;			// Type if object on the clipboard
-		internal ArrayList	Formats;		// list of formats available in the clipboard
-		internal bool		Retrieving;		// true if we are requesting an item
-		internal bool		Enumerating;		// true if we are enumerating through all known types
-		internal XplatUI.ObjectToClipboard Converter;
-	}
-
 	internal delegate int  XErrorHandler(IntPtr DisplayHandle, ref XErrorEvent error_event);
 
 	internal enum XRequest : byte {
