@@ -30,9 +30,19 @@ using System;
 using System.Drawing;
 using System.Runtime.InteropServices;
 
-namespace System.Windows.Forms {
+#if PUBLIC_TYPES
+namespace Mono.Winforms 
+#else
+namespace System.Windows.Forms
+#endif
+{
 	[Flags]
-	internal enum WindowStyles : int {
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum WindowStyles : int {
 		WS_OVERLAPPED		= 0x00000000,
 		WS_POPUP		= unchecked((int)0x80000000),
 		WS_CHILD		= 0x40000000,
@@ -63,7 +73,12 @@ namespace System.Windows.Forms {
 	}
 
 	[Flags]
-	internal enum WindowExStyles : int {
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum WindowExStyles : int {
 		// Extended Styles
 		WS_EX_DLGMODALFRAME	= 0x00000001,
 		WS_EX_DRAGDETECT	= 0x00000002,
@@ -99,14 +114,24 @@ namespace System.Windows.Forms {
 	}
 
 	[Flags]
-	internal enum ToolBarStyles : int {
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum ToolBarStyles : int {
 		TBSTYLE_TOOLTIPS    = 0x100,
 		TBSTYLE_FLAT        = 0x800,
 		TBSTYLE_LIST        = 0x1000,
 		TBSTYLE_TRANSPARENT = 0x8000
 	}
-	
-	internal enum Msg {
+
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum Msg {
 		WM_NULL                   = 0x0000,
 		WM_CREATE                 = 0x0001,
 		WM_DESTROY                = 0x0002,
@@ -358,7 +383,12 @@ namespace System.Windows.Forms {
 		NIN_BALLOONUSERCLICK      = WM_USER + 0x0005 
 	}
 
-	internal enum MsgButtons {
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum MsgButtons {
 		MK_LBUTTON		= 0x0001,
 		MK_RBUTTON		= 0x0002,
 		MK_SHIFT		= 0x0004,
@@ -367,8 +397,13 @@ namespace System.Windows.Forms {
 		MK_XBUTTON1		= 0x0020,
 		MK_XBUTTON2		= 0x0040,
 	}
-	
-	internal enum MsgUIState {
+
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum MsgUIState {
 		UIS_SET        = 1,
 		UIS_CLEAR      = 2,
 		UIS_INITIALIZE = 3,
@@ -378,17 +413,22 @@ namespace System.Windows.Forms {
 	}
 
 	[StructLayout(LayoutKind.Sequential)]
-	internal struct POINT {
-		internal int x;
-		internal int y;
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	struct POINT {
+		public int x;
+		public int y;
 
-		internal POINT (int x, int y)
+		public POINT (int x, int y)
 		{
 			this.x = x;
 			this.y = y;
 		}
 
-		internal Point ToPoint ()
+		public Point ToPoint ()
 		{
 			return new Point (x, y);
 		}
@@ -400,7 +440,12 @@ namespace System.Windows.Forms {
 	}
 	
 	[StructLayout(LayoutKind.Sequential)] 
-	internal struct MSG {
+#if PUBLIC_TYPES
+        public
+#else
+	internal
+#endif
+	struct MSG {
 		internal IntPtr	hwnd;
 		internal Msg	message;
 		internal IntPtr	wParam; 
@@ -416,26 +461,46 @@ namespace System.Windows.Forms {
 	}
 
 	[Flags]
-	internal enum TransparencySupport {
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum TransparencySupport {
 		None = 0x00,
 		Get = 0x01,
 		Set = 0x02,
 		GetSet = 0x03
 	}
 
-	internal enum WindowActiveFlags {
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum WindowActiveFlags {
 		WA_INACTIVE		= 0,
 		WA_ACTIVE		= 1,
 		WA_CLICKACTIVE		= 2
 	}
 
-	internal enum KeybdEventFlags {
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum KeybdEventFlags {
 		None			= 0,
 		ExtendedKey		= 0x0001,
 		KeyUp			= 0x0002
 	}
 
-	internal enum VirtualKeys {
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum VirtualKeys {
 		VK_LBUTTON		= 0x01,
 		VK_RBUTTON              = 0x02,
 		VK_CANCEL		= 0x03,
@@ -590,7 +655,12 @@ namespace System.Windows.Forms {
 		VK_OEM_CLEAR		= 0xFE,
 	}
 
-	internal enum TtyKeys {
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum TtyKeys {
 		XK_BackSpace		= 0xff08,  /* Back space, back char */
 		XK_Tab			= 0xff09,
 		XK_Linefeed		= 0xff0a,  /* Linefeed, LF */
@@ -603,7 +673,12 @@ namespace System.Windows.Forms {
 		XK_Delete		= 0xffff  /* Delete, rubout */
 	}
 
-	internal enum MiscKeys {
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum MiscKeys {
 		XK_ISO_Lock             = 0xfe01,
 		XK_ISO_Last_Group_Lock	= 0xfe0f,
 		XK_Select		= 0xff60,
@@ -622,7 +697,12 @@ namespace System.Windows.Forms {
 		XK_Num_Lock		= 0xff7f
 	}
 
-	internal enum KeypadKeys {
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum KeypadKeys {
 		XK_KP_Space		= 0xff80,
 		XK_KP_Tab		= 0xff89,
 		XK_KP_Enter		= 0xff8d,  /* Enter */
@@ -663,7 +743,12 @@ namespace System.Windows.Forms {
 		XK_KP_9			= 0xffb9
 	}
 
-        internal enum DeadKeys {
+#if PUBLIC_TYPES
+	public
+#else
+        internal
+#endif
+	enum DeadKeys {
                 XK_dead_grave           = 0xfe50,
                 XK_dead_acute           = 0xfe51,
                 XK_dead_circumflex      = 0xfe52,
@@ -687,7 +772,12 @@ namespace System.Windows.Forms {
         }
 
 	[StructLayout(LayoutKind.Sequential)]
-	internal struct HELPINFO {
+#if PUBLIC_TYPES
+	 public
+#else
+	internal
+#endif
+	struct HELPINFO {
 		internal uint		cbSize;
 		internal int		iContextType;
 		internal int		iCtrlId;
@@ -696,13 +786,23 @@ namespace System.Windows.Forms {
 		internal POINT		MousePos;
 	}
 
-	internal enum PeekMessageFlags {
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum PeekMessageFlags {
 		PM_NOREMOVE			= 0x00000000,
 		PM_REMOVE			= 0x00000001,
 		PM_NOYIELD			= 0x00000002
 	}
 
-	internal enum StdCursor {
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum StdCursor {
 		Default				= 0,
 		AppStarting			= 1,
 		Arrow				= 2,
@@ -733,7 +833,12 @@ namespace System.Windows.Forms {
 		WaitCursor			= 27
 	}
 
-	internal enum HitTest {
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum HitTest {
 		HTERROR				= -2,
 		HTTRANSPARENT			= -1,
 		HTNOWHERE			= 0,
@@ -765,14 +870,24 @@ namespace System.Windows.Forms {
 		HTHELP				= 21
 	}
 
-	internal enum TitleStyle {
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum TitleStyle {
 		None				= 0,
 		Normal				= 1,
 		Tool				= 2
 	}
 
 	[StructLayout(LayoutKind.Sequential)]
-	internal struct BITMAPINFOHEADER {
+#if PUBLIC_TYPES
+	 public
+#else
+	internal
+#endif
+	struct BITMAPINFOHEADER {
 		internal uint     biSize;
 		internal int      biWidth;
 		internal int      biHeight;
@@ -786,7 +901,12 @@ namespace System.Windows.Forms {
 		internal uint     biClrImportant;
 	}
 
-	internal enum ClipboardFormats : ushort {
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum ClipboardFormats : ushort {
 		CF_TEXT				= 1,
 		CF_BITMAP           		= 2,
 		CF_METAFILEPICT     		= 3,
@@ -807,7 +927,12 @@ namespace System.Windows.Forms {
 	}
 
 	[StructLayout(LayoutKind.Sequential, CharSet=CharSet.Unicode)]
-	internal struct MINMAXINFO {
+#if PUBLIC_TYPES
+	 public
+#else
+	internal
+#endif
+	struct MINMAXINFO {
 		internal POINT			ptReserved;
 		internal POINT			ptMaxSize;
 		internal POINT			ptMaxPosition;
@@ -815,7 +940,26 @@ namespace System.Windows.Forms {
 		internal POINT			ptMaxTrackSize;
 	}
 
-	internal enum ScrollBarCommands {
+	[StructLayout(LayoutKind.Sequential, CharSet=CharSet.Unicode)]
+#if PUBLIC_TYPES
+	 public
+#else
+	internal
+#endif
+	struct KeyFilterData {
+			internal bool Down;
+			internal int keycode;
+			internal int keysym;
+			internal Keys ModifierKeys;
+			internal String str;
+		}
+
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum ScrollBarCommands {
 		SB_LINEUP		= 0,
 		SB_LINELEFT		= 0,
 		SB_LINEDOWN		= 1,
@@ -833,7 +977,12 @@ namespace System.Windows.Forms {
 		SB_ENDSCROLL		= 8
 	}
 
-	internal enum ClipCombineMode {
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum ClipCombineMode {
 		RGN_AND = 1,
 		RGN_OR,
 		RGN_XOR,
@@ -844,7 +993,12 @@ namespace System.Windows.Forms {
 		RGN_MAX = RGN_COPY
 	}
 
-	internal enum SystemCommands {
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum SystemCommands {
 		SC_SIZE         = 0xF000,
 		SC_MOVE         = 0xF010,
 		SC_MINIMIZE     = 0xF020,
@@ -865,5 +1019,18 @@ namespace System.Windows.Forms {
 		SC_MONITORPOWER = 0xF170,
 		SC_CONTEXTHELP  = 0xF180
 	}
+
+#if PUBLIC_TYPES
+	public
+#else
+	internal
+#endif
+	enum AlertType {
+		Default		= 1,
+		Error		= 2,
+		Question	= 3,
+		Warning		= 4,
+		Information	= 5
+	}
 }
 
