@@ -159,6 +159,7 @@
 using System;
 using System.Collections;
 using System.Data;
+using System.Data.Common;
 using System.Globalization;
 using System.Xml;
 using System.Xml.Schema;
@@ -207,7 +208,7 @@ namespace System.Data
 			}
 		}
 	}
-
+	
 	internal class TableStructure
 	{
 		public TableStructure (DataTable table)
@@ -320,7 +321,11 @@ namespace System.Data
 		ArrayList targetElements = new ArrayList ();
 
 		TableStructure currentTable;
-
+	
+#if NET_2_0
+		// TODO: Do we need a collection here?
+		TableAdapterSchemaInfo currentAdapter;
+#endif
 		#endregion
 
 		// .ctor()
@@ -336,6 +341,13 @@ namespace System.Data
 			schema.Compile (null);
 		}
 
+#if NET_2_0
+		// properties
+		internal TableAdapterSchemaInfo CurrentAdapter {
+			get { return currentAdapter; }
+		}
+#endif
+		
 		// methods
 
 		public void Process ()
@@ -523,6 +535,13 @@ namespace System.Data
 						attr.NamespaceURI == XmlConstants.MsdataNamespace)
 						useCurrent = true;
 #endif
+
+					if (attr.NamespaceURI == XmlConstants.MspropNamespace && 
+					   !dataset.ExtendedProperties.ContainsKey(attr.Name)) {
+					       dataset.ExtendedProperties.Add (attr.Name, attr.Value);
+					       continue;
+					}
+					
 					if (attr.LocalName == "Locale" &&
 						attr.NamespaceURI == XmlConstants.MsdataNamespace) {
 						CultureInfo ci = new CultureInfo (attr.Value);
@@ -581,6 +600,12 @@ namespace System.Data
 			// Find Locale
 			if (el.UnhandledAttributes != null) {
 				foreach (XmlAttribute attr in el.UnhandledAttributes) {
+					
+					if (attr.NamespaceURI == XmlConstants.MspropNamespace) {
+					       table.ExtendedProperties.Add (attr.Name, attr.Value);
+					       continue;
+					}
+
 					if (attr.LocalName == "Locale" &&
 						attr.NamespaceURI == XmlConstants.MsdataNamespace)
 						table.Locale = new CultureInfo (attr.Value);
@@ -771,6 +796,11 @@ namespace System.Data
 		{
 			if (obj.UnhandledAttributes != null) {
 				foreach (XmlAttribute attr in obj.UnhandledAttributes) {
+					if (attr.NamespaceURI == XmlConstants.MspropNamespace) {
+					       col.ExtendedProperties.Add (attr.Name, attr.Value);
+					       continue;
+					}
+
 					if (attr.NamespaceURI != XmlConstants.MsdataNamespace)
 						continue;
 					switch (attr.LocalName) {
@@ -843,17 +873,28 @@ namespace System.Data
 
 		private void AddParentKeyColumn (XmlSchemaElement parent, XmlSchemaElement el, DataColumn col)
 		{
-			if (currentTable.PrimaryKey != null)
+			// check existing primary key
+			if (currentTable.Table.PrimaryKey.Length > 0)
+				throw new DataException (String.Format ("There is already primary key columns in the table \"{0}\".", currentTable.Table.TableName));
+
+			if (currentTable.PrimaryKey != null) {
+				// fill pk column info and return
+				col.ColumnName = currentTable.PrimaryKey.ColumnName;
+				col.ColumnMapping = currentTable.PrimaryKey.ColumnMapping;
+				col.Namespace = currentTable.PrimaryKey.Namespace;
+				col.DataType = currentTable.PrimaryKey.DataType;
+				col.AutoIncrement = currentTable.PrimaryKey.AutoIncrement;
+				col.AllowDBNull = currentTable.PrimaryKey.AllowDBNull;
+				
+				ImportColumnMetaInfo (el, el.QualifiedName, col);
 				return;
+			}
 
 			// check name identity
 			string name = XmlHelper.Decode (parent.QualifiedName.Name) + "_Id";
 			int count = 0;
 			while (currentTable.ContainsColumn (name))
 				name = String.Format ("{0}_{1}", name, count++);
-			// check existing primary key
-			if (currentTable.Table.PrimaryKey.Length > 0)
-				throw new DataException (String.Format ("There is already primary key columns in the table \"{0}\".", currentTable.Table.TableName));
 
 			col.ColumnName = name;
 			col.ColumnMapping = MappingType.Hidden;
@@ -861,8 +902,8 @@ namespace System.Data
 			col.DataType = typeof (int);
 			col.AutoIncrement = true;
 			col.AllowDBNull = false;
-
 			ImportColumnMetaInfo (el, el.QualifiedName, col);
+			
 			AddColumn (col);
 			currentTable.PrimaryKey = col;
 		}
@@ -1207,13 +1248,320 @@ namespace System.Data
 				if (ai != null) {
 					foreach (XmlNode n in ai.Markup) {
 						XmlElement el = n as XmlElement;
+						
+						// #325464 debugging
+						//Console.WriteLine ("Name: " + el.LocalName + " NS: " + el.NamespaceURI + " Const: " + XmlConstants.MsdataNamespace);
 						if (el != null && el.LocalName == "Relationship" && el.NamespaceURI == XmlConstants.MsdataNamespace)
 							HandleRelationshipAnnotation (el, nested);
+#if NET_2_0
+						if (el != null && el.LocalName == "DataSource" && el.NamespaceURI == XmlConstants.MsdatasourceNamespace)
+							HandleDataSourceAnnotation (el, nested);
+#endif
 					}
 				}
 			}
 		}
 
+#if NET_2_0
+		private void HandleDataSourceAnnotation (XmlElement el, bool nested)
+		{
+			// Handle: Connections and Tables
+			// For Tables: extract the provider information from connection and use
+			// the corresponding providerfactory to create the adapter and et al objects 
+			// and populate them
+			
+			// #325464 debugging
+			//Console.WriteLine ("In HandleDataSourceAnnotation... ");
+			string providerName = null;
+			string connString = null;
+			DbProviderFactory provider = null;
+			XmlElement e, tablesElement = null, firstChild;
+			
+			foreach (XmlNode n in el.ChildNodes) {
+				e = n as XmlElement;
+				
+				if (e == null)
+					continue;
+				
+#if !MONOTOUCH
+				if (e.LocalName == "Connections" && (firstChild = e.FirstChild as XmlElement) != null) {
+					providerName = firstChild.GetAttribute ("Provider");
+					connString = firstChild.GetAttribute ("AppSettingsPropertyName");
+					provider = DbProviderFactories.GetFactory (providerName);
+					continue;
+				}
+#endif // !MONOTOUCH
+				// #325464 debugging
+				//Console.WriteLine ("ProviderName: " + providerName + "Connstr: " + connString);
+				
+				if (e.LocalName == "Tables")
+					tablesElement = e;
+			}
+				
+			if (tablesElement != null && provider != null) {
+				foreach (XmlNode node in tablesElement.ChildNodes) {
+					ProcessTableAdapter (node as XmlElement, provider, connString);
+				}
+			}
+		}
+		
+		private void ProcessTableAdapter (XmlElement el, DbProviderFactory provider, string connStr)
+		{
+			XmlElement e;
+			string datasetTableName = null;
+			
+			if (el == null)
+				return;
+			
+			// #325464 debugging
+			//Console.WriteLine ("in ProcessTableAdapters...");
+			currentAdapter = new TableAdapterSchemaInfo (provider); 
+			currentAdapter.ConnectionString = connStr;
+			
+			//Console.WriteLine ("Provider: {0}, connection: {1}, adapter: {2}", 
+			//                   provider, currentAdapter.Connection, currentAdapter.Adapter);
+			currentAdapter.BaseClass = el.GetAttribute ("BaseClass");
+			datasetTableName = el.GetAttribute ("Name");
+			currentAdapter.Name = el.GetAttribute ("GeneratorDataComponentClassName");
+			
+			if (String.IsNullOrEmpty (currentAdapter.Name))
+				currentAdapter.Name = el.GetAttribute ("DataAccessorName");
+
+			//Console.WriteLine ("Name: "+currentAdapter.Name);
+			foreach (XmlNode n in el.ChildNodes) {
+				e = n as XmlElement;
+				
+				//Console.WriteLine ("Children of Tables: "+e.LocalName);
+				if (e == null)
+					continue;
+				
+				switch (e.LocalName) {
+					case "MainSource": 
+					case "Sources": 
+						foreach (XmlNode msn in e.ChildNodes)
+							ProcessDbSource (msn as XmlElement);
+						break;
+					
+					case "Mappings":
+						DataTableMapping tableMapping = new DataTableMapping ();
+						tableMapping.SourceTable = "Table";
+						tableMapping.DataSetTable = datasetTableName;
+						
+						foreach (XmlNode mps in e.ChildNodes)
+							ProcessColumnMapping (mps as XmlElement, tableMapping);
+						
+						currentAdapter.Adapter.TableMappings.Add (tableMapping);
+						break;						
+				}
+			}
+		}
+		
+		private void ProcessDbSource (XmlElement el)
+		{
+			
+			string cmdType;
+			string tmp = null;
+			XmlElement e;
+			
+			if (el == null)
+				return;
+			
+			//Console.WriteLine ("ProcessDbSources: "+el.LocalName);
+
+			tmp = el.GetAttribute ("GenerateShortCommands");
+			//Console.WriteLine ("GenerateShortCommands: {0}", tmp);
+			if (!String.IsNullOrEmpty (tmp))
+				currentAdapter.ShortCommands = Convert.ToBoolean (tmp);
+		
+			DbCommandInfo cmdInfo = new DbCommandInfo ();
+			tmp = el.GetAttribute ("GenerateMethods");
+			if (!String.IsNullOrEmpty (tmp)) {
+				DbSourceMethodInfo mthdInfo = null;
+				
+				switch ((GenerateMethodsType) Enum.Parse (typeof (GenerateMethodsType), tmp)) {
+				case GenerateMethodsType.Get:
+					mthdInfo = new DbSourceMethodInfo ();
+					mthdInfo.Name = el.GetAttribute ("GetMethodName");
+					mthdInfo.Modifier = el.GetAttribute ("GetMethodModifier");
+					if (String.IsNullOrEmpty (mthdInfo.Modifier))
+						mthdInfo.Modifier = "Public";
+					mthdInfo.ScalarCallRetval = el.GetAttribute ("ScalarCallRetval");
+					mthdInfo.QueryType = el.GetAttribute ("QueryType");
+					mthdInfo.MethodType = GenerateMethodsType.Get;
+					cmdInfo.Methods = new DbSourceMethodInfo [1];
+					cmdInfo.Methods[0] = mthdInfo;
+					break;
+					
+				case GenerateMethodsType.Fill:
+					mthdInfo = new DbSourceMethodInfo ();
+					mthdInfo.Name = el.GetAttribute ("FillMethodName");
+					mthdInfo.Modifier = el.GetAttribute ("FillMethodModifier");
+					if (String.IsNullOrEmpty (mthdInfo.Modifier))
+						mthdInfo.Modifier = "Public";
+					mthdInfo.ScalarCallRetval = null;
+					mthdInfo.QueryType = null;
+					mthdInfo.MethodType = GenerateMethodsType.Fill;
+					cmdInfo.Methods = new DbSourceMethodInfo [1];
+					cmdInfo.Methods[0] = mthdInfo;
+					break;
+					
+				case GenerateMethodsType.Both:
+					mthdInfo = new DbSourceMethodInfo ();
+					// Get
+					mthdInfo.Name = el.GetAttribute ("GetMethodName");
+					mthdInfo.Modifier = el.GetAttribute ("GetMethodModifier");
+					if (String.IsNullOrEmpty (mthdInfo.Modifier))
+						mthdInfo.Modifier = "Public";
+					mthdInfo.ScalarCallRetval = el.GetAttribute ("ScalarCallRetval");
+					mthdInfo.QueryType = el.GetAttribute ("QueryType");
+					mthdInfo.MethodType = GenerateMethodsType.Get;
+					cmdInfo.Methods = new DbSourceMethodInfo [2];
+					cmdInfo.Methods[0] = mthdInfo;
+					
+					// Fill
+					mthdInfo = new DbSourceMethodInfo ();
+					mthdInfo.Name = el.GetAttribute ("FillMethodName");
+					mthdInfo.Modifier = el.GetAttribute ("FillMethodModifier");
+					if (String.IsNullOrEmpty (mthdInfo.Modifier))
+						mthdInfo.Modifier = "Public";
+					mthdInfo.ScalarCallRetval = null;
+					mthdInfo.QueryType = null;
+					mthdInfo.MethodType = GenerateMethodsType.Fill;
+					cmdInfo.Methods[1] = mthdInfo;
+					break;
+				}
+			} else {
+				// no Get or Fill methods - non <MainSource> sources
+				DbSourceMethodInfo mthdInfo = new DbSourceMethodInfo ();
+				mthdInfo.Name = el.GetAttribute ("Name");
+				mthdInfo.Modifier = el.GetAttribute ("Modifier");
+				if (String.IsNullOrEmpty (mthdInfo.Modifier))
+					mthdInfo.Modifier = "Public";
+				mthdInfo.ScalarCallRetval = el.GetAttribute ("ScalarCallRetval");
+				mthdInfo.QueryType = el.GetAttribute ("QueryType");
+				mthdInfo.MethodType = GenerateMethodsType.None;
+				// Add MethodInfo to DbCommandInfo
+				cmdInfo.Methods = new DbSourceMethodInfo [1];
+				cmdInfo.Methods[0] = mthdInfo;
+			}
+			
+			foreach (XmlNode n in el.ChildNodes) {
+				e = n as XmlElement;
+				
+				if (e == null) 
+					continue;
+				
+				switch (e.LocalName) {
+					case "SelectCommand": 
+						cmdInfo.Command = ProcessDbCommand (e.FirstChild as XmlElement);
+						currentAdapter.Commands.Add (cmdInfo);
+						break;
+					case "InsertCommand": 
+						currentAdapter.Adapter.InsertCommand = ProcessDbCommand (e.FirstChild as XmlElement);
+						break;
+					case "UpdateCommand": 
+						currentAdapter.Adapter.UpdateCommand = ProcessDbCommand (e.FirstChild as XmlElement);
+						break;
+					case "DeleteCommand": 
+						currentAdapter.Adapter.DeleteCommand = ProcessDbCommand (e.FirstChild as XmlElement);
+						break;
+				}
+			}
+		}
+		
+		private DbCommand ProcessDbCommand (XmlElement el)
+		{
+			XmlElement e;
+			//Console.WriteLine (el.LocalName);
+			string cmdText = null;
+			string cmdType = null;
+			ArrayList parameters = null;
+			
+			if (el == null)
+				return null;
+			
+			cmdType = el.GetAttribute ("CommandType");
+			foreach (XmlNode n in el.ChildNodes) {
+				e = n as XmlElement;
+				if (e != null && e.LocalName == "CommandText")
+					cmdText = e.InnerText;
+				else if (e != null && e.LocalName == "Parameters" && !e.IsEmpty)
+					parameters = ProcessDbParameters (e);
+			}
+			
+			DbCommand cmd = currentAdapter.Provider.CreateCommand ();
+			cmd.CommandText = cmdText;
+			if (cmdType == "StoredProcedure")
+				cmd.CommandType = CommandType.StoredProcedure;
+			else
+				cmd.CommandType = CommandType.Text;
+
+			if (parameters != null)
+				cmd.Parameters.AddRange (parameters.ToArray ());
+			
+			//Console.WriteLine ("Parameters count: {0}", cmd.Parameters.Count);
+			return cmd;
+		}
+		
+		private ArrayList ProcessDbParameters (XmlElement el)
+		{
+			//Console.WriteLine ("ProcessDbParameters: "+el.LocalName);
+			string tmp = null;
+			XmlElement e;
+			DbParameter param = null;
+			ArrayList parameters = new ArrayList ();
+			
+			if (el == null)
+				return parameters;
+			
+			foreach (XmlNode n in el.ChildNodes) {
+				e = n as XmlElement;
+				
+				if (e == null)
+					continue;
+				param = currentAdapter.Provider.CreateParameter ();
+
+				tmp = e.GetAttribute ("AllowDbNull");
+				if (!String.IsNullOrEmpty (tmp))
+					param.IsNullable = Convert.ToBoolean (tmp);
+				
+				param.ParameterName = e.GetAttribute ("ParameterName");
+				tmp = e.GetAttribute ("ProviderType");
+				if (!String.IsNullOrEmpty (tmp))
+					tmp = e.GetAttribute ("DbType");
+				param.FrameworkDbType = tmp;
+				
+				tmp = e.GetAttribute ("Direction");
+				param.Direction = (ParameterDirection) Enum.Parse (typeof (ParameterDirection), tmp);
+				
+				((IDbDataParameter)param).Precision = Convert.ToByte (e.GetAttribute ("Precision"));
+				((IDbDataParameter)param).Scale = Convert.ToByte (e.GetAttribute ("Scale"));
+				param.Size = Convert.ToInt32 (e.GetAttribute ("Size"));
+				param.SourceColumn = e.GetAttribute ("SourceColumn");
+				
+				tmp = e.GetAttribute ("SourceColumnNullMapping");
+				if (!String.IsNullOrEmpty (tmp))
+					param.SourceColumnNullMapping = Convert.ToBoolean (tmp);
+				
+				tmp = e.GetAttribute ("SourceVersion");
+				param.SourceVersion = (DataRowVersion) Enum.Parse (typeof (DataRowVersion), tmp);				
+				parameters.Add (param);
+			}
+			
+			return parameters;
+		}
+
+		private void ProcessColumnMapping (XmlElement el, DataTableMapping tableMapping)
+		{
+			if (el == null)
+				return;
+			
+			tableMapping.ColumnMappings.Add (el.GetAttribute ("SourceColumn"), 
+			                                 el.GetAttribute ("DataSetColumn"));
+		}
+		
+#endif
+		
 		private void HandleRelationshipAnnotation (XmlElement el, bool nested)
 		{
 			string name = el.GetAttribute ("name");
