@@ -1,10 +1,12 @@
 //
 // ScriptManager.cs
 //
-// Author:
+// Authors:
 //   Igor Zelmanovich <igorz@mainsoft.com>
+//   Marek Habersack <grendel@twistedcode.net>
 //
 // (C) 2007 Mainsoft, Inc.  http://www.mainsoft.com
+// (C) 2007-2010 Novell, Inc (http://novell.com/)
 //
 //
 // Permission is hereby granted, free of charge, to any person obtaining
@@ -244,7 +246,7 @@ namespace System.Web.UI
 			}
 		}
 
-		bool IsDeploymentRetail {
+		internal bool IsDeploymentRetail {
 			get {
 #if TARGET_J2EE
 				return false;
@@ -395,9 +397,17 @@ namespace System.Web.UI
 		public static ScriptManager GetCurrent (Page page) {
 			if (page == null)
 				throw new ArgumentNullException ("page");
-			return (ScriptManager) page.Items [ScriptManagerKey];
+			return GetCurrentInternal (page);
 		}
 
+		static ScriptManager GetCurrentInternal (Page page)
+		{
+			if (page == null)
+				return null;
+
+			return (ScriptManager) page.Items [ScriptManagerKey];
+		}
+		
 		static void SetCurrent (Page page, ScriptManager instance) {
 			page.Items [ScriptManagerKey] = instance;
 			page.ClientScript.RegisterWebFormClientScript ();
@@ -446,7 +456,7 @@ namespace System.Web.UI
 		protected override void OnInit (EventArgs e) {
 			base.OnInit (e);
 
-			if (GetCurrent (Page) != null)
+			if (GetCurrentInternal (Page) != null)
 				throw new InvalidOperationException ("Only one instance of a ScriptManager can be added to the page.");
 
 			SetCurrent (Page, this);
@@ -537,7 +547,7 @@ namespace System.Web.UI
 				}
 
 				StringBuilder sb = new StringBuilder ();
-				sb.AppendLine ("if (typeof(Sys) === 'undefined') throw new Error('ASP.NET Ajax client-side framework failed to load.');");
+   				sb.AppendLine ("if (typeof(Sys) === 'undefined') throw new Error('ASP.NET Ajax client-side framework failed to load.');");
 
 				ScriptingProfileServiceSection profileService = (ScriptingProfileServiceSection) WebConfigurationManager.GetSection ("system.web.extensions/scripting/webServices/profileService");
 				if (profileService != null && profileService.Enabled)
@@ -563,12 +573,6 @@ namespace System.Web.UI
 				if (IsMultiForm)
 					RegisterScriptReference (this, ajaxWebFormsExtensionScript, true);
 			}
-
-			if (!String.IsNullOrEmpty (_panelToRefreshID)) {
-				UpdatePanel panel = FindPanelWithId (_panelToRefreshID);
-				if (panel != null)
-					RegisterPanelForRefresh (panel);
-			}
 			
 			// Register Scripts
 			if (_scriptToRegister != null)
@@ -690,14 +694,23 @@ namespace System.Web.UI
 		}
 #endif
 
-		static bool HasBeenRendered (Control control) {
+		bool PanelRequiresUpdate (UpdatePanel panel)
+		{
+			if (panel == null || _panelsToRefresh == null || _panelsToRefresh.Count == 0)
+				return false;
+
+			return _panelsToRefresh.Contains (panel);
+		}
+		
+		bool HasBeenRendered (Control control)
+		{
 			if (control == null)
 				return false;
 
 			UpdatePanel panel = control as UpdatePanel;
-			if (panel != null && panel.RequiresUpdate)
-					return true;
-			
+			if (PanelRequiresUpdate (panel))
+				return true;
+
 			return HasBeenRendered (control.Parent);
 		}
 
@@ -740,10 +753,9 @@ namespace System.Web.UI
 				ResolveScriptReference (this, e);
 		}
 
-		protected virtual void RaisePostDataChangedEvent () {
-			UpdatePanel up = Page.FindControl (_panelToRefreshID) as UpdatePanel;
-			if (up != null && up.ChildrenAsTriggers)
-				up.Update ();
+		protected virtual void RaisePostDataChangedEvent ()
+		{
+			// Why override?
 		}
 
 		public static void RegisterArrayDeclaration (Page page, string arrayName, string arrayValue) {
@@ -752,8 +764,11 @@ namespace System.Web.UI
 
 		public static void RegisterArrayDeclaration (Control control, string arrayName, string arrayValue) {
 			Page page = control.Page;
-			ScriptManager sm = GetCurrent (page);
+			ScriptManager sm = GetCurrentInternal (page);
 
+			if (sm == null)
+				return;
+			
 			if (sm._arrayDeclarations == null)
 				sm._arrayDeclarations = new List<RegisteredArrayDeclaration> ();
 
@@ -782,8 +797,11 @@ namespace System.Web.UI
 
 		public static void RegisterClientScriptBlock (Control control, Type type, string key, string script, bool addScriptTags) {
 			Page page = control.Page;
-			ScriptManager sm = GetCurrent (page);
+			ScriptManager sm = GetCurrentInternal (page);
 
+			if (sm == null)
+				return;
+			
 			RegisterScript (ref sm._clientScriptBlocks, control, type, key, script, null, addScriptTags, RegisteredScriptType.ClientScriptBlock);
 
 			if (!sm.IsInAsyncPostBack)
@@ -796,8 +814,11 @@ namespace System.Web.UI
 
 		public static void RegisterClientScriptInclude (Control control, Type type, string key, string url) {
 			Page page = control.Page;
-			ScriptManager sm = GetCurrent (page);
+			ScriptManager sm = GetCurrentInternal (page);
 
+			if (sm == null)
+				return;
+			
 			RegisterScript (ref sm._clientScriptBlocks, control, type, key, null, url, false, RegisteredScriptType.ClientScriptInclude);
 
 			if (!sm.IsInAsyncPostBack)
@@ -812,56 +833,19 @@ namespace System.Web.UI
 			RegisterClientScriptInclude (control, type, resourceName, ScriptResourceHandler.GetResourceUrl (type.Assembly, resourceName, true));
 		}
 
-		void RegisterScriptReference (Control control, ScriptReference script, bool loadScriptsBeforeUI) {
-
-			bool isDebugMode = IsDeploymentRetail ? false : (script.ScriptModeInternal == ScriptMode.Inherit ? IsDebuggingEnabled : (script.ScriptModeInternal == ScriptMode.Debug));
-			string url;
-			if (!String.IsNullOrEmpty (script.Path)) {
-				url = GetScriptName (control.ResolveClientUrl (script.Path), isDebugMode, EnableScriptLocalization ? script.ResourceUICultures : null);
-			}
-			else if (!String.IsNullOrEmpty (script.Name)) {
-				Assembly assembly;
-				if (String.IsNullOrEmpty (script.Assembly))
-					assembly = typeof (ScriptManager).Assembly;
-				else
-					assembly = Assembly.Load (script.Assembly);
-				string name = GetScriptName (script.Name, isDebugMode, null);
-				if (script.IgnoreScriptPath || String.IsNullOrEmpty (ScriptPath))
-					url = ScriptResourceHandler.GetResourceUrl (assembly, name, script.NotifyScriptLoaded);
-				else {
-					AssemblyName an = assembly.GetName ();
-					url = ResolveClientUrl (String.Concat (VirtualPathUtility.AppendTrailingSlash (ScriptPath), an.Name, '/', an.Version, '/', name));
-				}
-			}
-			else {
-				throw new InvalidOperationException ("Name and Path cannot both be empty.");
-			}
-
+		void RegisterScriptReference (Control control, ScriptReference script, bool loadScriptsBeforeUI)
+		{
+			string scriptPath = script.Path;
+			string url = script.GetUrl (this, false);
+			if (control != this && !String.IsNullOrEmpty (scriptPath))
+				url = control.ResolveClientUrl (url);
+			
 			if (loadScriptsBeforeUI)
 				RegisterClientScriptInclude (control, typeof (ScriptManager), url, url);
 			else
 				RegisterStartupScript (control, typeof (ScriptManager), url, String.Format ("<script src=\"{0}\" type=\"text/javascript\"></script>", url), false);
 		}
 
-		static string GetScriptName (string releaseName, bool isDebugMode, string [] supportedUICultures) {
-			if (!isDebugMode && (supportedUICultures == null || supportedUICultures.Length == 0))
-				return releaseName;
-
-			if (releaseName.Length < 3 || !releaseName.EndsWith (".js", StringComparison.OrdinalIgnoreCase))
-				throw new InvalidOperationException (String.Format ("'{0}' is not a valid script path.  The path must end in '.js'.", releaseName));
-
-			StringBuilder sb = new StringBuilder (releaseName);
-			sb.Length -= 3;
-			if (isDebugMode)
-				sb.Append (".debug");
-			string culture = Thread.CurrentThread.CurrentUICulture.Name;
-			if (supportedUICultures != null && Array.IndexOf<string> (supportedUICultures, culture) >= 0)
-				sb.AppendFormat (".{0}", culture);
-			sb.Append (".js");
-
-			return sb.ToString ();
-		}
-
 		void RegisterServiceReference (Control control, ServiceReference serviceReference)
 		{
 			if (serviceReference.InlineScript) {
@@ -936,8 +920,11 @@ namespace System.Web.UI
 
 		public static void RegisterExpandoAttribute (Control control, string controlId, string attributeName, string attributeValue, bool encode) {
 			Page page = control.Page;
-			ScriptManager sm = GetCurrent (page);
+			ScriptManager sm = GetCurrentInternal (page);
 
+			if (sm == null)
+				return;
+			
 			if (sm._expandoAttributes == null)
 				sm._expandoAttributes = new List<RegisteredExpandoAttribute> ();
 
@@ -953,8 +940,11 @@ namespace System.Web.UI
 
 		public static void RegisterHiddenField (Control control, string hiddenFieldName, string hiddenFieldInitialValue) {
 			Page page = control.Page;
-			ScriptManager sm = GetCurrent (page);
+			ScriptManager sm = GetCurrentInternal (page);
 
+			if (sm == null)
+				return;
+			
 			if (sm._hiddenFields == null)
 				sm._hiddenFields = new List<RegisteredHiddenField> ();
 
@@ -970,8 +960,11 @@ namespace System.Web.UI
 
 		public static void RegisterOnSubmitStatement (Control control, Type type, string key, string script) {
 			Page page = control.Page;
-			ScriptManager sm = GetCurrent (page);
+			ScriptManager sm = GetCurrentInternal (page);
 
+			if (sm == null)
+				return;
+			
 			RegisterScript (ref sm._onSubmitStatements, control, type, key, script, null, false, RegisteredScriptType.OnSubmitStatement);
 
 			if (!sm.IsInAsyncPostBack)
@@ -1022,7 +1015,8 @@ namespace System.Web.UI
 			RegisterScriptDescriptors ((Control) scriptControl, scriptControl.GetScriptDescriptors ());
 		}
 
-		void RegisterScriptDescriptors (Control control, IEnumerable<ScriptDescriptor> scriptDescriptors) {
+		void RegisterScriptDescriptors (Control control, IEnumerable<ScriptDescriptor> scriptDescriptors)
+		{
 			if (scriptDescriptors == null)
 				return;
 
@@ -1034,6 +1028,7 @@ namespace System.Web.UI
 				}
 				else
 					sb.AppendLine ("Sys.Application.add_init(function() {");
+				sb.Append ("\t");
 				sb.AppendLine (scriptDescriptor.GetScript ());
 				sb.AppendLine ("});");
 			}
@@ -1047,8 +1042,11 @@ namespace System.Web.UI
 
 		public static void RegisterStartupScript (Control control, Type type, string key, string script, bool addScriptTags) {
 			Page page = control.Page;
-			ScriptManager sm = GetCurrent (page);
+			ScriptManager sm = GetCurrentInternal (page);
 
+			if (sm == null)
+				return;
+			
 			RegisterScript (ref sm._startupScriptBlocks, control, type, key, script, null, addScriptTags, RegisteredScriptType.ClientStartupScript);
 
 			if (!sm.IsInAsyncPostBack)
@@ -1096,9 +1094,18 @@ namespace System.Web.UI
 				writer.WriteLine ("//<![CDATA[");
 				writer.WriteLine ("Sys.WebForms.PageRequestManager._initialize('{0}', document.getElementById('{1}'));", UniqueID, Page.Form.ClientID);
 				if (IsMultiForm)
-					writer.WriteLine ("Sys.WebForms.PageRequestManager.getInstance($get(\"{0}\"))._updateControls([{1}], [{2}], [{3}], {4});", Page.Form.ClientID, FormatUpdatePanelIDs (_updatePanels, true), FormatListIDs (_asyncPostBackControls, true), FormatListIDs (_postBackControls, true), AsyncPostBackTimeout);
+					writer.WriteLine ("Sys.WebForms.PageRequestManager.getInstance($get(\"{0}\"))._updateControls([{1}], [{2}], [{3}], {4});",
+							  Page.Form.ClientID,
+							  FormatUpdatePanelIDs (_updatePanels, true),
+							  FormatListIDs (_asyncPostBackControls, true, false),
+							  FormatListIDs (_postBackControls, true, false),
+							  AsyncPostBackTimeout);
 				else
-					writer.WriteLine ("Sys.WebForms.PageRequestManager.getInstance()._updateControls([{0}], [{1}], [{2}], {3});", FormatUpdatePanelIDs (_updatePanels, true), FormatListIDs (_asyncPostBackControls, true), FormatListIDs (_postBackControls, true), AsyncPostBackTimeout);
+					writer.WriteLine ("Sys.WebForms.PageRequestManager.getInstance()._updateControls([{0}], [{1}], [{2}], {3});",
+							  FormatUpdatePanelIDs (_updatePanels, true),
+							  FormatListIDs (_asyncPostBackControls, true, false),
+							  FormatListIDs (_postBackControls, true, false),
+							  AsyncPostBackTimeout);
 				writer.WriteLine ("//]]");
 				writer.WriteLine ("</script>");
 			}
@@ -1110,21 +1117,29 @@ namespace System.Web.UI
 				return null;
 
 			StringBuilder sb = new StringBuilder ();
-			for (int i = 0; i < list.Count; i++) {
-				sb.AppendFormat ("{0}{1}{2}{0},", useSingleQuote ? "'" : String.Empty, list [i].ChildrenAsTriggers ? "t" : "f", list [i].UniqueID);
+			foreach (UpdatePanel panel in list) {
+				if (!panel.Visible)
+					continue;
+				
+				sb.AppendFormat ("{0}{1}{2}{0},", useSingleQuote ? "'" : String.Empty, panel.ChildrenAsTriggers ? "t" : "f", panel.UniqueID);
 			}
+			
 			if (sb.Length > 0)
 				sb.Length--;
 
 			return sb.ToString ();
 		}
 
-		static string FormatListIDs<T> (List<T> list, bool useSingleQuote) where T : Control {
+		static string FormatListIDs<T> (List<T> list, bool useSingleQuote, bool skipInvisible) where T : Control
+		{
 			if (list == null || list.Count == 0)
 				return null;
 
 			StringBuilder sb = new StringBuilder ();
 			for (int i = 0; i < list.Count; i++) {
+				if (skipInvisible && !list [i].Visible)
+					continue;
+				
 				sb.AppendFormat ("{0}{1}{0},", useSingleQuote ? "'" : String.Empty, list [i].UniqueID);
 			}
 			if (sb.Length > 0)
@@ -1205,12 +1220,11 @@ namespace System.Web.UI
 			else if (_panelsToRefresh.Contains (panel))
 				return;
 
-			panel.Update ();
 			_panelsToRefresh.Add (panel);
 		}
 		
-		internal void WriteCallbackPanel (TextWriter output, UpdatePanel panel, StringBuilder panelOutput) {
-			RegisterPanelForRefresh (panel);
+		internal void WriteCallbackPanel (TextWriter output, UpdatePanel panel, StringBuilder panelOutput)
+		{
 			WriteCallbackOutput (output, updatePanel, panel.ClientID, panelOutput);
 		}
 
@@ -1222,7 +1236,7 @@ namespace System.Web.UI
 
 		static void WriteCallbackOutput (TextWriter output, string type, string name, object value) {
 			string str = value as string;
-			StringBuilder sb = value as StringBuilder;
+			StringBuilder sb = str == null ? value as StringBuilder : null;
 			int length = 0;
 			if (str != null)
 				length = str.Length;
@@ -1244,9 +1258,59 @@ namespace System.Web.UI
 			output.Write ('|');
 		}
 
-		void RenderPageCallback (HtmlTextWriter output, Control container) {
+		void RenderPageCallback (HtmlTextWriter output, Control container)
+		{
 			Page page = (Page) container;
 
+			// MSDN: http://msdn.microsoft.com/en-us/library/system.web.ui.updatepanel.aspx
+			//
+			// Refreshing UpdatePanel Content
+			//
+			// When partial-page rendering is enabled, a control can perform a postback
+			// that updates the whole page or an asynchronous postback that updates the
+			// content of one or more UpdatePanel controls. Whether a control causes an
+			// asynchronous postback and updates an UpdatePanel control depends on the
+			// following:
+			//
+			// * If the UpdateMode property of the UpdatePanel control is set to Always,
+			//   the UpdatePanel control's content is updated on every postback that
+			//   originates from the page. This includes asynchronous postbacks from
+			//   controls that are inside other UpdatePanel controls and postbacks from
+			//   controls that are not inside UpdatePanel controls.
+			//
+			// * If the UpdateMode property is set to Conditional, the UpdatePanel
+			//   control's content is updated in the following circumstances:
+			//
+			//       o When you call the Update method of the UpdatePanel control
+			//         explicitly.
+			//       o When the UpdatePanel control is nested inside another UpdatePanel
+			//         control, and the parent panel is updated.
+			//       o When a postback is caused by a control that is defined as a
+			//         trigger by using the Triggers property of the UpdatePanel
+			//         control. In this scenario, the control explicitly triggers an
+			//         update of the panel content. The control can be either inside or
+			//         outside the UpdatePanel control that the trigger is associated
+			//         with.
+			//       o When the ChildrenAsTriggers property is set to true and a child
+			//         control of the UpdatePanel control causes a postback. Child
+			//         controls of nested UpdatePanel controls do not cause an update to
+			//         the outer UpdatePanel control unless they are explicitly defined
+			//         as triggers.
+			//
+			if (_updatePanels != null && _updatePanels.Count > 0) {
+				bool needsUpdate;
+				foreach (UpdatePanel panel in _updatePanels) {
+					if (panel.RequiresUpdate || (!String.IsNullOrEmpty (_panelToRefreshID) && String.Compare (_panelToRefreshID, panel.UniqueID, StringComparison.Ordinal) == 0))
+						needsUpdate = true;
+					else
+						needsUpdate = false;
+
+					panel.SetInPartialRendering (needsUpdate);
+					if (needsUpdate)
+						RegisterPanelForRefresh (panel);
+				}
+			}
+			
 			page.Form.SetRenderMethodDelegate (RenderFormCallback);
 			HtmlTextParser parser = new HtmlTextParser (output);
 			page.Form.RenderControl (parser);
@@ -1256,11 +1320,11 @@ namespace System.Web.UI
 				foreach (KeyValuePair <string, string> kvp in pageHiddenFields)
 					WriteCallbackOutput (output, hiddenField, kvp.Key, kvp.Value);
 			
-			WriteCallbackOutput (output, asyncPostBackControlIDs, null, FormatListIDs (_asyncPostBackControls, false));
-			WriteCallbackOutput (output, postBackControlIDs, null, FormatListIDs (_postBackControls, false));
+			WriteCallbackOutput (output, asyncPostBackControlIDs, null, FormatListIDs (_asyncPostBackControls, false, false));
+			WriteCallbackOutput (output, postBackControlIDs, null, FormatListIDs (_postBackControls, false, false));
 			WriteCallbackOutput (output, updatePanelIDs, null, FormatUpdatePanelIDs (_updatePanels, false));
-			WriteCallbackOutput (output, childUpdatePanelIDs, null, FormatListIDs (_childUpdatePanels, false));
-			WriteCallbackOutput (output, panelsToRefreshIDs, null, FormatListIDs (_panelsToRefresh, false));
+			WriteCallbackOutput (output, childUpdatePanelIDs, null, FormatListIDs (_childUpdatePanels, false, true));
+			WriteCallbackOutput (output, panelsToRefreshIDs, null, FormatListIDs (_panelsToRefresh, false, true));
 			WriteCallbackOutput (output, asyncPostBackTimeout, null, AsyncPostBackTimeout.ToString ());
 			if (!IsMultiForm)
 				WriteCallbackOutput (output, pageTitle, null, Page.Title);
@@ -1323,6 +1387,7 @@ namespace System.Web.UI
 				return;
 			var registeredScripts = new Dictionary <string, RegisteredScript> ();
 			Control control;
+			Page page = Page;
 			
 			for (int i = 0; i < scriptList.Count; i++) {
 				RegisteredScript scriptEntry = scriptList [i];
@@ -1330,27 +1395,27 @@ namespace System.Web.UI
 					continue;
 
  				control = scriptEntry.Control;
-				if (Page == control || HasBeenRendered (control)) {
+				if (page == control || HasBeenRendered (control)) {
 					registeredScripts.Add (scriptEntry.Key, scriptEntry);
 					switch (scriptEntry.ScriptType) {
-					case RegisteredScriptType.ClientScriptBlock:
-						if (scriptEntry.AddScriptTags)
-							WriteCallbackOutput (output, scriptBlock, scriptContentNoTags, scriptEntry.Script);
-						else
-							WriteCallbackOutput (output, scriptBlock, scriptContentWithTags, SerializeScriptBlock (scriptEntry));
-						break;
-					case RegisteredScriptType.ClientStartupScript:
-						if (scriptEntry.AddScriptTags)
-							WriteCallbackOutput (output, scriptStartupBlock, scriptContentNoTags, scriptEntry.Script);
-						else
-							WriteCallbackOutput (output, scriptStartupBlock, scriptContentWithTags, SerializeScriptBlock (scriptEntry));
-						break;
-					case RegisteredScriptType.ClientScriptInclude:
-						WriteCallbackOutput (output, scriptBlock, scriptPath, scriptEntry.Url);
-						break;
-					case RegisteredScriptType.OnSubmitStatement:
-						WriteCallbackOutput (output, onSubmit, null, scriptEntry.Script);
-						break;
+						case RegisteredScriptType.ClientScriptBlock:
+							if (scriptEntry.AddScriptTags)
+								WriteCallbackOutput (output, scriptBlock, scriptContentNoTags, scriptEntry.Script);
+							else
+								WriteCallbackOutput (output, scriptBlock, scriptContentWithTags, SerializeScriptBlock (scriptEntry));
+							break;
+						case RegisteredScriptType.ClientStartupScript:
+							if (scriptEntry.AddScriptTags)
+								WriteCallbackOutput (output, scriptStartupBlock, scriptContentNoTags, scriptEntry.Script);
+							else
+								WriteCallbackOutput (output, scriptStartupBlock, scriptContentWithTags, SerializeScriptBlock (scriptEntry));
+							break;
+						case RegisteredScriptType.ClientScriptInclude:
+							WriteCallbackOutput (output, scriptBlock, scriptPath, scriptEntry.Url);
+							break;
+						case RegisteredScriptType.OnSubmitStatement:
+							WriteCallbackOutput (output, onSubmit, null, scriptEntry.Script);
+							break;
 					}
 				}
 			}
@@ -1397,14 +1462,15 @@ namespace System.Web.UI
 			throw new InvalidOperationException (String.Format ("The script tag registered for type '{0}' and key '{1}' has invalid characters outside of the script tags: {2}. Only properly formatted script tags can be registered.", scriptEntry.Type, scriptEntry.Key, scriptEntry.Script));
 		}
 
-		void RenderFormCallback (HtmlTextWriter output, Control container) {
+		void RenderFormCallback (HtmlTextWriter output, Control container)
+		{
 			output = ((HtmlTextParser) output).ResponseOutput;
 			HtmlForm form = (HtmlForm) container;
 			HtmlTextWriter writer = new HtmlDropWriter (output);
+			
 			if (form.HasControls ()) {
-				for (int i = 0; i < form.Controls.Count; i++) {
-					form.Controls [i].RenderControl (writer);
-				}
+				foreach (Control control in form.Controls)
+					control.RenderControl (writer);
 			}
 		}
 
