@@ -34,10 +34,10 @@ using System.Reflection;
 using System.IO;
 using System.Resources;
 using System.Collections;
-using System.Security.Cryptography;
 using System.Text;
 using System.Text.RegularExpressions;
 using System.Web.Configuration;
+using System.Web.Util;
 
 namespace System.Web.Handlers {
 #if SYSTEM_WEB_EXTENSIONS
@@ -58,11 +58,6 @@ namespace System.Web.Handlers {
 		const char QueryParamSeparator = '&';
 
 		static readonly Hashtable _embeddedResources = Hashtable.Synchronized (new Hashtable ());
-#if SYSTEM_WEB_EXTENSIONS
-		static ScriptResourceHandler () {
-			MachineKeySectionUtils.AutoGenKeys ();
-		}
-#endif
 
 		static void InitEmbeddedResourcesUrls (Assembly assembly, Hashtable hashtable)
 		{
@@ -93,67 +88,25 @@ namespace System.Web.Handlers {
 		}
 #endif
 
-		static string GetHexString (byte [] bytes)
-		{
-			const int letterPart = 55;
-			const int numberPart = 48;
-			char [] result = new char [bytes.Length * 2];
-			for (int i = 0; i < bytes.Length; i++) {
-				int tmp = (int) bytes [i];
-				int second = tmp & 15;
-				int first = (tmp >> 4) & 15;
-				result [(i * 2)] = (char) (first > 9 ? letterPart + first : numberPart + first);
-				result [(i * 2) + 1] = (char) (second > 9 ? letterPart + second : numberPart + second);
-			}
-			return new string (result);
-		}
-		
-		static byte[] GetEncryptionKey ()
-		{
-#if NET_2_0
-			return MachineKeySectionUtils.DecryptionKey192Bits ();
-#else
-			MachineKeyConfig config = HttpContext.GetAppConfig ("system.web/machineKey") as MachineKeyConfig;
-			return config.DecryptionKey192Bits;
-#endif
-		}
-
-		static byte[] GetBytes (string val)
-		{
-#if NET_2_0
-			return MachineKeySectionUtils.GetBytes (val, val.Length);
-#else
-			return MachineKeyConfig.GetBytes (val, val.Length);
-#endif
-		}		
-		
-		static byte [] init_vector = { 0xD, 0xE, 0xA, 0xD, 0xB, 0xE, 0xE, 0xF };
-		
 		static string EncryptAssemblyResource (string asmName, string resName)
 		{
-			byte[] key = GetEncryptionKey ();
 			byte[] bytes = Encoding.UTF8.GetBytes (String.Concat (asmName, ";", resName));
-			string result;
-			
-			ICryptoTransform encryptor = TripleDES.Create ().CreateEncryptor (key, init_vector);
-			result = GetHexString (encryptor.TransformFinalBlock (bytes, 0, bytes.Length));
-			bytes = null;
-
-			return String.Concat ("d=", result.ToLower (CultureInfo.InvariantCulture));
+			bytes = MachineKeySectionUtils.Encrypt (MachineKeySection.Config, bytes);
+			return Convert.ToBase64String (bytes);
 		}
 
 		static void DecryptAssemblyResource (string val, out string asmName, out string resName)
 		{
-			byte[] key = GetEncryptionKey ();
-			byte[] bytes = GetBytes (val);
-			byte[] result;
+			byte[] bytes = Convert.FromBase64String (val);
 
 			asmName = null;
 			resName = null;			
 
-			ICryptoTransform decryptor = TripleDES.Create ().CreateDecryptor (key, init_vector);
-			result = decryptor.TransformFinalBlock (bytes, 0, bytes.Length);
+			byte[] result = MachineKeySectionUtils.Decrypt (MachineKeySection.Config, bytes);
 			bytes = null;
+			// null will be returned if, for any reason, decryption fails
+			if (result == null)
+				return;
 
 			string data = Encoding.UTF8.GetString (result);
 			result = null;
@@ -201,7 +154,7 @@ namespace System.Web.Handlers {
 			if (apath != String.Empty)
 				atime = String.Concat (QueryParamSeparator, "t=", File.GetLastWriteTimeUtc (apath).Ticks);
 #endif
-			string href = HandlerFileName + "?" + EncryptAssemblyResource (aname, resourceName) + atime + extra;
+			string href = HandlerFileName + "?d=" + EncryptAssemblyResource (aname, resourceName) + atime + extra;
 
 			HttpContext ctx = HttpContext.Current;
 			if (ctx != null && ctx.Request != null) {
@@ -212,7 +165,6 @@ namespace System.Web.Handlers {
 			return href;
 		}
 
-	
 #if SYSTEM_WEB_EXTENSIONS
 		protected virtual void ProcessRequest (HttpContext context)
 #else
@@ -220,19 +172,27 @@ namespace System.Web.Handlers {
 		void System.Web.IHttpHandler.ProcessRequest (HttpContext context)
 #endif
 		{
+			HttpRequest request = context.Request;
+			HttpResponse response = context.Response;
 			string resourceName;
 			string asmName;
 			Assembly assembly;
 
-			DecryptAssemblyResource (context.Request.QueryString ["d"], out asmName, out resourceName);
+			// val is URL-decoded, which means every + has been replaced with ' ', we
+			// need to revert that or the base64 conversion will fail.
+			string d = request.QueryString ["d"];
+			if (!String.IsNullOrEmpty (d))
+				d = d.Replace (' ', '+');
+
+			DecryptAssemblyResource (d, out asmName, out resourceName);
 			if (resourceName == null)
 				throw new HttpException (404, "No resource name given");
-			
+
 			if (asmName == null || asmName == "s")
 				assembly = currAsm;
 			else
 				assembly = Assembly.Load (asmName);
-			
+
 			WebResourceAttribute wra = null;
 			WebResourceAttribute [] attrs = (WebResourceAttribute []) assembly.GetCustomAttributes (typeof (WebResourceAttribute), false);
 			for (int i = 0; i < attrs.Length; i++) {
@@ -255,11 +215,58 @@ namespace System.Web.Handlers {
 			if (wra == null)
 				throw new HttpException (404, String.Concat ("Resource ", resourceName, " not found"));
 			
-			context.Response.ContentType = wra.ContentType;
+			string req_cache = request.Headers ["Cache-Control"];
+			if (req_cache == "max-age=0") {
+				long atime;
+#if NET_2_0
+				if (Int64.TryParse (request.QueryString ["t"], out atime)) {
+#else
+				atime = -1;
+				try {
+					atime = Int64.Parse (request.QueryString ["t"]);
+				} catch {}
+				if (atime > -1) {
+#endif
+					if (atime == File.GetLastWriteTimeUtc (assembly.Location).Ticks) {
+						response.Clear ();
+						response.StatusCode = 304;
+						response.ContentType = null;
+						response.CacheControl = "public"; // easier to set it to public as MS than remove it
+						context.ApplicationInstance.CompleteRequest ();
+						return;
+					}
+				}
+			}
+			string modif_since = request.Headers ["If-Modified-Since"];
+			if (modif_since != null && modif_since != "") {
+				try {
+					DateTime modif;
+#if NET_2_0
+					if (DateTime.TryParseExact (modif_since, "r", null, 0, out modif))
+#else
+					modif = DateTime.MinValue;
+					try {
+						modif = DateTime.ParseExact (modif_since, "r", null, 0);
+					} catch { }
+					if (modif != DateTime.MinValue)
+#endif
+						if (File.GetLastWriteTimeUtc (assembly.Location) <= modif) {
+							response.Clear ();
+							response.StatusCode = 304;
+							response.ContentType = null;
+							response.CacheControl = "public"; // easier to set it to public as MS than remove it
+							context.ApplicationInstance.CompleteRequest ();
+							return;
+						}
+				} catch {}
+			}
 
-			/* tell the client they can cache resources for 1 year */
-			context.Response.ExpiresAbsolute = DateTime.Now.AddYears (1);
-			context.Response.CacheControl = "private";
+			response.ContentType = wra.ContentType;
+
+			DateTime utcnow = DateTime.UtcNow;
+			response.Headers.Add ("Last-Modified", utcnow.ToString ("r"));
+			response.ExpiresAbsolute = utcnow.AddYears (1);
+			response.CacheControl = "public";
 
 			Stream s = assembly.GetManifestResourceStream (resourceName);
 			if (s == null)
@@ -267,13 +274,20 @@ namespace System.Web.Handlers {
 
 			if (wra.PerformSubstitution) {
 				using (StreamReader r = new StreamReader (s)) {
-					TextWriter w = context.Response.Output;
+					TextWriter w = response.Output;
 					new PerformSubstitutionHelper (assembly).PerformSubstitution (r, w);
 				}
-			}
-			else {
+#if NET_2_0
+			} else if (response.OutputStream is HttpResponseStream) {
+				UnmanagedMemoryStream st = (UnmanagedMemoryStream) s;
+				HttpResponseStream hstream = (HttpResponseStream) response.OutputStream;
+				unsafe {
+					hstream.WritePtr (new IntPtr (st.PositionPointer), (int) st.Length);
+				}
+#endif
+			} else {
 				byte [] buf = new byte [1024];
-				Stream output = context.Response.OutputStream;
+				Stream output = response.OutputStream;
 				int c;
 				do {
 					c = s.Read (buf, 0, 1024);
@@ -281,7 +295,7 @@ namespace System.Web.Handlers {
 				} while (c > 0);
 			}
 #if SYSTEM_WEB_EXTENSIONS
-			TextWriter writer = context.Response.Output;
+			TextWriter writer = response.Output;
 			foreach (ScriptResourceAttribute sra in assembly.GetCustomAttributes (typeof (ScriptResourceAttribute), false)) {
 				if (sra.ScriptName == resourceName) {
 					string scriptResourceName = sra.ScriptResourceName;
@@ -329,7 +343,7 @@ namespace System.Web.Handlers {
 				}
 			}
 
-			bool notifyScriptLoaded = context.Request.QueryString ["n"] == "t";
+			bool notifyScriptLoaded = request.QueryString ["n"] == "t";
 			if (notifyScriptLoaded) {
 				writer.WriteLine ();
 				writer.WriteLine ("if(typeof(Sys)!=='undefined')Sys.Application.notifyScriptLoaded();");
