@@ -5,10 +5,11 @@
 // Author:
 //	Miguel de Icaza (miguel@novell.com)
 //	Gonzalo Paniagua Javier (gonzalo@novell.com)
+//      Marek Habersack <mhabersack@novell.com>
 //
 
 //
-// Copyright (C) 2005 Novell, Inc (http://www.novell.com)
+// Copyright (C) 2005-2010 Novell, Inc (http://www.novell.com)
 //
 // Permission is hereby granted, free of charge, to any person obtaining
 // a copy of this software and associated documentation files (the
@@ -38,15 +39,22 @@ using System.Security;
 using System.Security.Permissions;
 using System.Security.Principal;
 using System.Web.Configuration;
+using System.Web.Management;
 using System.Web.UI;
 using System.Web.Util;
 using System.Globalization;
 
-namespace System.Web {
-	
+#if NET_4_0
+using System.Security.Authentication.ExtendedProtection;
+using System.Web.Routing;
+#endif
+
+namespace System.Web
+{	
 	// CAS - no InheritanceDemand here as the class is sealed
 	[AspNetHostingPermission (SecurityAction.LinkDemand, Level = AspNetHostingPermissionLevel.Minimal)]
-	public sealed partial class HttpRequest {
+	public sealed partial class HttpRequest
+	{
 		HttpWorkerRequest worker_request;
 		HttpContext context;
 		WebROCollection query_string_nvc;
@@ -62,14 +70,16 @@ namespace System.Web {
 		// On-demand computed values
 		//
 		HttpBrowserCapabilities browser_capabilities;
-		string file_path, base_virtual_dir, root_virtual_dir;
+		string file_path, base_virtual_dir, root_virtual_dir, client_file_path;
 		string content_type;
 		int content_length = -1;
 		Encoding encoding;
 		string current_exe_path;
 		string physical_path;
 		string unescaped_path;
+		string original_path;
 		string path_info;
+		string raw_url;
 		WebROCollection all_params;
 		WebROCollection headers;
 		Stream input_stream;
@@ -95,11 +105,68 @@ namespace System.Web {
 		// Validations
 		bool validate_cookies, validate_query_string, validate_form;
 		bool checked_cookies, checked_query_string, checked_form;
-
+		static readonly UrlMappingCollection urlMappings;
 		readonly static char [] queryTrimChars = {'?'};
+#if NET_4_0
+		RequestContext requestContext;
 		
+		static bool validateRequestNewMode;
+		internal static bool ValidateRequestNewMode {
+			get { return validateRequestNewMode; }
+		}
+
+		private static char[] RequestPathInvalidCharacters {
+			get; set;
+		}
+
+		private static char[] CharsFromList (string list)
+		{
+			// List format is very strict and enforced by the Configuration	
+			// there must be a single char separated by commas with no trailing comma
+			// whitespace is allowed though and should be trimmed.
+			
+			string [] pieces = list.Split (',');
+
+			char [] chars = new char [pieces.Length];
+			for (int i = 0; i < chars.Length; i++) {
+				string trimmed = pieces [i].Trim ();
+				if (trimmed.Length != 1) {
+					// This should have been caught by System.Web.Configuration
+					// and throw a configuration error. This is just here for sanity
+					throw new System.Configuration.ConfigurationErrorsException ();
+				}
+
+				chars [i] = trimmed [0];
+			}
+
+			return chars;
+		}
+#endif
+
 		static HttpRequest ()
 		{
+			try {
+				UrlMappingsSection ums = WebConfigurationManager.GetWebApplicationSection ("system.web/urlMappings") as UrlMappingsSection;
+				if (ums != null && ums.IsEnabled) {
+					urlMappings = ums.UrlMappings;
+					if (urlMappings.Count == 0)
+						urlMappings = null;
+				}
+
+#if NET_4_0
+				Version validationMode = HttpRuntime.Section.RequestValidationMode;
+
+				if (validationMode >= new Version (4, 0)) {
+					validateRequestNewMode = true;
+					string invalidChars = HttpRuntime.Section.RequestPathInvalidCharacters;
+					if (!String.IsNullOrEmpty (invalidChars))
+						RequestPathInvalidCharacters = CharsFromList (invalidChars);
+				}
+#endif
+			} catch {
+				// unlikely to happen
+			}
+			
 			host_addresses = GetLocalHostAddresses ();
 		}
 		
@@ -135,13 +202,7 @@ namespace System.Web {
 					else
 						query = worker_request.GetQueryString();
 					
-					BuildUrlComponents (
-#if NET_2_0
-						ApplyUrlMapping (worker_request.GetUriPath ()),
-#else
-						worker_request.GetUriPath (),
-#endif
-						query);
+					BuildUrlComponents (ApplyUrlMapping (worker_request.GetUriPath ()), query);
 				}
 				return url_components;
 			}
@@ -160,22 +221,15 @@ namespace System.Web {
 				url_components.Query = query.TrimStart (queryTrimChars);
 		}
 
-#if NET_2_0
 		internal string ApplyUrlMapping (string url)
 		{
-			if (WebConfigurationManager.HasConfigErrors)
-				return url;
-			
-			UrlMappingsSection ums = WebConfigurationManager.GetSection ("system.web/urlMappings", ApplicationPath) as UrlMappingsSection;
-			UrlMappingCollection umc;
-
-			if (ums == null || !ums.IsEnabled || (umc = ums.UrlMappings).Count == 0)
+			if (urlMappings == null)
 				return url;
 
 			string relUrl = VirtualPathUtility.ToAppRelative (url);
 			UrlMapping um = null;
 			
-			foreach (UrlMapping u in umc) {
+			foreach (UrlMapping u in urlMappings) {
 				if (u == null)
 					continue;
 				if (String.Compare (relUrl, u.Url, StringComparison.Ordinal) == 0) {
@@ -201,7 +255,6 @@ namespace System.Web {
 
 			return url_components.Path;
 		}
-#endif
 
 		string [] SplitHeader (int header_index)
 		{
@@ -227,7 +280,6 @@ namespace System.Web {
 			}
 		}
 
-#if NET_2_0
 #if !TARGET_JVM
 		public WindowsIdentity LogonUserIdentity {
 			get { throw new NotImplementedException (); }
@@ -243,7 +295,6 @@ namespace System.Web {
 				anonymous_id = value;
 			}
 		}
-#endif
 
 		public string ApplicationPath {
 			get {
@@ -267,7 +318,6 @@ namespace System.Web {
 			}
 		}
 
-#if NET_2_0		
 		internal bool BrowserMightHaveSpecialWriter {
 			get {
 				return (browser_capabilities != null 
@@ -281,7 +331,6 @@ namespace System.Web {
 					|| HttpApplicationFactory.AppBrowsersFiles.Length > 0);
 			}
 		}
-#endif
 
 		public HttpClientCertificate ClientCertificate {
 			get {
@@ -316,7 +365,7 @@ namespace System.Web {
 			get {
 				if (encoding == null){
 					if (worker_request == null)
-						throw new HttpException ("No HttpWorkerRequest");
+						throw HttpException.NewWithCode ("No HttpWorkerRequest", WebEventCodes.RuntimeErrorRequestAbort);
 					
 					string content_type = ContentType;
 					string parameter = GetParameter (content_type, "; charset=");
@@ -395,9 +444,15 @@ namespace System.Web {
 				// For J2EE portal support we emulate cookies using the session.
 				GetSessionCookiesForPortal (cookies);
 #endif
-				if (validate_cookies && !checked_cookies){
-					ValidateCookieCollection (cookies);
+				bool needValidation = validate_cookies;
+#if NET_4_0
+				needValidation |= validateRequestNewMode;
+#endif
+				if (needValidation && !checked_cookies) {
+					// Setting this before calling the validator prevents
+					// possible endless recursion
 					checked_cookies = true;
+					ValidateCookieCollection (cookies);
 				}
 
 				return cookies;
@@ -413,14 +468,16 @@ namespace System.Web {
 				return FilePath;
 			}
 		}
-
-#if NET_2_0
+#if NET_4_0
+		public string CurrentExecutionFilePathExtension {
+			get { return global::System.IO.Path.GetExtension (CurrentExecutionFilePath); }
+		}
+#endif
 		public string AppRelativeCurrentExecutionFilePath {
 			get {
 				return VirtualPathUtility.ToAppRelative (CurrentExecutionFilePath);
 			}
 		}
-#endif
 
 		public string FilePath {
 			get {
@@ -428,18 +485,32 @@ namespace System.Web {
 					return "/"; // required for 2.0
 
 				if (file_path == null)
-					file_path = UrlUtils.Canonic (
-#if NET_2_0
-						ApplyUrlMapping (worker_request.GetFilePath ())
-#else
-						worker_request.GetFilePath ()
-#endif
-					);
+					file_path = UrlUtils.Canonic (ApplyUrlMapping (worker_request.GetFilePath ()));
 
 				return file_path;
 			}
 		}
 
+		internal string ClientFilePath {
+			get {
+				if (client_file_path == null) {
+					if (worker_request == null)
+						return "/";
+					
+					return UrlUtils.Canonic (ApplyUrlMapping (worker_request.GetFilePath ()));
+				}
+				
+				return client_file_path;
+			}
+
+			set {
+				if (value == null || value.Length == 0)
+					client_file_path = null;
+				else
+					client_file_path = value;
+			}
+		}
+		
 		internal string BaseVirtualDir {
 			get {
 				if (base_virtual_dir == null){
@@ -613,7 +684,7 @@ namespace System.Web {
 			if (starts_with)
 				return StrUtils.StartsWith (ContentType, ct, true);
 
-			return String.Compare (ContentType, ct, true, CultureInfo.InvariantCulture) == 0;
+			return String.Compare (ContentType, ct, true, Helpers.InvariantCulture) == 0;
 		}
 		
 		public NameValueCollection Form {
@@ -631,10 +702,18 @@ namespace System.Web {
 					form.Protect ();
 				}
 
-				if (validate_form && !checked_form){
+#if NET_4_0
+				if (validateRequestNewMode && !checked_form) {
+					// Setting this before calling the validator prevents
+					// possible endless recursion
 					checked_form = true;
-					ValidateNameValueCollection ("Form", form);
-				}
+					ValidateNameValueCollection ("Form", query_string_nvc, RequestValidationSource.Form);
+				} else
+#endif
+					if (validate_form && !checked_form){
+						checked_form = true;
+						ValidateNameValueCollection ("Form", form);
+					}
 				
 				return form;
 			}
@@ -642,8 +721,22 @@ namespace System.Web {
 
 		public NameValueCollection Headers {
 			get {
-				if (headers == null)
+				if (headers == null) {
 					headers = new HeadersCollection (this);
+#if NET_4_0
+					if (validateRequestNewMode) {
+						RequestValidator validator = RequestValidator.Current;
+						int validationFailureIndex;
+
+						foreach (string hkey in headers.AllKeys) {
+							string value = headers [hkey];
+							
+							if (!validator.IsValidRequestString (HttpContext.Current, value, RequestValidationSource.Headers, hkey, out validationFailureIndex))
+								ThrowValidationException ("Headers", hkey, value);
+						}
+					}
+#endif
+				}
 				
 				return headers;
 			}
@@ -724,14 +817,10 @@ namespace System.Web {
 			// upload
 			//
 			int content_length = ContentLength;
-
-#if NET_2_0
-			HttpRuntimeSection config = (HttpRuntimeSection) WebConfigurationManager.GetWebApplicationSection ("system.web/httpRuntime");
-#else
-			HttpRuntimeConfig config = (HttpRuntimeConfig) HttpContext.GetAppConfig ("system.web/httpRuntime");
-#endif
-			if ((content_length / 1024) > config.MaxRequestLength)
-				throw new HttpException (400, "Upload size exceeds httpRuntime limit.");
+			int content_length_kb = content_length / 1024;
+			HttpRuntimeSection config = HttpRuntime.Section;
+			if (content_length_kb > config.MaxRequestLength)
+				throw HttpException.NewWithCode (400, "Upload size exceeds httpRuntime limit.", WebEventCodes.RuntimeErrorPostTooLarge);
 
 			int total = 0;
 			byte [] buffer;
@@ -750,7 +839,7 @@ namespace System.Web {
 			if (buffer != null)
 				total = buffer.Length;
 
-			if (content_length > 0 && (content_length / 1024) >= config.RequestLengthDiskThreshold) {
+			if (content_length > 0 && content_length_kb >= config.RequestLengthDiskThreshold) {
 				// Writes the request to disk
 				total = Math.Min (content_length, total);
 				request_file = GetTempStream ();
@@ -778,8 +867,9 @@ namespace System.Web {
 				total = Math.Min (content_length, total);
 				IntPtr content = Marshal.AllocHGlobal (content_length);
 				if (content == (IntPtr) 0)
-					throw new HttpException (String.Format ("Not enough memory to allocate {0} bytes.",
-									content_length));
+					throw HttpException.NewWithCode (
+						String.Format ("Not enough memory to allocate {0} bytes.", content_length),
+						WebEventCodes.WebErrorOtherError);
 
 				if (total > 0)
 					Marshal.Copy (buffer, 0, content, total);
@@ -815,7 +905,7 @@ namespace System.Web {
 						break;
 					total += n;
 					if (total < 0 || total > maxlength)
-						throw new HttpException (400, "Upload size exceeds httpRuntime limit.");
+						throw HttpException.NewWithCode (400, "Upload size exceeds httpRuntime limit.", WebEventCodes.RuntimeErrorPostTooLarge);
 
 					if (ms != null && total > disk_th) {
 						// Swith to on-disk file.
@@ -837,7 +927,7 @@ namespace System.Web {
 			DoFilter (buffer);
 
 			if (total < content_length)
-				throw new HttpException (411, "The request body is incomplete.");
+				throw HttpException.NewWithCode (411, "The request body is incomplete.", WebEventCodes.WebErrorOtherError);
 		}
 #endif
 
@@ -860,7 +950,24 @@ namespace System.Web {
 				} catch {}
 			}
 		}
-		
+#if NET_4_0
+		public RequestContext RequestContext {
+			get {
+				if (requestContext == null)
+					requestContext = new RequestContext (new HttpContextWrapper (this.context ?? HttpContext.Current), new RouteData ());
+
+				return requestContext;
+			}
+			
+			internal set { requestContext = value; }	
+		}
+
+		public ChannelBinding HttpChannelBinding {
+			get {
+				throw new PlatformNotSupportedException ("This property is not supported.");
+			}
+		}
+#endif
 		public Stream InputStream {
 			get {
 				if (input_stream == null)
@@ -916,26 +1023,33 @@ namespace System.Web {
 			}
 		}
 
+		internal string PathNoValidation {
+			get {
+				if (original_path == null) {
+					if (url_components != null)
+						// use only if it's already been instantiated, so that we can't go into endless
+						// recursion in some scenarios
+						original_path = UrlComponents.Path;
+					else
+						original_path = ApplyUrlMapping (worker_request.GetUriPath ());
+				}
+
+				return original_path;
+			}
+		}
+		
 		public string Path {
 			get {
 				if (unescaped_path == null) {
-					string path;
-					if (url_components != null) {
-						// use only if it's already been instantiated, so that we can't go into endless
-						// recursion in some scenarios
-						path = UrlComponents.Path;
-					} else {
-#if NET_2_0
-						path = ApplyUrlMapping (worker_request.GetUriPath ());
-#else
-						path = worker_request.GetUriPath ();
-#endif
-					}
+					unescaped_path = Uri.UnescapeDataString (PathNoValidation);
+#if NET_4_0
+					if (validateRequestNewMode) {
+						RequestValidator validator = RequestValidator.Current;
+						int validationFailureIndex;
 						
-#if NET_2_0
-					unescaped_path = Uri.UnescapeDataString (path);
-#else
-					unescaped_path = HttpUtility.UrlDecode (path);
+						if (!validator.IsValidRequestString (HttpContext.Current, unescaped_path, RequestValidationSource.Path, null, out validationFailureIndex))
+							ThrowValidationException ("Path", "Path", unescaped_path);
+					}
 #endif
 				}
 				
@@ -948,7 +1062,16 @@ namespace System.Web {
 				if (path_info == null) {
 					if (worker_request == null)
 						return String.Empty;
-					path_info = worker_request.GetPathInfo ();
+					path_info = worker_request.GetPathInfo () ?? String.Empty;
+#if NET_4_0
+					if (validateRequestNewMode) {
+						RequestValidator validator = RequestValidator.Current;
+						int validationFailureIndex;
+						
+						if (!validator.IsValidRequestString (HttpContext.Current, path_info, RequestValidationSource.PathInfo, null, out validationFailureIndex))
+							ThrowValidationException ("PathInfo", "PathInfo", path_info);
+					}
+#endif
 				}
 
 				return path_info;
@@ -1014,11 +1137,18 @@ namespace System.Web {
 					
 					query_string_nvc.Protect();
 				}
-				
-				if (validate_query_string && !checked_query_string) {
-					ValidateNameValueCollection ("QueryString", query_string_nvc);
+#if NET_4_0
+				if (validateRequestNewMode && !checked_query_string) {
+					// Setting this before calling the validator prevents
+					// possible endless recursion
 					checked_query_string = true;
-				}
+					ValidateNameValueCollection ("QueryString", query_string_nvc, RequestValidationSource.QueryString);
+				} else
+#endif
+					if (validate_query_string && !checked_query_string) {
+						ValidateNameValueCollection ("QueryString", query_string_nvc);
+						checked_query_string = true;
+					}
 				
 				return query_string_nvc;
 			}
@@ -1026,10 +1156,26 @@ namespace System.Web {
 
 		public string RawUrl {
 			get {
-				if (worker_request != null)
-					return worker_request.GetRawUrl ();
-				else
-					return UrlComponents.Path + UrlComponents.Query;
+				if (raw_url == null) {
+					if (worker_request != null)
+						raw_url = worker_request.GetRawUrl ();
+					else
+						raw_url = UrlComponents.Path + UrlComponents.Query;
+					
+					if (raw_url == null)
+						raw_url = String.Empty;
+#if NET_4_0
+					if (validateRequestNewMode) {
+						RequestValidator validator = RequestValidator.Current;
+						int validationFailureIndex;
+
+						if (!validator.IsValidRequestString (HttpContext.Current, raw_url, RequestValidationSource.RawUrl, null, out validationFailureIndex))
+							ThrowValidationException ("RawUrl", "RawUrl", raw_url);
+					}
+#endif
+				}
+				
+				return raw_url;
 			}
 		}
 
@@ -1196,7 +1342,7 @@ namespace System.Web {
 		public string MapPath (string virtualPath, string baseVirtualDir, bool allowCrossAppMapping)
 		{
 			if (worker_request == null)
-				throw new HttpException ("No HttpWorkerRequest");
+				throw HttpException.NewWithCode ("No HttpWorkerRequest", WebEventCodes.RuntimeErrorRequestAbort);
 
 			if (virtualPath == null)
 				virtualPath = "~";
@@ -1207,22 +1353,23 @@ namespace System.Web {
 			}
 
 			if (!VirtualPathUtility.IsValidVirtualPath (virtualPath))
-				throw new HttpException (String.Format ("'{0}' is not a valid virtual path.", virtualPath));
+				throw HttpException.NewWithCode (String.Format ("'{0}' is not a valid virtual path.", virtualPath), WebEventCodes.RuntimeErrorRequestAbort);
 
 			string appVirtualPath = HttpRuntime.AppDomainAppVirtualPath;
-
 			if (!VirtualPathUtility.IsRooted (virtualPath)) {
 				if (StrUtils.IsNullOrEmpty (baseVirtualDir))
 					baseVirtualDir = appVirtualPath;
 				virtualPath = VirtualPathUtility.Combine (VirtualPathUtility.AppendTrailingSlash (baseVirtualDir), virtualPath);
-			}
-			virtualPath = VirtualPathUtility.ToAbsolute (virtualPath);
-			
+				if (!VirtualPathUtility.IsAbsolute (virtualPath))
+					virtualPath = VirtualPathUtility.ToAbsolute (virtualPath);
+			} else if (!VirtualPathUtility.IsAbsolute (virtualPath))
+				virtualPath = VirtualPathUtility.ToAbsolute (virtualPath);
+
 			if (!allowCrossAppMapping){
 				if (!StrUtils.StartsWith (virtualPath, appVirtualPath, true))
-					throw new HttpException ("MapPath: Mapping across applications not allowed");
+					throw HttpException.NewWithCode ("MapPath: Mapping across applications not allowed", WebEventCodes.RuntimeErrorRequestAbort);
 				if (appVirtualPath.Length > 1 && virtualPath.Length > 1 && virtualPath [0] != '/')
-					throw new HttpException ("MapPath: Mapping across applications not allowed");
+					throw HttpException.NewWithCode ("MapPath: Mapping across applications not allowed", WebEventCodes.RuntimeErrorRequestAbort);
 			}
 #if TARGET_JVM
 			return worker_request.MapPath (virtualPath);
@@ -1286,7 +1433,32 @@ namespace System.Web {
 			validate_query_string = true;
 			validate_form = true;
 		}
-
+#if NET_4_0
+		internal void Validate ()
+		{
+			var cfg = HttpRuntime.Section;
+			string query = UrlComponents.Query;
+			
+			if (query != null && query.Length > cfg.MaxQueryStringLength)
+				throw new HttpException (400, "The length of the query string for this request exceeds the configured maxQueryStringLength value.");
+			
+			string path = PathNoValidation;
+			if (path != null) {
+				if (path.Length > cfg.MaxUrlLength)
+					throw new HttpException (400, "The length of the URL for this request exceeds the configured maxUrlLength value.");
+				
+				char[] invalidChars = RequestPathInvalidCharacters;
+				if (invalidChars != null) {
+					int idx = path.IndexOfAny (invalidChars);
+					if (idx != -1)
+						throw HttpException.NewWithCode (
+							String.Format ("A potentially dangerous Request.Path value was detected from the client ({0}).", path [idx]),
+							WebEventCodes.RuntimeErrorValidationFailure
+						);
+				}
+			}
+		}
+#endif
 #region internal routines
 		internal string ClientTarget {
 			get {
@@ -1297,13 +1469,8 @@ namespace System.Web {
 				client_target = value;
 			}
 		}
-
-#if NET_2_0
-		public
-#else
-		internal
-#endif
-		bool IsLocal {
+		
+		public bool IsLocal {
 			get {
 				string address = worker_request.GetRemoteAddress ();
 
@@ -1329,24 +1496,30 @@ namespace System.Web {
 		{
 			file_path = path;
 			physical_path = null;
+			original_path = null;
 		}
 
 		internal void SetCurrentExePath (string path)
 		{
 			cached_url = null;
 			current_exe_path = path;
-			UrlComponents.Path = path;
+			UrlComponents.Path = path + PathInfo;
 			// recreated on demand
 			root_virtual_dir = null;
 			base_virtual_dir = null;
 			physical_path = null;
 			unescaped_path = null;
+			original_path = null;
 		}
 
 		internal void SetPathInfo (string pi)
 		{
 			cached_url = null;
 			path_info = pi;
+			original_path = null;
+
+			string path = UrlComponents.Path;
+			UrlComponents.Path = path + PathInfo;
 		}
 
 		// Headers is ReadOnly, so we need this hack for cookie-less sessions.
@@ -1398,9 +1571,8 @@ namespace System.Web {
 		}
 
 		internal HttpContext Context {
-			get {
-				return context;
-			}
+			get { return context; }
+			set { context = value; }
 		}
 
 		static void ValidateNameValueCollection (string name, NameValueCollection coll)
@@ -1410,11 +1582,39 @@ namespace System.Web {
 		
 			foreach (string key in coll.Keys) {
 				string val = coll [key];
-				if (val != null && val.Length > 0 && CheckString (val))
+				if (val != null && val.Length > 0 && IsInvalidString (val))
 					ThrowValidationException (name, key, val);
 			}
 		}
-		
+#if NET_4_0
+		static void ValidateNameValueCollection (string name, NameValueCollection coll, RequestValidationSource source)
+		{
+			if (coll == null)
+				return;
+
+			RequestValidator validator = RequestValidator.Current;
+			int validationFailureIndex;
+			HttpContext context = HttpContext.Current;
+
+			foreach (string key in coll.Keys) {
+				string val = coll [key];
+				if (val != null && val.Length > 0 && !validator.IsValidRequestString (context, val, source, key, out validationFailureIndex))
+					ThrowValidationException (name, key, val);
+			}
+		}
+
+		[AspNetHostingPermission (SecurityAction.Demand, Level = AspNetHostingPermissionLevel.High)]
+		public void InsertEntityBody ()
+		{
+			throw new PlatformNotSupportedException ("This method is not supported.");
+		}
+
+		[AspNetHostingPermission (SecurityAction.Demand, Level = AspNetHostingPermissionLevel.High)]
+		public void InsertEntityBody (byte[] buffer, int offset, int count)
+		{
+			throw new PlatformNotSupportedException ("This method is not supported.");
+		}
+#endif
 		static void ValidateCookieCollection (HttpCookieCollection cookies)
 		{
 			if (cookies == null)
@@ -1422,15 +1622,35 @@ namespace System.Web {
 		
 			int size = cookies.Count;
 			HttpCookie cookie;
+#if NET_4_0
+			RequestValidator validator = RequestValidator.Current;
+			int validationFailureIndex;
+			HttpContext context = HttpContext.Current;
+#endif
+			bool invalid;
+			
 			for (int i = 0 ; i < size ; i++) {
 				cookie = cookies[i];
-				string value = cookie.Value;
+				if (cookie == null)
+					continue;
 				
-				if (value != null && value != "" && CheckString (value))
-					ThrowValidationException ("Cookies", cookie.Name, cookie.Value);
+				string value = cookie.Value;
+				string name = cookie.Name;
+
+				if (!String.IsNullOrEmpty (value)) {
+#if NET_4_0
+					if (validateRequestNewMode)
+						invalid = !validator.IsValidRequestString (context, value, RequestValidationSource.Cookies, name, out validationFailureIndex);
+					else
+#endif
+						invalid = IsInvalidString (value);
+
+					if (invalid)
+						ThrowValidationException ("Cookies", name, value);
+				}
 			}
 		}
-		
+
 		static void ThrowValidationException (string name, string key, string value)
 		{
 			string v = "\"" + value + "\"";
@@ -1442,9 +1662,19 @@ namespace System.Web {
 		
 			throw new HttpRequestValidationException (msg);
 		}
-		
-		static bool CheckString (string val)
+
+
+		internal static bool IsInvalidString (string val)
 		{
+			int validationFailureIndex;
+
+			return IsInvalidString (val, out validationFailureIndex);
+		}
+
+		internal static bool IsInvalidString (string val, out int validationFailureIndex)
+		{
+			validationFailureIndex = 0;
+
 			int len = val.Length;
 			if (len < 2)
 				return false;
@@ -1456,9 +1686,12 @@ namespace System.Web {
 				if (current == '<' || current == '\xff1c') {
 					if (next == '!' || next < ' '
 					    || (next >= 'a' && next <= 'z')
-					    || (next >= 'A' && next <= 'Z'))
+					    || (next >= 'A' && next <= 'Z')) {
+						validationFailureIndex = idx - 1;
 						return true;
+					}
 				} else if (current == '&' && next == '#') {
+					validationFailureIndex = idx - 1;
 					return true;
 				}
 
@@ -1467,20 +1700,14 @@ namespace System.Web {
 
 			return false;
 		}
-
+		
 		static System.Net.IPAddress [] GetLocalHostAddresses ()
 		{
 			try {
 				string hostName = System.Net.Dns.GetHostName ();
-#if NET_2_0
 				System.Net.IPAddress [] ipaddr = System.Net.Dns.GetHostAddresses (hostName);
-#else
-				System.Net.IPAddress [] ipaddr = System.Net.Dns.GetHostByName (hostName).AddressList;
-#endif
 				return ipaddr;
-			}
-			catch
-			{
+			} catch {
 				return new System.Net.IPAddress[0];
 			}
 		}
