@@ -1,3 +1,4 @@
+
 //
 // Tests for System.Web.UI.Page
 //
@@ -35,15 +36,18 @@ using System.Threading;
 using System.Security.Principal;
 using System.Web;
 using System.Web.UI;
+#if NET_2_0
+using System.Web.UI.Adapters;
+#endif
 using MonoTests.SystemWeb.Framework;
 using MonoTests.stand_alone.WebHarness;
 using System.Web.UI.WebControls;
 using System.Web.UI.HtmlControls;
 using System.Collections;
+using System.Collections.Specialized;
 using System.Net;
 
 namespace MonoTests.System.Web.UI {
-
 	class TestPage : Page {
 
 		private HttpContext ctx;
@@ -109,25 +113,19 @@ namespace MonoTests.System.Web.UI {
 		}
 	}
 
-	[TestFixture]	
+	[TestFixture]
 	public class PageTest {
 
 		[TestFixtureSetUp]
 		public void SetUpTest ()
 		{
-#if VISUAL_STUDIO
-			WebTest.CopyResource (GetType (), "MonoTests.System.Web.UI.WebControls.Resources.PageCultureTest.aspx", "PageCultureTest.aspx");
-			WebTest.CopyResource (GetType (), "MonoTests.System.Web.UI.WebControls.Resources.PageLifecycleTest.aspx", "PageLifecycleTest.aspx");
-			WebTest.CopyResource (GetType (), "MonoTests.System.Web.UI.WebControls.Resources.PageValidationTest.aspx", "PageValidationTest.aspx");
-			WebTest.CopyResource (GetType (), "MonoTests.System.Web.UI.WebControls.Resources.AsyncPage.aspx", "AsyncPage.aspx");
-			WebTest.CopyResource (GetType (), "MonoTests.System.Web.UI.WebControls.Resources.AsyncPage.aspx", "AsyncPage.aspx");
-#else
 			WebTest.CopyResource (GetType (), "PageCultureTest.aspx", "PageCultureTest.aspx");
 			WebTest.CopyResource (GetType (), "PageLifecycleTest.aspx", "PageLifecycleTest.aspx");
 			WebTest.CopyResource (GetType (), "PageValidationTest.aspx", "PageValidationTest.aspx");
 			WebTest.CopyResource (GetType (), "AsyncPage.aspx", "AsyncPage.aspx");
-			WebTest.CopyResource (GetType (), "AsyncPage.aspx", "AsyncPage.aspx");
-#endif
+			WebTest.CopyResource (GetType (), "PageWithAdapter.aspx", "PageWithAdapter.aspx");
+			WebTest.CopyResource (GetType (), "RedirectOnError.aspx", "RedirectOnError.aspx");
+			WebTest.CopyResource (GetType (), "ClearErrorOnError.aspx", "ClearErrorOnError.aspx");
 		}
 
 		[Test]
@@ -167,6 +165,37 @@ namespace MonoTests.System.Web.UI {
 			Assert.IsNotNull (page.HttpContext.Request, "Request");
 		}
 
+		[Test]
+		[ExpectedException (typeof (HttpException))]
+		public void Response_OverridenContext ()
+		{
+			TestPage2 page = new TestPage2 ();
+			Assert.IsNotNull (page.Response, "Response");
+		}
+		
+		[Test]
+		[ExpectedException (typeof (HttpException))]
+		public void Cache_OverridenContext ()
+		{
+			TestPage2 page = new TestPage2 ();
+			Assert.IsNotNull (page.Cache, "Cache");
+		}
+		
+		[Test]
+		[ExpectedException (typeof (HttpException))]
+		public void Session_OverridenContext ()
+		{
+			TestPage2 page = new TestPage2 ();
+			Assert.IsNotNull (page.Session, "Session");
+		}
+
+		[Test]
+		public void Application_OverridenContext ()
+		{
+			TestPage2 page = new TestPage2 ();
+			Assert.IsNull (page.Application, "Application");
+		}
+
 #if NET_2_0
 		[Test]
 		[Category ("NunitWeb")]
@@ -297,8 +326,6 @@ namespace MonoTests.System.Web.UI {
 			HtmlDiff.AssertAreEqual (origHtml, newHtml, "HeaderRenderPreRenderComplete");
 		}
 
-
-
 		public static void CheckHeader (Page p)
 		{
 			Assert.AreEqual ("Untitled Page", p.Title, "CheckHeader#1");
@@ -307,20 +334,19 @@ namespace MonoTests.System.Web.UI {
 			Assert.AreEqual ("Test0", p.Header.Title, "CheckHeader#3");
 			p.Header.Title = "Test";
 			Assert.AreEqual ("Test", p.Title, "CheckHeader#4");
-		}     
+		}
 #endif
 
-
-
-
 #if NET_2_0
 		[Test]
 		[Category ("NunitWeb")]
-		public void Page_ValidationGroup () {
+		public void Page_ValidationGroup ()
+		{
 			new WebTest (PageInvoker.CreateOnLoad (Page_ValidationGroup_Load)).Run ();
 		}
 
-		public static void Page_ValidationGroup_Load (Page page) {
+		public static void Page_ValidationGroup_Load (Page page)
+		{
 			TextBox textbox;
 			BaseValidator val;
 
@@ -356,9 +382,88 @@ namespace MonoTests.System.Web.UI {
 			Assert.AreEqual (1, page.GetValidators ("VG1").Count, "Page_ValidationGroup#5");
 			Assert.AreEqual (1, page.GetValidators ("VG2").Count, "Page_ValidationGroup#6");
 		}
-#endif
-#if NET_2_0
 
+		[Test]
+		[Category("NunitWeb")]
+		public void InitOutputCache_UsesAdapter ()
+		{
+			WebTest t = new WebTest ("PageWithAdapter.aspx");
+			
+			t.Invoker = PageInvoker.CreateOnLoad (InitOutputCache_UsesAdapter_OnLoad);
+			t.Run ();
+		}
+
+		public static void InitOutputCache_UsesAdapter_OnLoad (Page p)
+		{
+			Assert.IsTrue (p.Response.Cache.VaryByHeaders ["header-from-aspx"], 
+				"InitOutputCache_UsesAdapter #1");
+			Assert.IsTrue (p.Response.Cache.VaryByParams ["param-from-aspx"], 
+				"InitOutputCache_UsesAdapter #2");
+			Assert.IsTrue (p.Response.Cache.VaryByHeaders ["header-from-adapter"], 
+				"InitOutputCache_UsesAdapter #3");
+			Assert.IsTrue (p.Response.Cache.VaryByParams ["param-from-adapter"], 
+				"InitOutputCache_UsesAdapter #4");
+		}
+		
+		[Test]
+		[Category("NunitWeb")]
+		public void PageStatePersister_UsesAdapter ()
+		{
+			WebTest t = new WebTest ("PageWithAdapter.aspx");
+			t.Invoker = PageInvoker.CreateOnLoad (PageStatePersister_UsesAdapter_OnLoad);
+			t.Run ();
+		}
+		
+		public static void PageStatePersister_UsesAdapter_OnLoad (Page p)
+		{
+			TestPageWithAdapter pageWithAdapter = (TestPageWithAdapter) p;
+			Assert.IsTrue (pageWithAdapter.PageStatePersister is TestPersister, 
+				"PageStatePersister_UsesAdapter #1");
+		}
+		
+		[Test]
+		[Category("NunitWeb")]
+		public void ScriptUsesAdapter ()
+		{
+			WebTest t = new WebTest ("PageWithAdapter.aspx");
+			string html = t.Run ();
+			Assert.IsTrue(html.IndexOf("var theForm = /* testFormReference */document.forms[") != -1, "ScriptUsesAdapter #1");
+		}
+
+		[Test]
+		[Category("NunitWeb")]
+		public void DeterminePostBackMode_UsesAdapter ()
+		{
+			WebTest t = new WebTest ("PageWithAdapter.aspx");
+			t.Run ();
+			t.Request = new FormRequest (t.Response, "form1");
+			t.Invoker = PageInvoker.CreateOnInit(DeterminePostBackMode_UsesAdapter_OnInit);
+			t.Run ();
+		}
+		
+		public static void DeterminePostBackMode_UsesAdapter_OnInit (Page p)
+		{
+			// We need to use this special version of HtmlInputHidden because the
+			// original class registers itself for event validation in RenderAttributes
+			// which is not called by WebTest after this OnInit handler returns (WebTest
+			// performs a fake postback which bypasses the rendering phase) and it would
+			// result in an event validation failure.
+			HtmlInputHidden h = new TestHtmlInputHidden();
+			h.ID = "DeterminePostBackModeTestField";
+			p.Controls.Add(h);
+			p.Load += new EventHandler(DeterminePostBackMode_UsesAdapter_OnLoad);
+		}
+		
+		public static void DeterminePostBackMode_UsesAdapter_OnLoad(object source, EventArgs args)
+		{
+			Page p = (Page)source;
+			HtmlInputHidden h = (HtmlInputHidden)p.FindControl("DeterminePostBackModeTestField");
+			Assert.AreEqual("DeterminePostBackModeTestValue", h.Value, 
+				"DeterminePostBackMode #1");
+		}
+#endif
+
+#if NET_2_0
 		// This test are testing validation fixture using RequiredFieldValidator for example
 		
 		[Test]
@@ -1031,12 +1136,26 @@ namespace MonoTests.System.Web.UI {
 		}
 
 		[Test]
-		public void PageAdapter ()
+		public void PageAdapterWithNoAdapter ()
 		{
 			Page p = new Page ();
 			Assert.AreEqual (null, p.PageAdapter, "PageAdapter");
 		}
 
+		[Test]
+		public void PageAdapterWithPageAdapter ()
+		{
+			TestPageWithAdapter p = new TestPageWithAdapter ();
+			Assert.AreEqual (p.page_adapter, p.PageAdapter, "PageAdapter");
+		}
+
+		[Test]
+		public void PageAdapterWithControlAdapter ()
+		{
+			TestPageWithControlAdapter p = new TestPageWithControlAdapter ();
+			Assert.AreEqual (null, p.PageAdapter, "PageAdapter");
+		}
+
 		[Test]
 		public void PreviousPage ()
 		{
@@ -1224,6 +1343,24 @@ namespace MonoTests.System.Web.UI {
 
 			#endregion
 		}
+
+		[Test]
+		[Category ("NunitWeb")]
+		public void ClearErrorOnErrorTest ()
+		{
+			WebTest t = new WebTest ("ClearErrorOnError.aspx");
+			string html = t.Run ();
+			Assert.AreEqual (HttpStatusCode.OK, t.Response.StatusCode);
+		}
+
+		[Test]
+		[Category ("NunitWeb")]
+		public void RedirectOnErrorTest ()
+		{
+			WebTest t = new WebTest ("RedirectOnError.aspx");
+			string html = t.Run ();
+			Assert.AreEqual (HttpStatusCode.Found, t.Response.StatusCode);
+		}
 #endif
 
 		[TestFixtureTearDown]
@@ -1232,4 +1369,129 @@ namespace MonoTests.System.Web.UI {
 			WebTest.Unload ();
 		}
 	}
+
+#if NET_2_0
+	class TestHtmlInputHidden : global::System.Web.UI.HtmlControls.HtmlInputHidden
+	{
+		protected override bool LoadPostData (string postDataKey, NameValueCollection postCollection)
+		{
+			string data = postCollection [postDataKey];
+			if (data != null && data != Value) {
+				Value = data;
+				return true;
+			}
+			return false;
+		}
+	}
+	
+	class TestAdapter : global::System.Web.UI.Adapters.PageAdapter
+	{
+		public override StringCollection CacheVaryByParams {
+			get {
+				StringCollection paramNames = new StringCollection();
+				paramNames.AddRange (new string[] {"param-from-adapter"});
+				return paramNames;
+			}
+		}
+
+		public override StringCollection CacheVaryByHeaders {
+			get {
+				StringCollection headerNames = new StringCollection();
+				headerNames.AddRange (new string[] {"header-from-adapter"});
+				return headerNames;
+			}
+		}
+		
+		PageStatePersister persister;
+		public override PageStatePersister GetStatePersister ()
+		{
+			if (persister == null)
+				persister = new TestPersister(Page);
+			return persister;
+		}
+
+		protected override string GetPostBackFormReference (string formId)
+		{
+			return String.Format("/* testFormReference */{0}", 
+				base.GetPostBackFormReference (formId));
+		}
+		
+		public override NameValueCollection DeterminePostBackMode ()
+		{
+			NameValueCollection origRequestValues = base.DeterminePostBackMode ();
+			if (origRequestValues == null)
+				return null;
+			NameValueCollection requestValues = new NameValueCollection ();
+			requestValues.Add (origRequestValues);
+			requestValues ["DeterminePostBackModeTestField"] 
+				= "DeterminePostBackModeTestValue";
+			return requestValues;
+		}
+		
+		internal new void RenderPostBackEvent (HtmlTextWriter w,
+		                                       string target,
+		                                       string argument,
+		                                       string softKeyLabel,
+		                                       string text,
+		                                       string postUrl,
+		                                       string accessKey,
+		                                       bool encode)
+		{
+			base.RenderPostBackEvent (w, target, argument, softKeyLabel, text, postUrl,
+				accessKey, encode);
+		}
+		
+	}
+	
+	class TestPersister : HiddenFieldPageStatePersister
+	{
+		public TestPersister (Page p) : base (p)
+		{
+		}
+	}
+
+	public class TestPageWithAdapter : Page
+	{
+		public global::System.Web.UI.Adapters.PageAdapter page_adapter;
+		
+		public TestPageWithAdapter () : base ()
+		{
+			page_adapter = new TestAdapter ();
+			WebTest t = WebTest.CurrentTest;
+			if (t != null)
+				t.Invoke (this);
+		}
+		
+		protected override global::System.Web.UI.Adapters.ControlAdapter ResolveAdapter ()
+		{
+			return page_adapter;
+		}
+		
+		public new PageStatePersister PageStatePersister {
+			get { return base.PageStatePersister; }
+		}
+	}
+	
+	public class TestControlAdapter : ControlAdapter
+	{
+	}
+
+	public class TestPageWithControlAdapter : Page
+	{
+		private global::System.Web.UI.Adapters.ControlAdapter control_adapter;
+		
+		public TestPageWithControlAdapter () : base ()
+		{
+			control_adapter = new TestControlAdapter ();
+			WebTest t = WebTest.CurrentTest;
+			if (t != null)
+				t.Invoke (this);
+		}
+		
+		protected override global::System.Web.UI.Adapters.ControlAdapter ResolveAdapter ()
+		{
+			return control_adapter;
+		}
+	}
+#endif
 }
