@@ -38,6 +38,8 @@ using System.IO;
 using System.Threading;
 using System.Globalization;
 
+using MonoTests.Common;
+
 namespace MonoTests.System.Web {
 
 	[TestFixture]
@@ -204,9 +206,38 @@ namespace MonoTests.System.Web {
 			Assert.AreEqual (Path.Combine (appBase, "DIR" + Path.DirectorySeparatorChar + "Web.config"),
 				r.MapPath ("Web.config", "~/DIR", false), "test11");
 		}
+	
+		[Test]
+		[Category ("NunitWeb")]
+		[ExpectedException (typeof (ArgumentException))]
+		public void Test_MapPath_InvalidBase_1 ()
+		{
+			WebTest t = new WebTest (new HandlerInvoker (new HandlerDelegate (
+				MapPathDelegate_InvalidBase_1)));
+			t.Run ();
+		}
+
+		static public void MapPathDelegate_InvalidBase_1 ()
+		{
+			HttpContext.Current.Request.MapPath ("Web.config", "something", true);
+		}
+
+		[Test]
+		[Category ("NunitWeb")]
+		[ExpectedException (typeof (ArgumentException))]
+		public void Test_MapPath_InvalidBase_2 ()
+		{
+			WebTest t = new WebTest (new HandlerInvoker (new HandlerDelegate (
+				MapPathDelegate_InvalidBase_2)));
+			t.Run ();
+		}
+
+		static public void MapPathDelegate_InvalidBase_2 ()
+		{
+			HttpContext.Current.Request.MapPath ("Web.config", "something", false);
+		}
 	}
 	
-
 	[TestFixture]
 	public class Test_HttpFakeRequest {
 		class FakeHttpWorkerRequest : HttpWorkerRequest {
@@ -224,6 +255,12 @@ namespace MonoTests.System.Web {
 				return_kind = re;
 			}
 			
+			public override string GetFilePath()
+			{
+				WhereAmI ();
+				return "/tmp/uri.aspx";
+			}
+
 			public override string GetUriPath()
 			{
 				WhereAmI ();
@@ -257,6 +294,8 @@ namespace MonoTests.System.Web {
 					return "PlainString";
 				case 51:
 					return "Plain&Arg=1";
+				case 52:
+					return "Empty=";
 				default:
 					return "GetQueryString";
 				}
@@ -409,6 +448,7 @@ namespace MonoTests.System.Web {
 					case 1: return null;
 					case 2: return "http://www.mono-project.com/test.aspx";
 					case 15: return "http://www.mono-project.com";
+					case 33: return "x";
 					}
 					break;
 				case HttpWorkerRequest.HeaderUserAgent:
@@ -497,8 +537,8 @@ namespace MonoTests.System.Web {
 		}
 		
 		[Test] 
-        [Category ("NotWorking")]
-        public void Test_RequestFields ()
+	[Category ("NotWorking")]
+	public void Test_RequestFields ()
 		{
 			HttpContext c = Cook (1);
 
@@ -654,6 +694,16 @@ namespace MonoTests.System.Web {
 			c = Cook (51);
 			Assert.AreEqual (c.Request.QueryString.ToString (), "Plain&Arg=1", "QTS#2");
 		}
+
+#if NET_2_0
+		[Test]
+		public void QueryString_NullTest ()
+		{
+			HttpRequest req = new HttpRequest ("file.aspx", "http://localhost/file.aspx", null);
+			
+			Assert.AreEqual (req.QueryString.ToString (), "", "QSNT#1");
+		}
+#endif
 		
 		[Test]
 		public void Leading_qm_in_QueryString ()
@@ -733,7 +783,10 @@ namespace MonoTests.System.Web {
 			Assert.AreEqual (null, c.Request.UrlReferrer, "REF1");
 
 			c = Cook (2);
-			Assert.AreEqual ("http://www.mono-project.com/test.aspx", c.Request.UrlReferrer.ToString (), "REF1");			
+			Assert.AreEqual ("http://www.mono-project.com/test.aspx", c.Request.UrlReferrer.ToString (), "REF1");
+
+			c = Cook (33);
+			Assert.AreEqual (null, c.Request.UrlReferrer, "REF1");			
 		}
 		
 
@@ -744,6 +797,13 @@ namespace MonoTests.System.Web {
 			HttpRequest req = c.Request;
 			Assert.AreEqual (0, req.ContentLength, "#01");
 		}
+
+		[Test]
+		public void EmptyQueryValueParams ()
+		{
+			HttpContext c = Cook (52);
+			Assert.AreEqual ("", c.Request.Params["Empty"]);
+		}
 	}
 
 	// This class is defined here to make it easy to create fake
