@@ -22,7 +22,7 @@ namespace MonoTests.SystemWeb.Framework
 	/// <seealso cref="Invoker"/>
 	/// <seealso cref="Run"/>
 	[Serializable]
-	public class WebTest
+	public partial class WebTest
 	{
 		/// <summary>
 		/// Thrown when trying to copy a resource after appdomain was created. Please call
@@ -123,8 +123,9 @@ namespace MonoTests.SystemWeb.Framework
 		/// <seealso cref="MonoTests.SystemWeb.Framework.Response.Body"/>
 		public string Run ()
 		{
+#if !DOTNET
 			SystemWebTestShim.BuildManager.SuppressDebugModeMessages ();
-
+#endif
 			if (Request.Url == null)
 				Request.Url = Invoker.GetDefaultUrl ();
 			_unloadHandler.StartingRequest();
@@ -330,8 +331,10 @@ namespace MonoTests.SystemWeb.Framework
 				
 				if (!resource.StartsWith (namePrefix))
 					continue;
-				
-				CopyResource (type, resource, Path.Combine (targetDir, resource.Substring (namePrefix.Length)));
+				 
+				// The Replace part is for VisualStudio which compiles .resx files despite them being marked as
+				// embedded resources, which breaks the tests.
+				CopyResource (type, resource, Path.Combine (targetDir, resource.Substring (namePrefix.Length).Replace (".remove_extension", String.Empty)));
 			}
 		}
 		
@@ -593,42 +596,18 @@ namespace MonoTests.SystemWeb.Framework
 		public static void CopyResources ()
 		{
 			Type myself = typeof (WebTest);
-
+			
 			CopyResource (myself, "My.ashx", "My.ashx");
 			CopyResource (myself, "Global.asax", "Global.asax");
-#if VISUAL_STUDIO
-			CopyResource (myself,
-				"MonoTests.SystemWeb.Framework.Resources.Web.config",
-				"Web.config");
-			CopyResource (myself,
-				"MonoTests.SystemWeb.Framework.Resources.MyPage.aspx",
-				"MyPage.aspx");
-			CopyResource (myself,
-				"MonoTests.SystemWeb.Framework.Resources.MyPage.aspx.cs",
-				"MyPage.aspx.cs");
-			CopyResource (myself,
-				"MonoTests.SystemWeb.Framework.Resources.MyPageWithMaster.aspx",
-				"MyPageWithMaster.aspx");
-			CopyResource (myself,
-				"MonoTests.SystemWeb.Framework.Resources.My.master",
-				"My.master");
-#else
-#if NET_2_0
-#if INSIDE_SYSTEM_WEB
-			CopyPrefixedResources (myself, "App_GlobalResources/", "App_GlobalResources");
-			CopyPrefixedResources (myself, "App_Code/", "App_Code");
-#endif
-			CopyResource (myself, "Web.mono.config", "Web.config");
 			CopyResource (myself, "MyPage.aspx", "MyPage.aspx");
 			CopyResource (myself, "MyPage.aspx.cs", "MyPage.aspx.cs");
-#else
-			CopyResource (myself, "MyPage_1.1.aspx", "MyPage.aspx");
-			CopyResource (myself, "Web.mono.config.1.1", "Web.config");
-#endif
 			CopyResource (myself, "MyPageWithMaster.aspx", "MyPageWithMaster.aspx");
 			CopyResource (myself, "My.master", "My.master");
-#endif
+
+			CopyResourcesLocal ();
 		}
+
+		static partial void CopyResourcesLocal ();
 #endif
 	}
 }
