@@ -42,10 +42,8 @@ using System.Diagnostics.SymbolStore;
 
 namespace System.Reflection.Emit
 {
-#if NET_2_0
 	[ComVisible (true)]
 	[ComDefaultInterface (typeof (_MethodBuilder))]
-#endif
 	[ClassInterface (ClassInterfaceType.None)]
 	public sealed class MethodBuilder : MethodInfo, _MethodBuilder
 	{
@@ -71,11 +69,7 @@ namespace System.Reflection.Emit
 		private CallingConventions call_conv;
 		private bool init_locals = true;
 		private IntPtr generic_container;
-#if NET_2_0 || BOOTSTRAP_NET_2_0
-		private GenericTypeParameterBuilder[] generic_params;
-#else
-		private Object generic_params; /* so offsets are the same */
-#endif
+		internal GenericTypeParameterBuilder[] generic_params;
 		private Type[] returnModReq;
 		private Type[] returnModOpt;
 		private Type[][] paramModReq;
@@ -122,11 +116,9 @@ namespace System.Reflection.Emit
 			charset = nativeCharset;
 		}
 
-#if NET_2_0
 		public override bool ContainsGenericParameters {
 			get { throw new NotSupportedException (); }
 		}
-#endif
 
 		public bool InitLocals {
 			get {return init_locals;}
@@ -339,12 +331,8 @@ namespace System.Reflection.Emit
 		internal void fixup ()
 		{
 			if (((attrs & (MethodAttributes.Abstract | MethodAttributes.PinvokeImpl)) == 0) && ((iattrs & (MethodImplAttributes.Runtime | MethodImplAttributes.InternalCall)) == 0)) {
-#if NET_2_0
 				// do not allow zero length method body on MS.NET 2.0 (and higher)
-				if (((ilgen == null) || (ILGenerator.Mono_GetCurrentOffset (ilgen) == 0)) && (code == null || code.Length == 0))
-#else
-				if (((ilgen == null) || (ILGenerator.Mono_GetCurrentOffset (ilgen) == 0)) && (code == null))
-#endif
+				if (((ilgen == null) || (ilgen.ILOffset == 0)) && (code == null || code.Length == 0))
 					throw new InvalidOperationException (
 									     String.Format ("Method '{0}.{1}' does not have a method body.",
 											    DeclaringType.FullName, Name));
@@ -415,12 +403,10 @@ namespace System.Reflection.Emit
 							SetLastError = (bool)value;
 						else if (name == "PreserveSig")
 							preserveSig = (bool)value;
-#if NET_1_1
 					else if (name == "BestFitMapping")
 						BestFitMapping = (bool)value;
 					else if (name == "ThrowOnUnmappableChar")
 						ThrowOnUnmappableChar = (bool)value;
-#endif
 					}
 
 					attrs |= MethodAttributes.PinvokeImpl;
@@ -431,11 +417,9 @@ namespace System.Reflection.Emit
 				case "System.Runtime.InteropServices.PreserveSigAttribute":
 					iattrs |= MethodImplAttributes.PreserveSig;
 					return;
-#if NET_2_0
 				case "System.Runtime.CompilerServices.SpecialNameAttribute":
 					attrs |= MethodAttributes.SpecialName;
 					return;
-#endif
 				case "System.Security.SuppressUnmanagedCodeSecurityAttribute":
 					attrs |= MethodAttributes.HasSecurity;
 					break;
@@ -452,9 +436,7 @@ namespace System.Reflection.Emit
 			}
 		}
 
-#if NET_2_0
 		[ComVisible (true)]
-#endif
 		public void SetCustomAttribute (ConstructorInfo con, byte[] binaryAttribute)
 		{
 			if (con == null)
@@ -472,6 +454,7 @@ namespace System.Reflection.Emit
 
 		public void AddDeclarativeSecurity (SecurityAction action, PermissionSet pset)
 		{
+#if !NET_2_1
 			if (pset == null)
 				throw new ArgumentNullException ("pset");
 			if ((action == SecurityAction.RequestMinimum) ||
@@ -496,11 +479,10 @@ namespace System.Reflection.Emit
 
 			permissions [permissions.Length - 1] = new RefEmitPermissionSet (action, pset.ToXml ().ToString ());
 			attrs |= MethodAttributes.HasSecurity;
+#endif
 		}
 
-#if NET_2_0
 		[Obsolete ("An alternate API is available: Emit the MarshalAs custom attribute instead.")]
-#endif
 		public void SetMarshal (UnmanagedMarshal unmanagedMarshal)
 		{
 			RejectIfCreated ();
@@ -551,9 +533,21 @@ namespace System.Reflection.Emit
 			return new NotSupportedException ("The invoked member is not supported in a dynamic module.");
 		}
 
-#if NET_2_0 || BOOTSTRAP_NET_2_0
-		[MethodImplAttribute(MethodImplOptions.InternalCall)]
-		public override extern MethodInfo MakeGenericMethod (params Type [] typeArguments);
+		public override MethodInfo MakeGenericMethod (params Type [] methodInstantiation)
+		{
+			if (!IsGenericMethodDefinition)
+				throw new InvalidOperationException ("Method is not a generic method definition");
+			if (methodInstantiation == null)
+				throw new ArgumentNullException ("methodInstantiation");
+			if (generic_params.Length != methodInstantiation.Length)
+				throw new ArgumentException ("Incorrect length", "methodInstantiation");
+			foreach (Type type in methodInstantiation) {
+				if (type == null)
+					throw new ArgumentNullException ("methodInstantiation");
+			}
+
+			return new MethodOnTypeBuilderInst (this, methodInstantiation);
+		}
 
 		public override bool IsGenericMethodDefinition {
 			get {
@@ -591,11 +585,16 @@ namespace System.Reflection.Emit
 		{
 			if (names == null)
 				throw new ArgumentNullException ("names");
+			if (names.Length == 0)
+				throw new ArgumentException ("names");
 
 			generic_params = new GenericTypeParameterBuilder [names.Length];
-			for (int i = 0; i < names.Length; i++)
-				generic_params [i] = new GenericTypeParameterBuilder (
-					type, this, names [i], i);
+			for (int i = 0; i < names.Length; i++) {
+				string item = names [i];
+				if (item == null)
+					throw new ArgumentNullException ("names");
+				generic_params [i] = new GenericTypeParameterBuilder (type, this, item, i);
+			}
 
 			return generic_params;
 		}
@@ -632,7 +631,6 @@ namespace System.Reflection.Emit
 				return base.Module;
 			}
 		}
-#endif
 
 		void _MethodBuilder.GetIDsOfNames([In] ref Guid riid, IntPtr rgszNames, uint cNames, uint lcid, IntPtr rgDispId)
 		{
