@@ -30,22 +30,14 @@
 // WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 //
 
-#if NET_1_1
-
-#if NET_2_0
 using System.Runtime.InteropServices;
-#endif
 
 namespace System.Runtime.Serialization.Formatters
 {
-#if NET_2_0
 	[ComVisible (true)]
-#endif
 	public enum TypeFilterLevel
 	{
 		Low = 2,
 		Full = 3
 	}
 }
-
-#endif
