@@ -50,19 +50,15 @@ using System.Security.Policy;
 using System.Security.Principal;
 using System.Configuration.Assemblies;
 
-#if NET_2_0
 using System.Collections.Generic;
 using System.Runtime.ConstrainedExecution;
 using System.Text;
-#endif
 
 namespace System {
 
-#if NET_2_0
 	[ComVisible (true)]
 #if !NET_2_1
 	[ComDefaultInterface (typeof (_AppDomain))]
-#endif
 #endif
 	[ClassInterface(ClassInterfaceType.None)]
 #if NET_2_1
@@ -85,7 +81,7 @@ namespace System {
 
 		[ThreadStatic]
 		static Hashtable assembly_resolve_in_progress_refonly;
-#if !NET_2_1 || MONOTOUCH
+#if !MOONLIGHT
 		// CAS
 		private Evidence _evidence;
 		private PermissionSet _granted;
@@ -119,13 +115,13 @@ namespace System {
 			}
 		}
 
-#if NET_2_0 && !NET_2_1
+#if !NET_2_1
 		[MonoTODO]
 		public ApplicationTrust ApplicationTrust {
 			get { throw new NotImplementedException (); }
 		}
 #endif
-#if !NET_2_1 || MONOTOUCH
+#if !MOONLIGHT
 		public string BaseDirectory {
 			get {
 				string path = SetupInformationNoCopy.ApplicationBase;
@@ -178,7 +174,7 @@ namespace System {
 				return getFriendlyName ();
 			}
 		}
-#if !NET_2_1 || MONOTOUCH
+#if !MOONLIGHT
 		public Evidence Evidence {
 			get {
 				// if the host (runtime) hasn't provided it's own evidence...
@@ -227,6 +223,13 @@ namespace System {
 			get { return _granted; }
 		}
 #endif
+
+#if NET_4_0
+		public PermissionSet PermissionSet {
+			get { return _granted ?? (_granted = new PermissionSet (PermissionState.Unrestricted)); }
+		}
+#endif
+
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private static extern AppDomain getCurDomain ();
 		
@@ -252,11 +255,9 @@ namespace System {
 			}
 		}
 
-#if !NET_2_1 || MONOTOUCH
+#if !MOONLIGHT
 
-#if NET_2_0
 		[Obsolete ("AppDomain.AppendPrivatePath has been deprecated. Please investigate the use of AppDomainSetup.PrivateBinPath instead.")]
-#endif
 		[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public void AppendPrivatePath (string path)
 		{
@@ -278,18 +279,14 @@ namespace System {
 			setup.PrivateBinPath = pp + path;
 		}
 
-#if NET_2_0
 		[Obsolete ("AppDomain.ClearPrivatePath has been deprecated. Please investigate the use of AppDomainSetup.PrivateBinPath instead.")]
-#endif
 		[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public void ClearPrivatePath ()
 		{
 			SetupInformationNoCopy.PrivateBinPath = String.Empty;
 		}
 
-#if NET_2_0
 		[Obsolete ("Use AppDomainSetup.ShadowCopyDirectories")]
-#endif
 		[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public void ClearShadowCopyPath ()
 		{
@@ -302,13 +299,11 @@ namespace System {
 			return Activator.CreateComInstanceFrom (assemblyName, typeName);
 		}
 
-#if NET_1_1
 		public ObjectHandle CreateComInstanceFrom (string assemblyFile, string typeName,
 			byte [] hashValue, AssemblyHashAlgorithm hashAlgorithm)
 		{
 			return Activator.CreateComInstanceFrom (assemblyFile, typeName, hashValue ,hashAlgorithm);
 		}
-#endif
 #endif
 
 		public ObjectHandle CreateInstance (string assemblyName, string typeName)
@@ -327,6 +322,9 @@ namespace System {
 			return Activator.CreateInstance (assemblyName, typeName, activationAttributes);
 		}
 
+#if NET_4_0
+		[Obsolete ("Use an overload that does not take an Evidence parameter")]
+#endif
 		public ObjectHandle CreateInstance (string assemblyName, string typeName, bool ignoreCase, BindingFlags bindingAttr,
 		                                    Binder binder, object[] args, CultureInfo culture, object[] activationAttributes,
 		                                    Evidence securityAttributes)
@@ -350,6 +348,9 @@ namespace System {
 			return (oh != null) ? oh.Unwrap () : null;
 		}
 
+#if NET_4_0
+		[Obsolete ("Use an overload that does not take an Evidence parameter")]
+#endif
 		public object CreateInstanceAndUnwrap (string assemblyName, string typeName, bool ignoreCase,
 		                                       BindingFlags bindingAttr, Binder binder, object[] args, CultureInfo culture,
 		                                       object[] activationAttributes, Evidence securityAttributes)
@@ -359,6 +360,47 @@ namespace System {
 			return (oh != null) ? oh.Unwrap () : null;
 		}
 
+#if NET_4_0
+		public ObjectHandle CreateInstance (string assemblyName, string typeName, bool ignoreCase, BindingFlags bindingAttr,
+		                                    Binder binder, object[] args, CultureInfo culture, object[] activationAttributes)
+		{
+			if (assemblyName == null)
+				throw new ArgumentNullException ("assemblyName");
+
+			return Activator.CreateInstance (assemblyName, typeName, ignoreCase, bindingAttr, binder, args,
+				culture, activationAttributes, null);
+		}
+		public object CreateInstanceAndUnwrap (string assemblyName, string typeName, bool ignoreCase,
+		                                       BindingFlags bindingAttr, Binder binder, object[] args, CultureInfo culture,
+		                                       object[] activationAttributes)
+		{
+			ObjectHandle oh = CreateInstance (assemblyName, typeName, ignoreCase, bindingAttr, binder, args,
+				culture, activationAttributes);
+			return (oh != null) ? oh.Unwrap () : null;
+		}
+
+		public ObjectHandle CreateInstanceFrom (string assemblyFile, string typeName, bool ignoreCase,
+		                                        BindingFlags bindingAttr, Binder binder, object[] args, CultureInfo culture,
+		                                        object[] activationAttributes)
+		{
+			if (assemblyFile == null)
+				throw new ArgumentNullException ("assemblyFile");
+
+			return Activator.CreateInstanceFrom (assemblyFile, typeName, ignoreCase, bindingAttr, binder, args,
+			                                     culture, activationAttributes, null);
+		}
+
+		public object CreateInstanceFromAndUnwrap (string assemblyName, string typeName, bool ignoreCase,
+		                                           BindingFlags bindingAttr, Binder binder, object[] args,
+		                                           CultureInfo culture, object[] activationAttributes)
+		{
+			ObjectHandle oh = CreateInstanceFrom (assemblyName, typeName, ignoreCase, bindingAttr, binder, args,
+				culture, activationAttributes);
+
+			return (oh != null) ? oh.Unwrap () : null;
+		}
+#endif
+
 		public ObjectHandle CreateInstanceFrom (string assemblyFile, string typeName)
 		{
 			if (assemblyFile == null)
@@ -375,6 +417,9 @@ namespace System {
 			return Activator.CreateInstanceFrom (assemblyFile, typeName, activationAttributes);
 		}
 
+#if NET_4_0
+		[Obsolete ("Use an overload that does not take an Evidence parameter")]
+#endif
 		public ObjectHandle CreateInstanceFrom (string assemblyFile, string typeName, bool ignoreCase,
 		                                        BindingFlags bindingAttr, Binder binder, object[] args, CultureInfo culture,
 		                                        object[] activationAttributes, Evidence securityAttributes)
@@ -398,6 +443,9 @@ namespace System {
 			return (oh != null) ? oh.Unwrap () : null;
 		}
 
+#if NET_4_0
+		[Obsolete ("Use an overload that does not take an Evidence parameter")]
+#endif
 		public object CreateInstanceFromAndUnwrap (string assemblyName, string typeName, bool ignoreCase,
 		                                           BindingFlags bindingAttr, Binder binder, object[] args,
 		                                           CultureInfo culture, object[] activationAttributes,
@@ -416,6 +464,9 @@ namespace System {
 			return DefineDynamicAssembly (name, access, null, null, null, null, null, false);
 		}
 
+#if NET_4_0
+		[Obsolete ("Declarative security for assembly level is no longer enforced")]
+#endif
 		public AssemblyBuilder DefineDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access, Evidence evidence)
 		{
 			return DefineDynamicAssembly (name, access, null, evidence, null, null, null, false);
@@ -426,12 +477,18 @@ namespace System {
 			return DefineDynamicAssembly (name, access, dir, null, null, null, null, false);
 		}
 
+#if NET_4_0
+		[Obsolete ("Declarative security for assembly level is no longer enforced")]
+#endif
 		public AssemblyBuilder DefineDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access, string dir,
 		                                              Evidence evidence)
 		{
 			return DefineDynamicAssembly (name, access, dir, evidence, null, null, null, false);
 		}
 
+#if NET_4_0
+		[Obsolete ("Declarative security for assembly level is no longer enforced")]
+#endif
 		public AssemblyBuilder DefineDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access,
 		                                              PermissionSet requiredPermissions,
 		                                              PermissionSet optionalPermissions,
@@ -441,6 +498,9 @@ namespace System {
 				refusedPermissions, false);
 		}
 
+#if NET_4_0
+		[Obsolete ("Declarative security for assembly level is no longer enforced")]
+#endif
 		public AssemblyBuilder DefineDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access, Evidence evidence,
 		                                              PermissionSet requiredPermissions,
 		                                              PermissionSet optionalPermissions,
@@ -450,6 +510,9 @@ namespace System {
 				refusedPermissions, false);
 		}
 
+#if NET_4_0
+		[Obsolete ("Declarative security for assembly level is no longer enforced")]
+#endif
 		public AssemblyBuilder DefineDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access, string dir,
 		                                              PermissionSet requiredPermissions,
 		                                              PermissionSet optionalPermissions,
@@ -459,6 +522,9 @@ namespace System {
 				refusedPermissions, false);
 		}
 
+#if NET_4_0
+		[Obsolete ("Declarative security for assembly level is no longer enforced")]
+#endif
 		public AssemblyBuilder DefineDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access, string dir,
 		                                              Evidence evidence,
 		                                              PermissionSet requiredPermissions,
@@ -469,6 +535,9 @@ namespace System {
 				refusedPermissions, false);
 		}
 
+#if NET_4_0
+		[Obsolete ("Declarative security for assembly level is no longer enforced")]
+#endif
 		public AssemblyBuilder DefineDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access, string dir,
 		                                              Evidence evidence,
 		                                              PermissionSet requiredPermissions,
@@ -486,8 +555,10 @@ namespace System {
 			return ab;
 		}
 
-#if NET_2_0
 		// NET 3.5 method
+#if NET_4_0
+		[Obsolete ("Declarative security for assembly level is no longer enforced")]
+#endif
 		public AssemblyBuilder DefineDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access, string dir,
 		                                              Evidence evidence,
 		                                              PermissionSet requiredPermissions,
@@ -503,9 +574,22 @@ namespace System {
 		}
 
 		// NET 3.5 method
-		public AssemblyBuilder DefineDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access, IEnumerable<CustomAttributeBuilder> assemblyAttributes) {
+		public AssemblyBuilder DefineDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access, IEnumerable<CustomAttributeBuilder> assemblyAttributes)
+		{
 			return DefineDynamicAssembly (name, access, null, null, null, null, null, false, assemblyAttributes);
 		}
+
+#if NET_4_0
+		public AssemblyBuilder DefineDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access, string dir, bool isSynchronized, IEnumerable<CustomAttributeBuilder> assemblyAttributes)
+		{
+			return DefineDynamicAssembly (name, access, dir, null, null, null, null, isSynchronized, assemblyAttributes);
+		}
+
+		[MonoLimitation ("The argument securityContextSource is ignored")]
+		public AssemblyBuilder DefineDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access, IEnumerable<CustomAttributeBuilder> assemblyAttributes, SecurityContextSource securityContextSource)
+		{
+			return DefineDynamicAssembly (name, access, assemblyAttributes);
+		}
 #endif
 
 		internal AssemblyBuilder DefineInternalDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access)
@@ -525,34 +609,54 @@ namespace System {
 
 		public int ExecuteAssembly (string assemblyFile)
 		{
-			return ExecuteAssembly (assemblyFile, null, null);
+			return ExecuteAssembly (assemblyFile, (Evidence)null, null);
 		}
 
+#if NET_4_0
+		[Obsolete ("Use an overload that does not take an Evidence parameter")]
+#endif
 		public int ExecuteAssembly (string assemblyFile, Evidence assemblySecurity)
 		{
 			return ExecuteAssembly (assemblyFile, assemblySecurity, null);
 		}
 
+#if NET_4_0
+		[Obsolete ("Use an overload that does not take an Evidence parameter")]
+#endif
 		public int ExecuteAssembly (string assemblyFile, Evidence assemblySecurity, string[] args)
 		{
 			Assembly a = Assembly.LoadFrom (assemblyFile, assemblySecurity);
 			return ExecuteAssemblyInternal (a, args);
 		}
 
+#if NET_4_0
+		[Obsolete ("Use an overload that does not take an Evidence parameter")]
+#endif
 		public int ExecuteAssembly (string assemblyFile, Evidence assemblySecurity, string[] args, byte[] hashValue, AssemblyHashAlgorithm hashAlgorithm)
 		{
 			Assembly a = Assembly.LoadFrom (assemblyFile, assemblySecurity, hashValue, hashAlgorithm);
 			return ExecuteAssemblyInternal (a, args);
 		}
 
+
+#if NET_4_0
+		public int ExecuteAssembly (string assemblyFile, string[] args)
+		{
+			Assembly a = Assembly.LoadFrom (assemblyFile, null);
+			return ExecuteAssemblyInternal (a, args);
+		}
+
+		public int ExecuteAssembly (string assemblyFile, string[] args, byte[] hashValue, AssemblyHashAlgorithm hashAlgorithm)
+		{
+			Assembly a = Assembly.LoadFrom (assemblyFile, null, hashValue, hashAlgorithm);
+			return ExecuteAssemblyInternal (a, args);
+		}
+#endif
+
 		int ExecuteAssemblyInternal (Assembly a, string[] args)
 		{
 			if (a.EntryPoint == null)
-#if NET_2_0
 				throw new MissingMethodException ("Entry point not found in assembly '" + a.FullName + "'.");
-#else
-				throw new COMException ("Unspecified error.", -2147467259);
-#endif
 			return ExecuteAssembly (a, args);
 		}
 
@@ -599,6 +703,9 @@ namespace System {
 			return result;
 		}
 
+#if NET_4_0
+		[Obsolete ("Use an overload that does not take an Evidence parameter")]
+#endif
 		public Assembly Load (AssemblyName assemblyRef, Evidence assemblySecurity)
 		{
 			if (assemblyRef == null)
@@ -655,6 +762,9 @@ namespace System {
 			return Load (assemblyString, null, false);
 		}
 
+#if NET_4_0
+		[Obsolete ("Use an overload that does not take an Evidence parameter")]
+#endif
 		public Assembly Load (string assemblyString, Evidence assemblySecurity)
 		{
 			return Load (assemblyString, assemblySecurity, false);
@@ -687,6 +797,9 @@ namespace System {
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		internal extern Assembly LoadAssemblyRaw (byte[] rawAssembly, byte[] rawSymbolStore, Evidence securityEvidence, bool refonly);
 
+#if NET_4_0
+		[Obsolete ("Use an overload that does not take an Evidence parameter")]
+#endif
 		public Assembly Load (byte[] rawAssembly, byte[] rawSymbolStore, Evidence securityEvidence)
 		{
 			return Load (rawAssembly, rawSymbolStore, securityEvidence, false);
@@ -701,7 +814,10 @@ namespace System {
 			assembly.FromByteArray = true;
 			return assembly;
 		}
-#if !NET_2_1 || MONOTOUCH
+#if !MOONLIGHT
+#if NET_4_0
+		[Obsolete ("AppDomain policy levels are obsolete")]
+#endif
 		[SecurityPermission (SecurityAction.Demand, ControlPolicy = true)]
 		public void SetAppDomainPolicy (PolicyLevel domainPolicy)
 		{
@@ -718,9 +834,7 @@ namespace System {
 			_granted = ps.PermissionSet;
 		}
 
-#if NET_2_0
 		[Obsolete ("Use AppDomainSetup.SetCachePath")]
-#endif
 		[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public void SetCachePath (string path)
 		{
@@ -737,18 +851,14 @@ namespace System {
 			_principal = null;
 		}
 
-#if NET_2_0
 		[Obsolete ("Use AppDomainSetup.ShadowCopyFiles")]
-#endif
 		[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public void SetShadowCopyFiles()
 		{
 			SetupInformationNoCopy.ShadowCopyFiles = "true";
 		}
 
-#if NET_2_0
 		[Obsolete ("Use AppDomainSetup.ShadowCopyDirectories")]
-#endif
 		[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public void SetShadowCopyPath (string path)
 		{
@@ -856,7 +966,7 @@ namespace System {
 			return _process_guid;
 		}
 
-#if !NET_2_1 || MONOTOUCH
+#if !MOONLIGHT
 
 		public static AppDomain CreateDomain (string friendlyName)
 		{
@@ -901,6 +1011,15 @@ namespace System {
 			} else if (info.ConfigurationFile == null)
 				info.ConfigurationFile = "[I don't have a config file]";
 
+#if !NET_2_1
+			if (info.AppDomainInitializer != null) {
+				if (!info.AppDomainInitializer.Method.IsStatic)
+					throw new ArgumentException ("Non-static methods cannot be invoked as an appdomain initializer");
+			}
+#endif
+
+			info.SerializeNonPrimitives ();
+
 			AppDomain ad = (AppDomain) RemotingServices.GetDomainProxy (createDomain (friendlyName, info));
 			if (securityInfo == null) {
 				// get default domain's Evidence (unless we're are the default!)
@@ -912,11 +1031,8 @@ namespace System {
 			else
 				ad._evidence = new Evidence (securityInfo);	// copy
 
-#if NET_2_0 && !NET_2_1
+#if !NET_2_1
 			if (info.AppDomainInitializer != null) {
-				if (!info.AppDomainInitializer.Method.IsStatic)
-					throw new ArgumentException ("Non-static methods cannot be invoked as an appdomain initializer");
-
 				Loader loader = new Loader (
 					info.AppDomainInitializer.Method.DeclaringType.Assembly.Location);
 				ad.DoCallBack (loader.Load);
@@ -931,7 +1047,7 @@ namespace System {
 			return ad;
 		}
 
-#if NET_2_0 && !NET_2_1
+#if !NET_2_1
 		[Serializable]
 		class Loader {
 
@@ -973,7 +1089,7 @@ namespace System {
 			return CreateDomain (friendlyName, securityInfo, CreateDomainSetup (appBasePath, appRelativeSearchPath, shadowCopyFiles));
 		}
 		
-#if NET_2_0 && !NET_2_1
+#if !NET_2_1
 		public static AppDomain CreateDomain (string friendlyName, Evidence securityInfo, AppDomainSetup info,
 		                                      PermissionSet grantSet, params StrongName [] fullTrustAssemblies)
 		{
@@ -995,12 +1111,7 @@ namespace System {
 			if (shadowCopyFiles)
 				info.ShadowCopyFiles = "true";
 			else
-#if NET_2_0
 				info.ShadowCopyFiles = "false";
-#else
-				info.ShadowCopyFiles = null;
-#endif
-
 
 			return info;
 		}
@@ -1025,9 +1136,7 @@ namespace System {
 		}
 
 		[SecurityPermission (SecurityAction.Demand, ControlAppDomain = true)]
-#if NET_2_0
 		[ReliabilityContractAttribute (Consistency.MayCorruptAppDomain, Cer.MayFail)]
-#endif
 		public static void Unload (AppDomain domain)
 		{
 			if (domain == null)
@@ -1040,18 +1149,14 @@ namespace System {
 		[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public extern void SetData (string name, object data);
 
-#if NET_2_0
-		[MonoTODO]
+		[MonoLimitation ("The permission field is ignored")]
 		public void SetData (string name, object data, IPermission permission)
 		{
-			throw new NotImplementedException ();
+			SetData (name, data);
 		}
-#endif
 
 #if !NET_2_1
-#if NET_2_0
 		[Obsolete ("Use AppDomainSetup.DynamicBase")]
-#endif
 		[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public void SetDynamicBase (string path)
 		{
@@ -1059,13 +1164,11 @@ namespace System {
 		}
 #endif // !NET_2_1
 
-#if NET_2_0
 		[Obsolete ("AppDomain.GetCurrentThreadId has been deprecated"
 			+ " because it does not provide a stable Id when managed"
 			+ " threads are running on fibers (aka lightweight"
 			+ " threads). To get a stable identifier for a managed"
 			+ " thread, use the ManagedThreadId property on Thread.'")]
-#endif
 		public static int GetCurrentThreadId ()
 		{
 			return Thread.CurrentThreadId;
@@ -1073,7 +1176,7 @@ namespace System {
 
 		public override string ToString ()
 		{
-#if !NET_2_1 || MONOTOUCH
+#if !MOONLIGHT
 			return getFriendlyName ();
 #else
 			StringBuilder sb = new StringBuilder ("Name:");
@@ -1128,7 +1231,7 @@ namespace System {
 		private Assembly DoAssemblyResolve (string name, bool refonly)
 		{
 			ResolveEventHandler del;
-#if NET_2_0 && !NET_2_1
+#if !NET_2_1
 			if (refonly)
 				del = ReflectionOnlyAssemblyResolve;
 			else
@@ -1250,24 +1353,6 @@ namespace System {
 
 		// End of methods called from the runtime
 		
-#if BOOTSTRAP_WITH_OLDLIB
-		// older MCS/corlib returns:
-		// _AppDomain.cs(138) error CS0592: Attribute 'SecurityPermission' is not valid on this declaration type.
-		// It is valid on 'assembly' 'class' 'constructor' 'method' 'struct'  declarations only.
-		public event AssemblyLoadEventHandler AssemblyLoad;
-
-		public event ResolveEventHandler AssemblyResolve;
-
-		public event EventHandler DomainUnload;
-
-		public event EventHandler ProcessExit;
-
-		public event ResolveEventHandler ResourceResolve;
-
-		public event ResolveEventHandler TypeResolve;
-
-		public event UnhandledExceptionEventHandler UnhandledException;
-#else
 		[method: SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public event AssemblyLoadEventHandler AssemblyLoad;
 
@@ -1288,16 +1373,19 @@ namespace System {
 
 		[method: SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public event UnhandledExceptionEventHandler UnhandledException;
-#endif
 
 #if NET_4_0 || BOOTSTRAP_NET_4_0
 		[MonoTODO]
 		public bool IsHomogenous {
 			get { return true; }
 		}
+
+		[MonoTODO]
+		public bool IsFullyTrusted {
+			get { return true; }
+		}
 #endif
 
-#if NET_2_0
         #pragma warning disable 649
 		private AppDomainManager _domain_manager;
         #pragma warning restore 649
@@ -1306,9 +1394,8 @@ namespace System {
 		public AppDomainManager DomainManager {
 			get { return _domain_manager; }
 		}
-#endif
 
-#if NET_2_0 && (!NET_2_1 || MONOTOUCH)
+#if (!MOONLIGHT)
 
 		public event ResolveEventHandler ReflectionOnlyAssemblyResolve;
 
@@ -1360,14 +1447,20 @@ namespace System {
 
 		public int ExecuteAssemblyByName (string assemblyName)
 		{
-			return ExecuteAssemblyByName (assemblyName, null, null);
+			return ExecuteAssemblyByName (assemblyName, (Evidence)null, null);
 		}
 
+#if NET_4_0
+		[Obsolete ("Use an overload that does not take an Evidence parameter")]
+#endif
 		public int ExecuteAssemblyByName (string assemblyName, Evidence assemblySecurity)
 		{
 			return ExecuteAssemblyByName (assemblyName, assemblySecurity, null);
 		}
 
+#if NET_4_0
+		[Obsolete ("Use an overload that does not take an Evidence parameter")]
+#endif
 		public int ExecuteAssemblyByName (string assemblyName, Evidence assemblySecurity, params string[] args)
 		{
 			Assembly a = Assembly.Load (assemblyName, assemblySecurity);
@@ -1375,6 +1468,9 @@ namespace System {
 			return ExecuteAssemblyInternal (a, args);
 		}
 
+#if NET_4_0
+		[Obsolete ("Use an overload that does not take an Evidence parameter")]
+#endif
 		public int ExecuteAssemblyByName (AssemblyName assemblyName, Evidence assemblySecurity, params string[] args)
 		{
 			Assembly a = Assembly.Load (assemblyName, assemblySecurity);
@@ -1382,6 +1478,22 @@ namespace System {
 			return ExecuteAssemblyInternal (a, args);
 		}
 
+#if NET_4_0
+		public int ExecuteAssemblyByName (string assemblyName, params string[] args)
+		{
+			Assembly a = Assembly.Load (assemblyName, null);
+
+			return ExecuteAssemblyInternal (a, args);
+		}
+
+		public int ExecuteAssemblyByName (AssemblyName assemblyName, params string[] args)
+		{
+			Assembly a = Assembly.Load (assemblyName, null);
+
+			return ExecuteAssemblyInternal (a, args);
+		}
+#endif
+
 		public bool IsDefaultAppDomain ()
 		{
 			return Object.ReferenceEquals (this, DefaultDomain);
@@ -1391,9 +1503,9 @@ namespace System {
 		{
 			return GetAssemblies (true);
 		}
-#endif
 
-#if NET_2_1 && !MONOTOUCH
+#else // MOONLIGHT
+
 		public int ExecuteAssemblyByName (string assemblyName)
 		{
 			// critical code in SL that we're not calling in ML
@@ -1401,7 +1513,7 @@ namespace System {
 		}
 #endif
 
-#if NET_1_1 && !NET_2_1
+#if !NET_2_1
 		void _AppDomain.GetIDsOfNames ([In] ref Guid riid, IntPtr rgszNames, uint cNames, uint lcid, IntPtr rgDispId)
 		{
 			throw new NotImplementedException ();
@@ -1423,5 +1535,42 @@ namespace System {
 			throw new NotImplementedException ();
 		}
 #endif
+
+#if NET_4_0 || MOONLIGHT
+		[MonoTODO ("Currently always returns false")]
+		public bool? IsCompatibilitySwitchSet (string value)
+		{
+			if (value == null)
+				throw new ArgumentNullException ("value");
+			// default (at least for SL4) is to return false for unknown values (can't get a null out of it)
+			return false;
+		}
+
+		[MonoTODO ("Currently always returns false")]
+		public static bool MonitoringIsEnabled { 
+			get { return false; }
+			set { throw new NotImplementedException (); }
+		}
+
+		[MonoTODO]
+		public long MonitoringSurvivedMemorySize {
+			get { throw new NotImplementedException (); }
+		}
+
+		[MonoTODO]
+		public static long MonitoringSurvivedProcessMemorySize {
+			get { throw new NotImplementedException (); }
+		}
+
+		[MonoTODO]
+		public long MonitoringTotalAllocatedMemorySize {
+			get { throw new NotImplementedException (); }
+		}
+
+		[MonoTODO]
+		public TimeSpan MonitoringTotalProcessorTime {
+			get { throw new NotImplementedException (); }
+		}
+#endif
 	}
 }
