@@ -54,6 +54,98 @@ namespace System
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private static extern void get_enum_info (Type enumType, out MonoEnumInfo info);
 
+		//
+		// These comparers are needed because enumerations must be compared
+		// using unsigned values so that negative numbers can be looked up
+		// See bug: #371559
+		//
+		internal static SByteComparer  sbyte_comparer = new SByteComparer ();
+		internal static ShortComparer short_comparer = new ShortComparer ();
+		internal static IntComparer   int_comparer = new IntComparer ();
+		internal static LongComparer  long_comparer = new LongComparer ();
+		
+		internal class SByteComparer : IComparer, System.Collections.Generic.IComparer<sbyte>
+		{
+			public int Compare (object x, object y)
+			{
+				sbyte ix = (sbyte) x;
+				sbyte iy = (sbyte) y;
+				
+				return ((byte) ix) - ((byte) iy);
+			}
+
+			public int Compare (sbyte ix, sbyte iy)
+			{
+				return ((byte) ix) - ((byte) iy);
+			}
+		}
+		
+		internal class ShortComparer : IComparer, System.Collections.Generic.IComparer<short>
+	  	{
+			public int Compare (object x, object y)
+			{
+				short ix = (short) x;
+				short iy = (short) y;
+				
+				return ((ushort) ix) - ((ushort) iy);
+			}
+
+			public int Compare (short ix, short iy)
+			{
+				return ((ushort) ix) - ((ushort) iy);
+			}
+		}
+		
+		internal class IntComparer : IComparer, System.Collections.Generic.IComparer<int>
+		  {
+			public int Compare (object x, object y)
+			{
+				int ix = (int) x;
+				int iy = (int) y;
+
+				if (ix == iy)
+					return 0;
+
+				if (((uint) ix) < ((uint) iy))
+					return -1;
+				return 1;
+			}
+
+			public int Compare (int ix, int iy)
+			{
+				if (ix == iy)
+					return 0;
+
+				if (((uint) ix) < ((uint) iy))
+					return -1;
+				return 1;
+			}
+		}
+
+		internal class LongComparer : IComparer, System.Collections.Generic.IComparer<long>
+		{
+			public int Compare (object x, object y)
+			{
+				long ix = (long) x;
+				long iy = (long) y;
+				
+				if (ix == iy)
+					return 0;
+				if (((ulong) ix) < ((ulong) iy))
+					return -1;
+				return 1;
+			}
+
+			public int Compare (long ix, long iy)
+			{
+				if (ix == iy)
+					return 0;
+				if (((ulong) ix) < ((ulong) iy))
+					return -1;
+				return 1;
+			}
+		}
+			
 		static MonoEnumInfo ()
 		{
 			global_cache_monitor = new object ();
@@ -86,14 +178,27 @@ namespace System
 			/* Threads could die, so keep a global cache too */
 			lock (global_cache_monitor) {
 				if (global_cache.ContainsKey (enumType)) {
-					info = (MonoEnumInfo) global_cache [enumType];
-					cache [enumType] = info;
+					object boxedInfo = global_cache [enumType];
+					cache [enumType] = boxedInfo;
+					info = (MonoEnumInfo)boxedInfo;
 					return;
 				}
 			}
 
 			get_enum_info (enumType, out info);
-			Array.Sort (info.values, info.names);
+
+			IComparer ic = null;
+			Type et = Enum.GetUnderlyingType (enumType);
+			if (et == typeof (int))
+				ic = int_comparer;
+			else if (et == typeof (short))
+				ic = short_comparer;
+			else if (et == typeof (sbyte))
+				ic = sbyte_comparer;
+			else if (et == typeof (long))
+				ic = long_comparer;
+			
+			Array.Sort (info.values, info.names, ic);
 			if (info.names.Length > 50) {
 				info.name_hash = new Hashtable (info.names.Length);
 				for (int i = 0; i <  info.names.Length; ++i)
@@ -107,9 +212,7 @@ namespace System
 	};
 
 	[Serializable]
-#if NET_2_0
 	[ComVisible (true)]
-#endif
 	public abstract class Enum : ValueType, IComparable, IConvertible, IFormattable
 	{
 		protected Enum ()
@@ -124,115 +227,94 @@ namespace System
 
 		bool IConvertible.ToBoolean (IFormatProvider provider)
 		{
-			return Convert.ToBoolean (get_value (), provider);
+			return Convert.ToBoolean (Value, provider);
 		}
 
 		byte IConvertible.ToByte (IFormatProvider provider)
 		{
-			return Convert.ToByte (get_value (), provider);
+			return Convert.ToByte (Value, provider);
 		}
 
 		char IConvertible.ToChar (IFormatProvider provider)
 		{
-			return Convert.ToChar (get_value (), provider);
+			return Convert.ToChar (Value, provider);
 		}
 
 		DateTime IConvertible.ToDateTime (IFormatProvider provider)
 		{
-			return Convert.ToDateTime (get_value (), provider);
+			return Convert.ToDateTime (Value, provider);
 		}
 
 		decimal IConvertible.ToDecimal (IFormatProvider provider)
 		{	
-			return Convert.ToDecimal (get_value (), provider);
+			return Convert.ToDecimal (Value, provider);
 		}
 
 		double IConvertible.ToDouble (IFormatProvider provider)
 		{	
-			return Convert.ToDouble (get_value (), provider);
+			return Convert.ToDouble (Value, provider);
 		}
 
 		short IConvertible.ToInt16 (IFormatProvider provider)
 		{
-			return Convert.ToInt16 (get_value (), provider);
+			return Convert.ToInt16 (Value, provider);
 		}
 
 		int IConvertible.ToInt32 (IFormatProvider provider)
 		{
-			return Convert.ToInt32 (get_value (), provider);
+			return Convert.ToInt32 (Value, provider);
 		}
 
 		long IConvertible.ToInt64 (IFormatProvider provider)
 		{
-			return Convert.ToInt64 (get_value (), provider);
+			return Convert.ToInt64 (Value, provider);
 		}
 
-#if ONLY_1_1
-#pragma warning disable 3019
-		[CLSCompliant (false)]
-#endif
 		sbyte IConvertible.ToSByte (IFormatProvider provider)
 		{
-			return Convert.ToSByte (get_value (), provider);
+			return Convert.ToSByte (Value, provider);
 		}
-#if ONLY_1_1
-#pragma warning restore 3019
-#endif
 
 		float IConvertible.ToSingle (IFormatProvider provider)
 		{
-			return Convert.ToSingle (get_value (), provider);
+			return Convert.ToSingle (Value, provider);
 		}
 
-		object IConvertible.ToType (Type type, IFormatProvider provider)
+		object IConvertible.ToType (Type targetType, IFormatProvider provider)
 		{
-			return Convert.ToType (get_value (), type, provider);
+			if (targetType == null)
+				throw new ArgumentNullException ("targetType");
+			if (targetType == typeof (string))
+				return ToString (provider);
+			return Convert.ToType (Value, targetType, provider, false);
 		}
 
-#if ONLY_1_1
-#pragma warning disable 3019
-		[CLSCompliant (false)]
-#endif
 		ushort IConvertible.ToUInt16 (IFormatProvider provider)
 		{
-			return Convert.ToUInt16 (get_value (), provider);
+			return Convert.ToUInt16 (Value, provider);
 		}
-#if ONLY_1_1
-#pragma warning restore 3019
-#endif
 
-#if ONLY_1_1
-#pragma warning disable 3019
-		[CLSCompliant (false)]
-#endif
 		uint IConvertible.ToUInt32 (IFormatProvider provider)
 		{
-			return Convert.ToUInt32 (get_value (), provider);
+			return Convert.ToUInt32 (Value, provider);
 		}
-#if ONLY_1_1
-#pragma warning restore 3019
-#endif
 
-#if ONLY_1_1
-#pragma warning disable 3019
-		[CLSCompliant (false)]
-#endif
 		ulong IConvertible.ToUInt64 (IFormatProvider provider)
 		{
-			return Convert.ToUInt64 (get_value (), provider);
+			return Convert.ToUInt64 (Value, provider);
 		}
-#if ONLY_1_1
-#pragma warning restore 3019
-#endif
 
 		// <-- End IConvertible methods
 
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private extern object get_value ();
 
-#if NET_2_0
+		// wrap the icall into a property so we don't hav to use the icall everywhere
+		private object Value {
+			get { return get_value (); }
+		}
+
 		[ComVisible (true)]
-#endif
 		public static Array GetValues (Type enumType)
 		{
 			if (enumType == null)
@@ -246,9 +328,7 @@ namespace System
 			return (Array) info.values.Clone ();
 		}
 
-#if NET_2_0
 		[ComVisible (true)]
-#endif
 		public static string[] GetNames (Type enumType)
 		{
 			if (enumType == null)
@@ -262,9 +342,54 @@ namespace System
 			return (string []) info.names.Clone ();
 		}
 
-#if NET_2_0
+		//
+		// The faster, non-boxing version.   It must use the special MonoEnumInfo.xxx_comparers
+		// to ensure that we are perfoming bitwise compares, and not signed compares.
+		//
+		// It also tries to use the non-boxing version of the various Array.BinarySearch methods
+		//
+		static int FindPosition (Type enumType, object value, Array values)
+		{
+			switch (Type.GetTypeCode (GetUnderlyingType (enumType))) {
+			case TypeCode.SByte:
+				sbyte [] sbyte_array = values as sbyte [];
+				return Array.BinarySearch (sbyte_array, (sbyte) value,  MonoEnumInfo.sbyte_comparer);
+
+			case TypeCode.Byte:
+				byte [] byte_array = values as byte [];
+				return Array.BinarySearch (byte_array, (byte) value);
+
+			case TypeCode.Int16:
+				short [] short_array = values as short [];
+				return Array.BinarySearch (short_array, (short)value, MonoEnumInfo.short_comparer);
+
+			case TypeCode.UInt16:
+				ushort [] ushort_array = values as ushort [];
+				return Array.BinarySearch (ushort_array, (ushort)value);
+		
+			case TypeCode.Int32:
+				int[] int_array = values as int[];
+				return Array.BinarySearch (int_array, (int)value, MonoEnumInfo.int_comparer);
+
+			case TypeCode.UInt32:
+				uint[] uint_array = values as uint [];
+				return Array.BinarySearch (uint_array, (uint)value);			
+			
+			case TypeCode.Int64:
+				long [] long_array = values as long [];
+				return Array.BinarySearch (long_array, (long) value,  MonoEnumInfo.long_comparer);
+
+			case TypeCode.UInt64:
+				ulong [] ulong_array = values as ulong [];
+				return Array.BinarySearch (ulong_array, (ulong) value);
+
+			default:
+				// This should never happen
+				return Array.BinarySearch (values, value);
+			}
+		}
+	
 		[ComVisible (true)]
-#endif
 		public static string GetName (Type enumType, object value)
 		{
 			if (enumType == null)
@@ -278,13 +403,12 @@ namespace System
 			MonoEnumInfo info;
 			value = ToObject (enumType, value);
 			MonoEnumInfo.GetInfo (enumType, out info);
-			int i = Array.BinarySearch (info.values, value);
+
+			int i = FindPosition (enumType, value, info.values);
 			return (i >= 0) ? info.names [i] : null;
 		}
 
-#if NET_2_0
 		[ComVisible (true)]
-#endif
 		public static bool IsDefined (Type enumType, object value)
 		{
 			if (enumType == null)
@@ -304,7 +428,8 @@ namespace System
 			} else if ((vType == info.utype) || (vType == enumType)) {
 				value = ToObject (enumType, value);
 				MonoEnumInfo.GetInfo (enumType, out info);
-				return (Array.BinarySearch (info.values, value) >= 0);
+
+				return FindPosition (enumType, value, info.values) >= 0;
 			} else {
 				throw new ArgumentException("The value parameter is not the correct type."
 					+ "It must be type String or the same type as the underlying type"
@@ -315,9 +440,7 @@ namespace System
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private static extern Type get_underlying_type (Type enumType);
 
-#if NET_2_0
 		[ComVisible (true)]
-#endif
 		public static Type GetUnderlyingType (Type enumType)
 		{
 			if (enumType == null)
@@ -329,9 +452,7 @@ namespace System
 			return get_underlying_type (enumType);
 		}
 
-#if NET_2_0
 		[ComVisible (true)]
-#endif
 		public static object Parse (Type enumType, string value)
 		{
 			// Note: Parameters are checked in the other overload
@@ -387,9 +508,7 @@ namespace System
 
 		private static char [] split_char = { ',' };
 
-#if NET_2_0
-		[ComVisible (true)]
-#endif
+		[ComVisible(true)]
 		public static object Parse (Type enumType, string value, bool ignoreCase)
 		{
 			if (enumType == null)
@@ -405,13 +524,26 @@ namespace System
 			if (value.Length == 0)
 				throw new ArgumentException ("An empty string is not considered a valid value.");
 
+			object result;
+			if (!Parse (enumType, value, ignoreCase, out result))
+				throw new ArgumentException (String.Format ("The requested value '{0}' was not found.", value));
+
+			return result;
+		}
+
+		static bool Parse<TEnum> (Type enumType, string value, bool ignoreCase, out TEnum result)
+		{
+			result = default (TEnum);
+
 			MonoEnumInfo info;
 			MonoEnumInfo.GetInfo (enumType, out info);
 
 			// is 'value' a named constant?
 			int loc = FindName (info.name_hash, info.names, value, ignoreCase);
-			if (loc >= 0)
-				return info.values.GetValue (loc);
+			if (loc >= 0) {
+				result = (TEnum) info.values.GetValue (loc);
+				return true;
+			}
 
 			TypeCode typeCode = ((Enum) info.values.GetValue (0)).GetTypeCode ();
 
@@ -422,20 +554,95 @@ namespace System
 				for (int i = 0; i < names.Length; ++i) {
 					loc = FindName (info.name_hash, info.names, names [i].Trim (), ignoreCase);
 					if (loc < 0)
-						throw new ArgumentException ("The requested value was not found.");
+						return false;
+
 					retVal |= GetValue (info.values.GetValue (loc), typeCode);
 				}
-				return ToObject (enumType, retVal);
+				result = (TEnum) ToObject (enumType, retVal);
+				return true;
 			}
 
 			// is 'value' a number?
-			try {
-				return ToObject (enumType, Convert.ChangeType (value, typeCode));
-			} catch (FormatException) {
-				throw new ArgumentException (String.Format ("The requested value '{0}' was not found.", value));
+			switch (typeCode) {
+			case TypeCode.SByte:
+				sbyte sb;
+				if (!SByte.TryParse (value, out sb))
+					return false;
+				result = (TEnum) ToObject (enumType, sb);
+				break;
+			case TypeCode.Byte:
+				byte b;
+				if (!Byte.TryParse (value, out b))
+					return false;
+				result = (TEnum) ToObject (enumType, b);
+				break;
+			case TypeCode.Int16:
+				short i16;
+				if (!Int16.TryParse (value, out i16))
+					return false;
+				result = (TEnum) ToObject (enumType, i16);
+				break;
+			case TypeCode.UInt16:
+				ushort u16;
+				if (!UInt16.TryParse (value, out u16))
+					return false;
+				result = (TEnum) ToObject (enumType, u16);
+				break;
+			case TypeCode.Int32:
+				int i32;
+				if (!Int32.TryParse (value, out i32))
+					return false;
+				result = (TEnum) ToObject (enumType, i32);
+				break;
+			case TypeCode.UInt32:
+				uint u32;
+				if (!UInt32.TryParse (value, out u32))
+					return false;
+				result = (TEnum) ToObject (enumType, u32);
+				break;
+			case TypeCode.Int64:
+				long i64;
+				if (!Int64.TryParse (value, out i64))
+					return false;
+				result = (TEnum) ToObject (enumType, i64);
+				break;
+			case TypeCode.UInt64:
+				ulong u64;
+				if (!UInt64.TryParse (value, out u64))
+					return false;
+				result = (TEnum) ToObject (enumType, u64);
+				break;
+			default:
+				break;
 			}
+
+			return true;
 		}
 
+#if BOOTSTRAP_NET_4_0 || NET_4_0 || MOONLIGHT
+		public static bool TryParse<TEnum> (string value, out TEnum result) where TEnum : struct
+		{
+			return TryParse (value, false, out result);
+		}
+
+		public static bool TryParse<TEnum> (string value, bool ignoreCase, out TEnum result) where TEnum : struct
+		{
+			Type tenum_type = typeof (TEnum);
+			if (!tenum_type.IsEnum)
+				throw new ArgumentException("TEnum is not an Enum type.", "enumType");
+
+			result = default (TEnum);
+
+			if (value == null || value.Trim ().Length == 0)
+				return false;
+
+			return Parse (tenum_type, value, ignoreCase, out result);
+		}
+#endif
+
+		[MethodImplAttribute (MethodImplOptions.InternalCall)]
+		private extern int compare_value_to (object other);
+
 		/// <summary>
 		///   Compares the enum value with another enum value of the same type.
 		/// </summary>
@@ -455,12 +662,7 @@ namespace System
 					target.GetType(), thisType));
 			}
 
-			object value1, value2;
-
-			value1 = this.get_value ();
-			value2 = ((Enum)target).get_value ();
-
-			return ((IComparable)value1).CompareTo (value2);
+			return compare_value_to (target);
 		}
 
 		public override string ToString ()
@@ -468,9 +670,7 @@ namespace System
 			return ToString ("G");
 		}
 
-#if NET_2_0
 		[Obsolete("Provider is ignored, just use ToString")]
-#endif
 		public string ToString (IFormatProvider provider)
 		{
 			return ToString ("G", provider);
@@ -481,12 +681,10 @@ namespace System
 			if (format == String.Empty || format == null)
 				format = "G";
 			
-			return Format (this.GetType (), this.get_value (), format);
+			return Format (this.GetType (), this.Value, format);
 		}
 
-#if NET_2_0
 		[Obsolete("Provider is ignored, just use ToString")]
-#endif
 		public string ToString (String format, IFormatProvider provider)
 		{
 			// provider is not used for Enums
@@ -494,77 +692,59 @@ namespace System
 			if (format == String.Empty || format == null) {
 				format = "G";
 			}
-			return Format (this.GetType(), this.get_value (), format);
+			return Format (this.GetType(), this.Value, format);
 		}
 
-#if NET_2_0
 		[ComVisible (true)]
-#endif
 		public static object ToObject (Type enumType, byte value)
 		{
 			return ToObject (enumType, (object)value);
 		}
 
-#if NET_2_0
 		[ComVisible (true)]
-#endif
 		public static object ToObject (Type enumType, short value)
 		{
 			return ToObject (enumType, (object)value);
 		}
 
-#if NET_2_0
 		[ComVisible (true)]
-#endif
 		public static object ToObject (Type enumType, int value)
 		{
 			return ToObject (enumType, (object)value);
 		}
 
-#if NET_2_0
 		[ComVisible (true)]
-#endif
 		public static object ToObject (Type enumType, long value)
 		{
 			return ToObject (enumType, (object)value);
 		}
 
-#if NET_2_0
 		[ComVisible (true)]
-#endif
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		public static extern object ToObject (Type enumType, object value);
 
-#if NET_2_0
 		[ComVisible (true)]
-#endif
 		[CLSCompliant (false)]
 		public static object ToObject (Type enumType, sbyte value)
 		{
 			return ToObject (enumType, (object)value);
 		}
 
-#if NET_2_0
 		[ComVisible (true)]
-#endif
 		[CLSCompliant (false)]
 		public static object ToObject (Type enumType, ushort value)
 		{
 			return ToObject (enumType, (object)value);
 		}
 
-#if NET_2_0
 		[ComVisible (true)]
-#endif
 		[CLSCompliant (false)]
 		public static object ToObject (Type enumType, uint value)
 		{
 			return ToObject (enumType, (object)value);
 		}
 
-#if NET_2_0
 		[ComVisible (true)]
-#endif
 		[CLSCompliant (false)]
 		public static object ToObject (Type enumType, ulong value)
 		{
@@ -759,9 +939,7 @@ namespace System
 			return retVal;
 		}
 
-#if NET_2_0
 		[ComVisible (true)]
-#endif
 		public static string Format (Type enumType, object value, string format)
 		{
 			if (enumType == null)
@@ -834,5 +1012,14 @@ namespace System
 			}
 			return retVal;
 		}
+#if NET_4_0 || MOONLIGHT
+		public bool HasFlag (Enum flag)
+		{
+			ulong mvalue = Convert.ToUInt64 (get_value (), null);
+			ulong fvalue = Convert.ToUInt64 (flag, null);
+
+			return ((mvalue & fvalue) == fvalue);
+		}
+#endif
 	}
 }
