@@ -10,6 +10,7 @@
 using System;
 using System.Reflection;
 using System.Reflection.Emit;
+using System.Text;
 
 using NUnit.Framework;
 
@@ -331,7 +332,73 @@ namespace MonoTests.System.Reflection.Emit
 			object[] invokeArgs = { "Hello, World!" };
 			object objRet = hello.Invoke (null, invokeArgs);
 			Assert.AreEqual (2, objRet);
-		}			
+		}
+
+		public delegate int IntInvoker();
+
+		public class Foo<T> {
+			public virtual int Test () { return 99; }
+		} 
+
+		[Test]
+		public void ConstrainedPrexixDoesntCrash () //bug #529238
+		{
+			Type foo = typeof (Foo<int>);
+
+			DynamicMethod dm = new DynamicMethod ("Hello", typeof (int), null);
+			ILGenerator ilgen = dm.GetILGenerator ();
+			ilgen.DeclareLocal (foo);
+			ilgen.Emit (OpCodes.Newobj, foo.GetConstructor (new Type [0]));
+			ilgen.Emit (OpCodes.Stloc_0);
+			ilgen.Emit (OpCodes.Ldloca_S, 0);
+			ilgen.Emit (OpCodes.Constrained, foo);
+			ilgen.Emit (OpCodes.Callvirt, foo.GetMethod ("Test"));
+			ilgen.Emit (OpCodes.Ret);
+
+			IntInvoker hi = (IntInvoker) dm.CreateDelegate (typeof (IntInvoker));
+			Assert.AreEqual (99, hi (), "#1");	
+		}
+
+		// #575955
+		[Test]
+		public void Module_GetMethod () {
+			AssemblyName assemblyName = new AssemblyName ();
+			assemblyName.Name = "foo";
+
+			AssemblyBuilder assembly =
+				AppDomain.CurrentDomain.DefineDynamicAssembly (
+															   assemblyName, AssemblyBuilderAccess.RunAndSave);
+
+			ModuleBuilder module = assembly.DefineDynamicModule ("foo.dll");
+
+			var d = new DynamicMethod ("foo", typeof (int), new Type [] { typeof (int[,]) }, module);
+			var ig = d.GetILGenerator ();
+			ig.Emit (OpCodes.Ldarg_0);
+			ig.Emit (OpCodes.Ldc_I4, 1);
+			ig.Emit (OpCodes.Ldc_I4, 1);
+			ig.Emit (OpCodes.Call, module.GetArrayMethod (typeof (int[,]), "Get", CallingConventions.Standard, typeof (int), new Type [] { typeof (int), typeof (int) }));
+			ig.Emit (OpCodes.Ret);
+		
+			var del = (Func<int[,], int>)d.CreateDelegate (typeof (Func<int[,], int>));
+			int[,] arr = new int [10, 10];
+			arr [1, 1] = 5;
+			Assert.AreEqual (5, del (arr));
+		}
+
+		[Test]
+		[Category ("NotWorking")]
+		public void InvalidUnicodeName ()
+		{
+			var name = new StringBuilder ().Append ('\udf45').Append ('\ud808');
+			var method = new DynamicMethod (name.ToString (), typeof (bool), new Type [0]);
+			var il = method.GetILGenerator ();
+			il.Emit (OpCodes.Ldc_I4_1);
+			il.Emit (OpCodes.Ret);
+
+			var function = (Func<bool>) method.CreateDelegate (typeof (Func<bool>));
+
+			Assert.IsTrue (function ());
+		}
 	}
 }
 
