@@ -31,6 +31,7 @@ using NUnit.Framework;
 using System;
 using System.Threading;
 using System.Reflection;
+using System.Reflection.Emit;
 using System.Runtime.InteropServices;
 using System.Runtime.CompilerServices;
 
@@ -38,6 +39,11 @@ using System.Runtime.CompilerServices;
 using System.Collections.Generic;
 #endif
 
+namespace A.B.C {
+	public struct MethodInfoTestStruct {
+		int p;
+	}
+}
 namespace MonoTests.System.Reflection
 {
 	[TestFixture]
@@ -232,7 +238,24 @@ namespace MonoTests.System.Reflection
 			Assert.AreEqual ("Int32* PtrFunc(Int32*)", this.GetType ().GetMethod ("PtrFunc").ToString ());
 		}
 
+
 #if NET_2_0
+		public struct SimpleStruct
+		{
+			int a;
+		}
+
+		public static unsafe SimpleStruct* PtrFunc2 (SimpleStruct* a, A.B.C.MethodInfoTestStruct *b)
+		{
+			return (SimpleStruct*) 0;
+		}
+
+		[Test]
+		public void ToStringWithPointerSignaturesToNonPrimitiveType ()
+		{
+			Assert.AreEqual ("SimpleStruct* PtrFunc2(SimpleStruct*, A.B.C.MethodInfoTestStruct*)", 
+				this.GetType ().GetMethod ("PtrFunc2").ToString ());
+		}	
 		[Test]
 		public void ToStringGenericMethod ()
 		{
@@ -255,6 +278,31 @@ namespace MonoTests.System.Reflection
 			Assert.AreEqual (typeof (GBD_D), typeof (GBD_E).GetMethod ("f").GetBaseDefinition ().DeclaringType);
 		}
 
+		class TestInheritedMethodA {
+			private void TestMethod ()
+			{
+			}
+
+			public void TestMethod2 ()
+			{
+			}
+		}
+
+		class TestInheritedMethodB : TestInheritedMethodA {
+		}
+
+		[Test]
+		public void InheritanceTestGetMethodTest ()
+		{
+			MethodInfo inheritedMethod = typeof(TestInheritedMethodB).GetMethod("TestMethod", BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic);
+			MethodInfo baseMethod = typeof(TestInheritedMethodB).GetMethod("TestMethod", BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic);
+			Assert.AreSame (inheritedMethod, baseMethod);
+
+			MethodInfo inheritedMethod2 = typeof(TestInheritedMethodB).GetMethod("TestMethod2", BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic);
+			MethodInfo baseMethod2 = typeof(TestInheritedMethodB).GetMethod("TestMethod2", BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic);
+			Assert.AreSame (inheritedMethod, baseMethod);
+		}
+
 #if NET_2_0
 #if !TARGET_JVM // MethodBody is not supported for TARGET_JVM
 		[Test]
@@ -479,6 +527,37 @@ namespace MonoTests.System.Reflection
 			}
 		}
 
+		[Test]
+		public void MakeGenericMethodWithNonGenericMethodDefinitionMethod ()
+		{
+			MethodInfo gmi = this.GetType ().GetMethod ("SimpleGenericMethod");
+			MethodInfo inst = gmi.MakeGenericMethod (typeof (int), typeof (double));
+			try {
+				inst.MakeGenericMethod (typeof (int), typeof (double));
+				Assert.Fail ("#1");
+			} catch (InvalidOperationException ex) {
+			}
+		}
+
+		public TFoo SimpleGenericMethod2<TFoo, TBar> () { return default (TFoo); }
+		/*Test for the uggly broken behavior of SRE.*/
+		[Test]
+		public void MakeGenericMethodWithSreTypeResultsInStupidMethodInfo ()
+		{
+			AssemblyName assemblyName = new AssemblyName ();
+			assemblyName.Name = "MonoTests.System.Reflection.Emit.MethodInfoTest";
+			AssemblyBuilder assembly = Thread.GetDomain ().DefineDynamicAssembly (assemblyName, AssemblyBuilderAccess.RunAndSave, ".");
+			ModuleBuilder module = assembly.DefineDynamicModule ("module1", "tst.dll");
+			TypeBuilder tb = module.DefineType ("Test", TypeAttributes.Public);
+
+			MethodInfo gmi = this.GetType ().GetMethod ("SimpleGenericMethod2");
+			MethodInfo ins = gmi.MakeGenericMethod (typeof (int), tb);
+
+			Assert.AreSame (tb, ins.GetGenericArguments () [1], "#1");
+			/*broken ReturnType*/
+			Assert.AreSame (gmi.GetGenericArguments () [0], ins.ReturnType, "#2");
+		}
+
 		public static int? pass_nullable (int? i)
 		{
 			return i;
@@ -591,6 +670,19 @@ namespace MonoTests.System.Reflection
 			Assert.AreSame (mi2, mi4.GetGenericMethodDefinition (), "#C3");
 		}
 
+		public void TestMethod123(int a, int b) {}
+
+		[Test]
+		public void GetParametersDontReturnInternedArray ()
+		{
+			var method = typeof (MethodInfoTest).GetMethod ("TestMethod123");
+			var parms = method.GetParameters ();
+			Assert.AreNotSame (parms, method.GetParameters (), "#1");
+
+			parms [0] = null;
+			Assert.IsNotNull (method.GetParameters () [0], "#2");
+		}
+
 		[Test]
 		public void Bug354757 ()
 		{
@@ -622,6 +714,32 @@ namespace MonoTests.System.Reflection
 			}
 		}
 #endif
+#if NET_4_0
+		interface IMethodInvoke<out T>
+		{
+		    T Test ();
+		}
+
+		class MethodInvoke : IMethodInvoke<string>
+		{
+		    public string Test ()
+		    {
+		        return "MethodInvoke";
+		    }
+		}
+
+		[Test]
+		public void GetInterfaceMapWorksWithVariantIfaces ()
+		{
+			var m0 = typeof (IMethodInvoke<object>).GetMethod ("Test");
+			var m1 = typeof (IMethodInvoke<string>).GetMethod ("Test");
+			var obj = new MethodInvoke ();
+
+			Assert.AreEqual ("MethodInvoke", m0.Invoke (obj, new Object [0]));
+			Assert.AreEqual ("MethodInvoke", m1.Invoke (obj, new Object [0]));
+		}
+#endif
+
 	}
 	
 #if NET_2_0
