@@ -535,8 +535,8 @@ static dis_map_t method_flags_map [] = {
 	{ METHOD_ATTRIBUTE_RT_SPECIAL_NAME,     "rtspecialname " },
 	{ METHOD_ATTRIBUTE_UNMANAGED_EXPORT,    "export " },
 /* MS ilasm doesn't compile this statement - is must be added automagically when permissionset are present */
-/*	{ METHOD_ATTRIBUTE_HAS_SECURITY,        "hassecurity" }, */
-	{ METHOD_ATTRIBUTE_REQUIRE_SEC_OBJECT,  "requiresecobj" },
+	{ METHOD_ATTRIBUTE_HAS_SECURITY,        "" /*"hassecurity"*/ },
+	{ METHOD_ATTRIBUTE_REQUIRE_SEC_OBJECT,  "requiresecobj " },
 	{ METHOD_ATTRIBUTE_PINVOKE_IMPL,        "pinvokeimpl " }, 
 	{ METHOD_ATTRIBUTE_STRICT,	            "strict " }, 
 	{ 0, NULL }
@@ -552,10 +552,11 @@ method_flags (guint32 f)
 {
 	GString *str = g_string_new ("");
 	int access = f & METHOD_ATTRIBUTE_MEMBER_ACCESS_MASK;
+	int rest = f & ~access;
 	char *s;
 	
 	g_string_append (str, map (access, method_access_map));
-	g_string_append (str, flags (f, method_flags_map));
+	g_string_append (str, flags (rest, method_flags_map));
 
 	s = str->str;
 	g_string_free (str, FALSE);
@@ -566,10 +567,10 @@ method_flags (guint32 f)
 static dis_map_t pinvoke_flags_map [] = {
 	{ PINVOKE_ATTRIBUTE_NO_MANGLE ,            "nomangle " },
 	{ PINVOKE_ATTRIBUTE_SUPPORTS_LAST_ERROR,   "lasterr " },
-	{ PINVOKE_ATTRIBUTE_BEST_FIT_ENABLED,      "bestfit:on" },
-	{ PINVOKE_ATTRIBUTE_BEST_FIT_DISABLED,      "bestfit:off" },
-	{ PINVOKE_ATTRIBUTE_THROW_ON_UNMAPPABLE_ENABLED, "charmaperror:on" },
-	{ PINVOKE_ATTRIBUTE_THROW_ON_UNMAPPABLE_DISABLED, "charmaperror:off" },
+	{ PINVOKE_ATTRIBUTE_BEST_FIT_ENABLED,      "bestfit:on " },
+	{ PINVOKE_ATTRIBUTE_BEST_FIT_DISABLED,      "bestfit:off " },
+	{ PINVOKE_ATTRIBUTE_THROW_ON_UNMAPPABLE_ENABLED, "charmaperror:on " },
+	{ PINVOKE_ATTRIBUTE_THROW_ON_UNMAPPABLE_DISABLED, "charmaperror:off " },
 	{ 0, NULL }
 };
 
@@ -602,11 +603,12 @@ pinvoke_flags (guint32 f)
 	GString *str = g_string_new ("");
 	int cset = f & PINVOKE_ATTRIBUTE_CHAR_SET_MASK;
 	int cconv = f & PINVOKE_ATTRIBUTE_CALL_CONV_MASK;
+	int rest = f & ~(cset | cconv);
 	char *s;
 
 	g_string_append (str, map (cset, pinvoke_char_set_map));
 	g_string_append (str, map (cconv, pinvoke_call_conv_map));
-	g_string_append (str, flags (f, pinvoke_flags_map));
+	g_string_append (str, flags (rest, pinvoke_flags_map));
 
 	s = g_strdup(str->str);
 	g_string_free (str, FALSE);
@@ -614,46 +616,6 @@ pinvoke_flags (guint32 f)
 	return s;
 }
 
-static dis_map_t method_impl_map [] = {
-	{ METHOD_IMPL_ATTRIBUTE_IL,              "cil " },
-	{ METHOD_IMPL_ATTRIBUTE_NATIVE,          "native " },
-	{ METHOD_IMPL_ATTRIBUTE_OPTIL,           "optil " },
-	{ METHOD_IMPL_ATTRIBUTE_RUNTIME,         "runtime " },
-	{ 0, NULL }
-};
-
-static dis_map_t managed_type_map [] = {
-	{ METHOD_IMPL_ATTRIBUTE_UNMANAGED,       "unmanaged " },
-	{ METHOD_IMPL_ATTRIBUTE_MANAGED,         "managed " },
-	{ 0, NULL }
-};
-
-static dis_map_t managed_impl_flags [] = {
-	{ METHOD_IMPL_ATTRIBUTE_FORWARD_REF,     "fwdref " },
-	{ METHOD_IMPL_ATTRIBUTE_PRESERVE_SIG,    "preservesig " },
-	{ METHOD_IMPL_ATTRIBUTE_INTERNAL_CALL,   "internalcall " },
-	{ METHOD_IMPL_ATTRIBUTE_SYNCHRONIZED,    "synchronized " },
-	{ METHOD_IMPL_ATTRIBUTE_NOINLINING,      "noinlining " },
-	{ 0, NULL }
-};
-
-static char *
-method_impl_flags (guint32 f)
-{
-	GString *str = g_string_new ("");
-	char *s;
-	int code_type = f & METHOD_IMPL_ATTRIBUTE_CODE_TYPE_MASK;
-	int managed_type = f & METHOD_IMPL_ATTRIBUTE_MANAGED_MASK;
-
-	g_string_append (str, map (code_type, method_impl_map));
-	g_string_append (str, map (managed_type, managed_type_map));
-	g_string_append (str, flags (f, managed_impl_flags));
-	
-	s = str->str;
-	g_string_free (str, FALSE);
-	return s;
-}
-
 static void
 dis_locals (MonoImage *m, MonoMethodHeader *mh, const char *ptr) 
 {
@@ -718,7 +680,8 @@ dis_code (MonoImage *m, guint32 token, guint32 rva, MonoGenericContainer *contai
 	}
 
 	mh = mono_metadata_parse_mh_full (m, container, ptr);
-	if ((entry_point = mono_image_get_entry_point (m)) && mono_metadata_token_index (entry_point)){
+	entry_point = mono_image_get_entry_point (m);
+	if (entry_point && mono_metadata_token_index (entry_point) && mono_metadata_token_table (entry_point) == MONO_TABLE_METHOD) {
 		loc = mono_metadata_locate_token (m, entry_point);
 		if (rva == read32 (loc))
 			fprintf (output, "\t.entrypoint\n");
@@ -884,7 +847,7 @@ dis_method_list (const char *klass_name, MonoImage *m, guint32 start, guint32 en
 		mono_metadata_decode_row (t, i, cols, MONO_METHOD_SIZE);
 
 		flags = method_flags (cols [MONO_METHOD_FLAGS]);
-		impl_flags = method_impl_flags (cols [MONO_METHOD_IMPLFLAGS]);
+		impl_flags = get_method_impl_flags (cols [MONO_METHOD_IMPLFLAGS]);
 
 		sig = mono_metadata_blob_heap (m, cols [MONO_METHOD_SIGNATURE]);
 		mono_metadata_decode_blob_size (sig, &sig);
@@ -896,8 +859,11 @@ dis_method_list (const char *klass_name, MonoImage *m, guint32 start, guint32 en
 			container = type_container;
 
 		ms = mono_metadata_parse_method_signature_full (m, container, i + 1, sig, &sig);
-		sig_str = dis_stringify_method_signature (m, ms, i + 1, container, FALSE);
-		method_name = mono_metadata_string_heap (m, cols [MONO_METHOD_NAME]);
+		if (ms != NULL){
+			sig_str = dis_stringify_method_signature (m, ms, i + 1, container, FALSE);
+			method_name = mono_metadata_string_heap (m, cols [MONO_METHOD_NAME]);
+		} else
+			method_name = g_strdup ("<NULL METHOD SIGNATURE>");
 
 		fprintf (output, "    // method line %d\n", i + 1);
 		fprintf (output, "    .method %s", flags);
@@ -1015,7 +981,7 @@ dis_property_signature (MonoImage *m, guint32 prop_idx, MonoGenericContainer *co
 	if (prop_flags & 0x0400)
 		g_string_append (res, "rtspecialname ");
 	qk = get_escaped_name (name);
-	g_string_sprintfa (res, "%s %s (", blurb, qk);
+	g_string_append_printf (res, "%s %s (", blurb, qk);
 	g_free (qk);
 	g_free (blurb);
 	for (i = 0; i < pcount; i++) {
@@ -1067,7 +1033,7 @@ dis_event_signature (MonoImage *m, guint32 event_idx, MonoGenericContainer *cont
 		g_string_append (res, "specialname ");
 	if (event_flags & 0x0400)
 		g_string_append (res, "rtspecialname ");
-	g_string_sprintfa (res, "%s %s", type, esname);
+	g_string_append_printf (res, "%s %s", type, esname);
 
 	g_free (type);
 	g_free (esname);
@@ -1204,8 +1170,12 @@ dis_type (MonoImage *m, int n, int is_nested, int forward)
 
 	name = mono_metadata_string_heap (m, cols [MONO_TYPEDEF_NAME]);
 	nspace = mono_metadata_string_heap (m, cols [MONO_TYPEDEF_NAMESPACE]);
-	if (*nspace && !is_nested) 
-		fprintf (output, ".namespace %s\n{\n", nspace);
+	if (*nspace && !is_nested) {
+		char *esnspace;
+		esnspace = get_escaped_name (nspace);
+		fprintf (output, ".namespace %s\n{\n", esnspace);
+		g_free (esnspace);
+	}
 
 	container = mono_metadata_load_generic_params (m, MONO_TOKEN_TYPE_DEF | (n + 1), NULL);
 	if (container)
@@ -1217,7 +1187,7 @@ dis_type (MonoImage *m, int n, int is_nested, int forward)
 		
                 param = get_generic_param (m, container);
 		if (param) {
-			fprintf (output, param);
+			fprintf (output, "%s", param);
 			g_free (param);
 		}
                 fprintf (output, "\n");
@@ -1231,7 +1201,7 @@ dis_type (MonoImage *m, int n, int is_nested, int forward)
 
                 param = get_generic_param (m, container);
 		if (param) {
-			fprintf (output, param);
+			fprintf (output, "%s", param);
 			g_free (param);
 		}
 		fprintf (output, "\n");
@@ -1335,8 +1305,7 @@ dis_globals (MonoImage *m)
 
 }
 
-static void
-dis_mresource (MonoImage *m)
+static void dis_resources_worker (MonoImage *m, gboolean just_print)
 {
 	MonoTableInfo *t = &m->tables [MONO_TABLE_MANIFESTRESOURCE];
 	int i;
@@ -1349,10 +1318,21 @@ dis_mresource (MonoImage *m)
 
 		mono_metadata_decode_row (t, i, cols, MONO_MANIFEST_SIZE);
 		name = mono_metadata_string_heap (m, cols [MONO_MANIFEST_NAME]);
-		
-		if (! (res = mono_image_get_resource (m, cols [MONO_MANIFEST_OFFSET], &size)))
-			continue;	
 
+		if (just_print)
+			fprintf (output, "%8x: %s", cols [MONO_MANIFEST_OFFSET], name);
+		
+		if (! (res = mono_image_get_resource (m, cols [MONO_MANIFEST_OFFSET], &size))) {
+			if (just_print)
+				fprintf (output, " (absent from image)\n");
+			continue;	
+		}
+
+		if (just_print) {
+			fprintf (output, " (size %u)\n", size);
+			continue;
+		}
+		
 		if ( (fp = fopen (name, "ab")) ) {
 			if (ftell (fp) == 0)
 				fwrite (res, size, 1, fp);
@@ -1365,6 +1345,107 @@ dis_mresource (MonoImage *m)
 	}		
 }
 
+static void
+dis_mresource (MonoImage *m)
+{
+	dis_resources_worker (m, FALSE);
+}
+
+static void
+dis_presource (MonoImage *m)
+{
+	dis_resources_worker (m, TRUE);
+}
+
+static char *
+exported_type_flags (guint32 flags)
+{
+	static char buffer [1024];
+	int visibility = flags & TYPE_ATTRIBUTE_VISIBILITY_MASK;
+
+	buffer [0] = 0;
+
+	if (flags & TYPE_ATTRIBUTE_FORWARDER) {
+		strcat (buffer, "forwarder ");
+		return buffer;
+	}
+
+	strcat (buffer, map (visibility, visibility_map));
+	return buffer;
+}
+
+static char *
+get_escaped_fullname (MonoImage *m, guint32 nspace_idx, guint32 name_idx)
+{
+	const char *name, *nspace;
+	char *fullname, *esfullname;
+
+	nspace = mono_metadata_string_heap (m, nspace_idx);
+	name = mono_metadata_string_heap (m, name_idx);
+
+	fullname = g_strdup_printf ("%s%s%s", nspace, *nspace ? "." : "", name);
+	esfullname = get_escaped_name (fullname);
+
+	g_free (fullname);
+
+	return esfullname;
+}
+
+static void
+dis_exported_types (MonoImage *m)
+{
+	MonoTableInfo *t = &m->tables [MONO_TABLE_EXPORTEDTYPE];
+	int i;
+
+	for (i = 1; i <= t->rows; i++) {
+		char *fullname;
+		guint32 impl, idx, type_token;
+		guint32 cols [MONO_EXP_TYPE_SIZE];
+
+		mono_metadata_decode_row (t, i - 1, cols, MONO_EXP_TYPE_SIZE);
+
+		fullname = get_escaped_fullname (m, cols [MONO_EXP_TYPE_NAMESPACE], cols [MONO_EXP_TYPE_NAME]);
+
+		fprintf (output, "\n");
+		fprintf (output, ".class extern %s%s\n", exported_type_flags (cols [MONO_EXP_TYPE_FLAGS]), fullname);
+		fprintf (output, "{\n");
+
+		g_free (fullname);
+
+		impl = cols [MONO_EXP_TYPE_IMPLEMENTATION];
+		if (impl) {
+			idx = impl >> MONO_IMPLEMENTATION_BITS;
+			switch (impl & MONO_IMPLEMENTATION_MASK) {
+			case MONO_IMPLEMENTATION_FILE:
+				fprintf (output, "    .file '%s'\n",
+					mono_metadata_string_heap (m, mono_metadata_decode_row_col (&m->tables [MONO_TABLE_FILE], idx - 1, MONO_FILE_NAME)));
+				break;
+			case MONO_IMPLEMENTATION_ASSEMBLYREF:
+				fprintf (output, "    .assembly extern '%s'\n",
+					mono_metadata_string_heap (m, mono_metadata_decode_row_col (&m->tables [MONO_TABLE_ASSEMBLYREF], idx - 1, MONO_ASSEMBLYREF_NAME)));
+				break;
+			case MONO_IMPLEMENTATION_EXP_TYPE:
+				fullname = get_escaped_fullname (
+					m,
+					mono_metadata_decode_row_col (&m->tables [MONO_TABLE_EXPORTEDTYPE], idx - 1, MONO_EXP_TYPE_NAMESPACE),
+					mono_metadata_decode_row_col (&m->tables [MONO_TABLE_EXPORTEDTYPE], idx - 1, MONO_EXP_TYPE_NAME));
+				fprintf (output, "    .class extern %s\n", fullname);
+				g_free (fullname);
+				break;
+			default:
+				g_assert_not_reached ();
+				break;
+			}
+		}
+
+		type_token = cols [MONO_EXP_TYPE_TYPEDEF];
+		if (type_token)
+			fprintf (output, "    .class 0x%08x\n", type_token | MONO_TOKEN_TYPE_DEF);
+
+		fprintf (output, "}\n");
+	}
+}
+
 /**
  * dis_types:
  * @m: metadata context
@@ -1476,6 +1557,7 @@ struct {
 	{ "--moduleref",   MONO_TABLE_MODULEREF,   	dump_table_moduleref },
 	{ "--module",      MONO_TABLE_MODULE,      	dump_table_module },
 	{ "--mresources",  0,	dis_mresource },
+	{ "--presources", 0, dis_presource },
 	{ "--nested",      MONO_TABLE_NESTEDCLASS, 	dump_table_nestedclass },
 	{ "--param",       MONO_TABLE_PARAM,       	dump_table_param },
 	{ "--parconst",    MONO_TABLE_GENERICPARAMCONSTRAINT, dump_table_parconstraint },
@@ -1485,6 +1567,7 @@ struct {
 	{ "--typeref",     MONO_TABLE_TYPEREF,     	dump_table_typeref },
 	{ "--typespec",    MONO_TABLE_TYPESPEC,     	dump_table_typespec },
 	{ "--implmap",     MONO_TABLE_IMPLMAP,     	dump_table_implmap },
+	{ "--fieldrva",    MONO_TABLE_FIELDRVA,     dump_table_fieldrva },
 	{ "--standalonesig", MONO_TABLE_STANDALONESIG,  dump_table_standalonesig },
 	{ "--methodptr", MONO_TABLE_METHOD_POINTER,  dump_table_methodptr },
 	{ "--fieldptr", MONO_TABLE_FIELD_POINTER,  dump_table_fieldptr },
@@ -1532,6 +1615,7 @@ disassemble_file (const char *file)
 		dis_directive_mresource (img);
 		dis_directive_module (img);
 		dis_directive_moduleref (img);
+		dis_exported_types (img);
 		dis_nt_header (img);
                 if (dump_managed_resources)
         		dis_mresource (img);
