@@ -14,14 +14,20 @@
 #include <unistd.h>
 #include <string.h>
 #include <sys/types.h>
-#include <sys/socket.h>
-#include <sys/un.h>
-#include <sys/mman.h>
-#include <dirent.h>
+#ifdef HAVE_SYS_SOCKET_H
+#  include <sys/socket.h>
+#endif
+#ifdef HAVE_SYS_UN_H
+#  include <sys/un.h>
+#endif
+#ifdef HAVE_SYS_MMAN_H
+#  include <sys/mman.h>
+#endif
+#ifdef HAVE_DIRENT_H
+#  include <dirent.h>
+#endif
 #include <sys/stat.h>
 
-#include <mono/os/gc_wrapper.h>
-
 #include <mono/io-layer/wapi.h>
 #include <mono/io-layer/wapi-private.h>
 #include <mono/io-layer/handles-private.h>
@@ -46,7 +52,9 @@ static struct _WapiHandleOps *handle_ops[WAPI_HANDLE_COUNT]={
 	&_wapi_sem_ops,
 	&_wapi_mutex_ops,
 	&_wapi_event_ops,
+#ifndef DISABLE_SOCKETS
 	&_wapi_socket_ops,
+#endif
 	&_wapi_find_ops,
 	&_wapi_process_ops,
 	&_wapi_pipe_ops,
@@ -104,12 +112,35 @@ static guint32 _wapi_private_handle_count = 0;
 static guint32 _wapi_private_handle_slot_count = 0;
 
 struct _WapiHandleSharedLayout *_wapi_shared_layout = NULL;
+
+/*
+ * If SHM is enabled, this will point to shared memory, otherwise it will be NULL.
+ */
 struct _WapiFileShareLayout *_wapi_fileshare_layout = NULL;
 
+/*
+ * If SHM is disabled, this will point to a hash of _WapiFileShare structures, otherwise
+ * it will be NULL. We use this instead of _wapi_fileshare_layout to avoid allocating a
+ * 4MB array.
+ */
+static GHashTable *file_share_hash;
+static CRITICAL_SECTION file_share_hash_mutex;
+
+#define file_share_hash_lock() EnterCriticalSection (&file_share_hash_mutex)
+#define file_share_hash_unlock() LeaveCriticalSection (&file_share_hash_mutex)
+
 guint32 _wapi_fd_reserve;
 
-mono_mutex_t _wapi_global_signal_mutex;
-pthread_cond_t _wapi_global_signal_cond;
+/* 
+ * This is an internal handle which is used for handling waiting for multiple handles.
+ * Threads which wait for multiple handles wait on this one handle, and when a handle
+ * is signalled, this handle is signalled too.
+ */
+static gpointer _wapi_global_signal_handle;
+
+/* Point to the mutex/cond inside _wapi_global_signal_handle */
+mono_mutex_t *_wapi_global_signal_mutex;
+pthread_cond_t *_wapi_global_signal_cond;
 
 int _wapi_sem_id;
 gboolean _wapi_has_shut_down = FALSE;
@@ -121,6 +152,8 @@ gboolean _wapi_has_shut_down = FALSE;
 static pid_t _wapi_pid;
 static mono_once_t pid_init_once = MONO_ONCE_INIT;
 
+static gpointer _wapi_handle_real_new (WapiHandleType type, gpointer handle_specific);
+
 static void pid_init (void)
 {
 	_wapi_pid = getpid ();
@@ -188,8 +221,16 @@ static void handle_cleanup (void)
 	
 	_wapi_shm_semaphores_remove ();
 
-	mono_mutex_destroy(&_wapi_global_signal_mutex);
-	pthread_cond_destroy(&_wapi_global_signal_cond);
+	_wapi_shm_detach (WAPI_SHM_DATA);
+	_wapi_shm_detach (WAPI_SHM_FILESHARE);
+
+	if (file_share_hash) {
+		g_hash_table_destroy (file_share_hash);
+		DeleteCriticalSection (&file_share_hash_mutex);
+	}
+
+	for (i = 0; i < _WAPI_PRIVATE_MAX_SLOTS; ++i)
+		g_free (_wapi_private_handles [i]);
 }
 
 void _wapi_cleanup ()
@@ -206,8 +247,6 @@ void _wapi_cleanup ()
 static mono_once_t shared_init_once = MONO_ONCE_INIT;
 static void shared_init (void)
 {
-	int thr_ret;
-	
 	g_assert ((sizeof (handle_ops) / sizeof (handle_ops[0]))
 		  == WAPI_HANDLE_COUNT);
 	
@@ -235,16 +274,22 @@ static void shared_init (void)
 	_wapi_shared_layout = _wapi_shm_attach (WAPI_SHM_DATA);
 	g_assert (_wapi_shared_layout != NULL);
 	
-	_wapi_fileshare_layout = _wapi_shm_attach (WAPI_SHM_FILESHARE);
-	g_assert (_wapi_fileshare_layout != NULL);
+	if (_wapi_shm_enabled ()) {
+		/* This allocates a 4mb array, so do it only if SHM is enabled */
+		_wapi_fileshare_layout = _wapi_shm_attach (WAPI_SHM_FILESHARE);
+		g_assert (_wapi_fileshare_layout != NULL);
+	}
 	
-	_wapi_collection_init ();
-	
-	thr_ret = pthread_cond_init(&_wapi_global_signal_cond, NULL);
-	g_assert (thr_ret == 0);
-	
-	thr_ret = mono_mutex_init(&_wapi_global_signal_mutex, NULL);
-	g_assert (thr_ret == 0);
+#if !defined (DISABLE_SHARED_HANDLES)
+	if (_wapi_shm_enabled ())
+		_wapi_collection_init ();
+#endif
+
+	/* Can't call wapi_handle_new as it calls us recursively */
+	_wapi_global_signal_handle = _wapi_handle_real_new (WAPI_HANDLE_EVENT, NULL);
+
+	_wapi_global_signal_cond = &_WAPI_PRIVATE_HANDLES (GPOINTER_TO_UINT (_wapi_global_signal_handle)).signal_cond;
+	_wapi_global_signal_mutex = &_WAPI_PRIVATE_HANDLES (GPOINTER_TO_UINT (_wapi_global_signal_handle)).signal_mutex;
 
 	/* Using g_atexit here instead of an explicit function call in
 	 * a cleanup routine lets us cope when a third-party library
@@ -402,16 +447,12 @@ again:
 	return(0);
 }
 
-gpointer _wapi_handle_new (WapiHandleType type, gpointer handle_specific)
+static gpointer _wapi_handle_real_new (WapiHandleType type, gpointer handle_specific)
 {
 	guint32 handle_idx = 0;
 	gpointer handle;
 	int thr_ret;
 	
-	g_assert (_wapi_has_shut_down == FALSE);
-	
-	mono_once (&shared_init_once, shared_init);
-	
 #ifdef DEBUG
 	g_message ("%s: Creating new handle of type %s", __func__,
 		   _wapi_handle_typename[type]);
@@ -464,6 +505,7 @@ gpointer _wapi_handle_new (WapiHandleType type, gpointer handle_specific)
 		ref = _wapi_handle_new_shared (type, handle_specific);
 		if (ref == 0) {
 			_wapi_handle_collect ();
+			_wapi_process_reap ();
 			ref = _wapi_handle_new_shared (type, handle_specific);
 			if (ref == 0) {
 				/* FIXME: grow the arrays */
@@ -483,6 +525,15 @@ done:
 	return(handle);
 }
 
+gpointer _wapi_handle_new (WapiHandleType type, gpointer handle_specific)
+{
+	g_assert (_wapi_has_shut_down == FALSE);
+	
+	mono_once (&shared_init_once, shared_init);
+
+	return _wapi_handle_real_new (type, handle_specific);
+}
+
 gpointer _wapi_handle_new_from_offset (WapiHandleType type, guint32 offset,
 				       gboolean timestamp)
 {
@@ -1511,14 +1562,14 @@ static int timedwait_signal_poll_cond (pthread_cond_t *cond, mono_mutex_t *mutex
 	return(ret);
 }
 
-int _wapi_handle_wait_signal (void)
+int _wapi_handle_wait_signal (gboolean poll)
 {
-	return timedwait_signal_poll_cond (&_wapi_global_signal_cond, &_wapi_global_signal_mutex, NULL, TRUE);
+	return _wapi_handle_timedwait_signal_handle (_wapi_global_signal_handle, NULL, TRUE, poll);
 }
 
-int _wapi_handle_timedwait_signal (struct timespec *timeout)
+int _wapi_handle_timedwait_signal (struct timespec *timeout, gboolean poll)
 {
-	return timedwait_signal_poll_cond (&_wapi_global_signal_cond, &_wapi_global_signal_mutex, timeout, TRUE);
+	return _wapi_handle_timedwait_signal_handle (_wapi_global_signal_handle, timeout, TRUE, poll);
 }
 
 int _wapi_handle_wait_signal_handle (gpointer handle, gboolean alertable)
@@ -1527,11 +1578,11 @@ int _wapi_handle_wait_signal_handle (gpointer handle, gboolean alertable)
 	g_message ("%s: waiting for %p", __func__, handle);
 #endif
 	
-	return _wapi_handle_timedwait_signal_handle (handle, NULL, alertable);
+	return _wapi_handle_timedwait_signal_handle (handle, NULL, alertable, FALSE);
 }
 
 int _wapi_handle_timedwait_signal_handle (gpointer handle,
-					  struct timespec *timeout, gboolean alertable)
+										  struct timespec *timeout, gboolean alertable, gboolean poll)
 {
 #ifdef DEBUG
 	g_message ("%s: waiting for %p (type %s)", __func__, handle,
@@ -1567,10 +1618,62 @@ int _wapi_handle_timedwait_signal_handle (gpointer handle,
 		
 	} else {
 		guint32 idx = GPOINTER_TO_UINT(handle);
-		return timedwait_signal_poll_cond (&_WAPI_PRIVATE_HANDLES(idx).signal_cond, &_WAPI_PRIVATE_HANDLES(idx).signal_mutex, timeout, alertable);
+		int res;
+		pthread_cond_t *cond;
+		mono_mutex_t *mutex;
+
+		if (alertable && !wapi_thread_set_wait_handle (handle))
+			return 0;
+
+		cond = &_WAPI_PRIVATE_HANDLES (idx).signal_cond;
+		mutex = &_WAPI_PRIVATE_HANDLES (idx).signal_mutex;
+
+		if (poll) {
+			/* This is needed when waiting for process handles */
+			res = timedwait_signal_poll_cond (cond, mutex, timeout, alertable);
+		} else {
+			if (timeout)
+				res = mono_cond_timedwait (cond, mutex, timeout);
+			else
+				res = mono_cond_wait (cond, mutex);
+		}
+
+		if (alertable)
+			wapi_thread_clear_wait_handle (handle);
+
+		return res;
 	}
 }
 
+void
+_wapi_free_share_info (_WapiFileShare *share_info)
+{
+	if (!_wapi_shm_enabled ()) {
+		file_share_hash_lock ();
+		g_hash_table_remove (file_share_hash, share_info);
+		file_share_hash_unlock ();
+	} else {
+		memset (share_info, '\0', sizeof(struct _WapiFileShare));
+	}
+}
+
+static gint
+wapi_share_info_equal (gconstpointer ka, gconstpointer kb)
+{
+	const _WapiFileShare *s1 = ka;
+	const _WapiFileShare *s2 = kb;
+
+	return (s1->device == s2->device && s1->inode == s2->inode) ? 1 : 0;
+}
+
+static guint
+wapi_share_info_hash (gconstpointer data)
+{
+	const _WapiFileShare *s = data;
+
+	return s->inode;
+}
+
 gboolean _wapi_handle_get_or_set_share (dev_t device, ino_t inode,
 					guint32 new_sharemode,
 					guint32 new_access,
@@ -1591,55 +1694,36 @@ gboolean _wapi_handle_get_or_set_share (dev_t device, ino_t inode,
 	/* Prevent new entries racing with us */
 	thr_ret = _wapi_shm_sem_lock (_WAPI_SHARED_SEM_FILESHARE);
 	g_assert (thr_ret == 0);
-	
-	/* If a linear scan gets too slow we'll have to fit a hash
-	 * table onto the shared mem backing store
-	 */
-	*share_info = NULL;
-	for (i = 0; i <= _wapi_fileshare_layout->hwm; i++) {
-		file_share = &_wapi_fileshare_layout->share_info[i];
 
-		/* Make a note of an unused slot, in case we need to
-		 * store share info
+	if (!_wapi_shm_enabled ()) {
+		_WapiFileShare tmp;
+
+		/*
+		 * Instead of allocating a 4MB array, we use a hash table to keep track of this
+		 * info. This is needed even if SHM is disabled, to track sharing inside
+		 * the current process.
 		 */
-		if (first_unused == -1 && file_share->handle_refs == 0) {
-			first_unused = i;
-			continue;
+		if (!file_share_hash) {
+			file_share_hash = g_hash_table_new_full (wapi_share_info_hash, wapi_share_info_equal, NULL, g_free);
+			InitializeCriticalSection (&file_share_hash_mutex);
 		}
-		
-		if (file_share->handle_refs == 0) {
-			continue;
-		}
-		
-		if (file_share->device == device &&
-		    file_share->inode == inode) {
+			
+		tmp.device = device;
+		tmp.inode = inode;
+
+		file_share_hash_lock ();
+
+		file_share = g_hash_table_lookup (file_share_hash, &tmp);
+		if (file_share) {
 			*old_sharemode = file_share->sharemode;
 			*old_access = file_share->access;
 			*share_info = file_share;
 			
-			/* Increment the reference count while we
-			 * still have sole access to the shared area.
-			 * This makes the increment atomic wrt
-			 * collections
-			 */
 			InterlockedIncrement ((gint32 *)&file_share->handle_refs);
-			
 			exists = TRUE;
-			break;
-		}
-	}
-	
-	if (!exists) {
-		if (i == _WAPI_FILESHARE_SIZE && first_unused == -1) {
-			/* No more space */
 		} else {
-			if (first_unused == -1) {
-				file_share = &_wapi_fileshare_layout->share_info[++i];
-				_wapi_fileshare_layout->hwm = i;
-			} else {
-				file_share = &_wapi_fileshare_layout->share_info[first_unused];
-			}
-			
+			file_share = g_new0 (_WapiFileShare, 1);
+
 			file_share->device = device;
 			file_share->inode = inode;
 			file_share->opened_by_pid = _wapi_getpid ();
@@ -1647,11 +1731,73 @@ gboolean _wapi_handle_get_or_set_share (dev_t device, ino_t inode,
 			file_share->access = new_access;
 			file_share->handle_refs = 1;
 			*share_info = file_share;
-		}
-	}
 
-	if (*share_info != NULL) {
-		InterlockedExchange ((gint32 *)&(*share_info)->timestamp, now);
+			g_hash_table_insert (file_share_hash, file_share, file_share);
+		}
+
+		file_share_hash_unlock ();
+	} else {
+		/* If a linear scan gets too slow we'll have to fit a hash
+		 * table onto the shared mem backing store
+		 */
+		*share_info = NULL;
+		for (i = 0; i <= _wapi_fileshare_layout->hwm; i++) {
+			file_share = &_wapi_fileshare_layout->share_info[i];
+
+			/* Make a note of an unused slot, in case we need to
+			 * store share info
+			 */
+			if (first_unused == -1 && file_share->handle_refs == 0) {
+				first_unused = i;
+				continue;
+			}
+		
+			if (file_share->handle_refs == 0) {
+				continue;
+			}
+		
+			if (file_share->device == device &&
+				file_share->inode == inode) {
+				*old_sharemode = file_share->sharemode;
+				*old_access = file_share->access;
+				*share_info = file_share;
+			
+				/* Increment the reference count while we
+				 * still have sole access to the shared area.
+				 * This makes the increment atomic wrt
+				 * collections
+				 */
+				InterlockedIncrement ((gint32 *)&file_share->handle_refs);
+			
+				exists = TRUE;
+				break;
+			}
+		}
+	
+		if (!exists) {
+			if (i == _WAPI_FILESHARE_SIZE && first_unused == -1) {
+				/* No more space */
+			} else {
+				if (first_unused == -1) {
+					file_share = &_wapi_fileshare_layout->share_info[++i];
+					_wapi_fileshare_layout->hwm = i;
+				} else {
+					file_share = &_wapi_fileshare_layout->share_info[first_unused];
+				}
+			
+				file_share->device = device;
+				file_share->inode = inode;
+				file_share->opened_by_pid = _wapi_getpid ();
+				file_share->sharemode = new_sharemode;
+				file_share->access = new_access;
+				file_share->handle_refs = 1;
+				*share_info = file_share;
+			}
+		}
+
+		if (*share_info != NULL) {
+			InterlockedExchange ((gint32 *)&(*share_info)->timestamp, now);
+		}
 	}
 	
 	thr_ret = _wapi_shm_sem_unlock (_WAPI_SHARED_SEM_FILESHARE);
@@ -1682,6 +1828,7 @@ static void _wapi_handle_check_share_by_pid (struct _WapiFileShare *share_info)
 	}
 }
 
+#ifdef __linux__
 /* Scan /proc/<pids>/fd/ for open file descriptors to the file in
  * question.  If there are none, reset the share info.
  *
@@ -1815,6 +1962,29 @@ done:
 
 	_wapi_handle_unlock_shared_handles ();
 }
+#else
+//
+// Other implementations (non-Linux)
+//
+void _wapi_handle_check_share (struct _WapiFileShare *share_info, int fd)
+{
+	int thr_ret;
+	
+	/* Prevents entries from expiring under us if we remove this
+	 * one */
+	thr_ret = _wapi_handle_lock_shared_handles ();
+	g_assert (thr_ret == 0);
+	
+	/* Prevent new entries racing with us */
+	thr_ret = _wapi_shm_sem_lock (_WAPI_SHARED_SEM_FILESHARE);
+	g_assert (thr_ret == 0);
+	
+	_wapi_handle_check_share_by_pid (share_info);
+
+	thr_ret = _wapi_shm_sem_unlock (_WAPI_SHARED_SEM_FILESHARE);
+	_wapi_handle_unlock_shared_handles ();
+}
+#endif
 
 void _wapi_handle_dump (void)
 {
