@@ -26,9 +26,6 @@ MonoBoolean ves_icall_System_ConsoleDriver_SetBreak (MonoBoolean want_break) MON
 MonoBoolean ves_icall_System_ConsoleDriver_TtySetup (MonoString *keypad, MonoString *teardown, MonoArray **control_characters, int **size) MONO_INTERNAL;
 void ves_icall_System_ConsoleDriver_Suspend (void) MONO_INTERNAL;
 
-void console_restore_signal_handlers (void);
-void console_set_signal_handlers (void);
-
 G_END_DECLS
 
 #endif /* _MONO_METADATA_CONSOLEIO_H */
